Type: Package
Package: nominatimlite
Title: Interface to the 'Nominatim' API
Version: 0.6.0
Authors@R: c(
    person("Diego", "Hernangómez", , "diego.hernangomezherrero@gmail.com", role = c("aut", "cre", "cph"),
           comment = c(ORCID = "0000-0001-8457-4658")),
    person("Jindra", "Lacko", role = c("ctb", "rev"),
           comment = c(ORCID = "0000-0002-0375-5156")),
    person("Alex", "White", role = "ctb"),
    person("OpenStreetMap", role = "cph",
           comment = "For the data")
  )
Description: Lightweight interface to the 'OpenStreetMap' 'Nominatim' API
    <https://nominatim.org/release-docs/latest/>. Extract coordinates from
    addresses, retrieve addresses from coordinates, look up amenities and
    addresses, and return results as 'tibble' or 'sf' objects.
License: MIT + file LICENSE
URL: https://dieghernan.github.io/nominatimlite/,
        https://github.com/dieghernan/nominatimlite
BugReports: https://github.com/dieghernan/nominatimlite/issues
Depends: R (>= 4.1.0)
Imports: dplyr (>= 1.0.0), jsonlite (>= 1.7.0), sf (>= 0.9.0), tools,
        utils
Suggests: arcgeocoder, ggplot2 (>= 3.0.0), knitr, quarto, testthat (>=
        3.0.0), tibble (>= 3.0.0), tidygeocoder
VignetteBuilder: quarto
Config/Needs/website: dieghernan/gitdevr, pak, tidyverse, leaflet,
        reactable, crosstalk, tidyr, htmltools
Config/roxygen2/markdown: TRUE
Config/roxygen2/version: 8.0.0
Config/testthat/edition: 3
Config/testthat/parallel: false
Copyright: Data © OpenStreetMap contributors, ODbL 1.0.
        <https://www.openstreetmap.org/copyright>
Encoding: UTF-8
LazyData: true
X-schema.org-applicationCategory: cartography
X-schema.org-keywords: r, geocoding, openstreetmap, address, nominatim,
        reverse-geocoding, rstats, shapefile, r-package, spatial, cran,
        api-wrapper, api, gis, geocoder, reverse-geocoder
NeedsCompilation: no
Packaged: 2026-06-02 15:38:13 UTC; diego
Author: Diego Hernangómez [aut, cre, cph] (ORCID:
    <https://orcid.org/0000-0001-8457-4658>),
  Jindra Lacko [ctb, rev] (ORCID:
    <https://orcid.org/0000-0002-0375-5156>),
  Alex White [ctb],
  OpenStreetMap [cph] (For the data)
Maintainer: Diego Hernangómez <diego.hernangomezherrero@gmail.com>
Repository: CRAN
Date/Publication: 2026-06-03 05:10:09 UTC
Built: R 4.6.0; ; 2026-06-03 08:44:56 UTC; unix

Type: Package
Package: nominatimlite
Title: Interface to the 'Nominatim' API
Version: 0.7.0
Authors@R: c(
    person("Diego", "Hernangómez", , "diego.hernangomezherrero@gmail.com", role = c("aut", "cre", "cph"),
           comment = c(ORCID = "0000-0001-8457-4658")),
    person("Jindra", "Lacko", role = c("ctb", "rev"),
           comment = c(ORCID = "0000-0002-0375-5156")),
    person("Alex", "White", role = "ctb"),
    person("OpenStreetMap contributors", role = "cph",
           comment = "For the osm_amenities data")
  )
Description: Provides a lightweight interface to the 'Nominatim' API
    <https://nominatim.org/release-docs/latest/>. It supports free-form
    and structured address searches, searches for addresses from
    coordinates, amenity lookup and address lookup by 'OpenStreetMap'
    object identifier. It returns results as 'tibble' data frames or 'sf'
    objects.
License: MIT + file LICENSE
URL: https://dieghernan.github.io/nominatimlite/,
        https://github.com/dieghernan/nominatimlite
BugReports: https://github.com/dieghernan/nominatimlite/issues
Depends: R (>= 4.1.0)
Imports: dplyr (>= 1.0.0), jsonlite (>= 1.7.0), sf (>= 0.9.0), tools,
        utils
Suggests: curl, ggplot2, knitr, quarto, testthat (>= 3.3.0), tibble,
        withr
VignetteBuilder: quarto
Config/Needs/website: dieghernan/gitdevr, pak, tidyverse, leaflet,
        reactable, crosstalk, tidyr, htmltools, arcgeocoder,
        tidygeocoder
Config/roxygen2/markdown: TRUE
Config/roxygen2/version: 8.1.0
Config/testthat/edition: 3
Config/testthat/parallel: false
Copyright: See file inst/COPYRIGHTS.
Encoding: UTF-8
LazyData: true
X-schema.org-applicationCategory: cartography
X-schema.org-keywords: r, openstreetmap, address, coordinates,
        nominatim, reverse-search, rstats, r-package, cran,
        api-wrapper, api, geocoder, geocoding, gis, reverse-geocoder,
        reverse-geocoding, shapefile, spatial
NeedsCompilation: no
Packaged: 2026-08-25 10:16:18 UTC; diego
Author: Diego Hernangómez [aut, cre, cph] (ORCID:
    <https://orcid.org/0000-0001-8457-4658>),
  Jindra Lacko [ctb, rev] (ORCID:
    <https://orcid.org/0000-0002-0375-5156>),
  Alex White [ctb],
  OpenStreetMap contributors [cph] (For the osm_amenities data)
Maintainer: Diego Hernangómez <diego.hernangomezherrero@gmail.com>
Repository: CRAN
Date/Publication: 2026-08-25 11:20:18 UTC

# Generated by roxygen2: do not edit by hand

export(bbox_to_poly)
export(geo_address_lookup)
export(geo_address_lookup_sf)
export(geo_amenity)
export(geo_amenity_sf)
export(geo_lite)
export(geo_lite_sf)
export(geo_lite_struct)
export(geo_lite_struct_sf)
export(nominatim_check_access)
export(reverse_geo_lite)
export(reverse_geo_lite_sf)
importFrom(utils,download.file)
importFrom(utils,setTxtProgressBar)
importFrom(utils,txtProgressBar)

<!-- README.md is generated from README.qmd. Please edit that file -->

# nominatimlite <a href="https://dieghernan.github.io/nominatimlite/"><img src="man/figures/logo.png" alt="nominatimlite website" align="right" height="139"/></a>

<!-- badges: start -->

[![CRAN
status](https://www.r-pkg.org/badges/version/nominatimlite)](https://CRAN.R-project.org/package=nominatimlite)
[![CRAN
results](https://badges.cranchecks.info/worst/nominatimlite.svg)](https://cran.r-project.org/web/checks/check_results_nominatimlite.html)
[![Downloads](https://cranlogs.r-pkg.org/badges/nominatimlite)](https://CRAN.R-project.org/package=nominatimlite)
[![Nominatim-version](https://img.shields.io/badge/Nominatim-%E2%89%A5_4.4.0-blue)](https://nominatim.org/release-docs/develop/api/Overview/)
[![R-CMD-check](https://github.com/dieghernan/nominatimlite/actions/workflows/check-full.yaml/badge.svg)](https://github.com/dieghernan/nominatimlite/actions/workflows/check-full.yaml)
[![codecov](https://codecov.io/gh/dieghernan/nominatimlite/branch/main/graph/badge.svg?token=jSZ4RIsj91)](https://app.codecov.io/gh/dieghernan/nominatimlite)
[![CodeFactor](https://www.codefactor.io/repository/github/dieghernan/nominatimlite/badge)](https://www.codefactor.io/repository/github/dieghernan/nominatimlite)
[![r-universe](https://dieghernan.r-universe.dev/badges/nominatimlite)](https://dieghernan.r-universe.dev/nominatimlite)
[![Project Status: Active – The project has reached a stable, usable
state and is being actively
developed.](https://www.repostatus.org/badges/latest/active.svg)](https://www.repostatus.org/#active)
[![DOI](https://img.shields.io/badge/DOI-10.32614/CRAN.package.nominatimlite-blue)](https://doi.org/10.32614/CRAN.package.nominatimlite)
[![status](https://tinyverse.netlify.app/status/nominatimlite)](https://CRAN.R-project.org/package=nominatimlite)

<!-- badges: end -->

The goal of **nominatimlite** is to provide a lightweight interface for
geocoding addresses with the [Nominatim
API](https://nominatim.org/release-docs/latest/). It also allows you to
return results as `sf` objects using the **sf** package.

The full site with examples and vignettes is available at
<https://dieghernan.github.io/nominatimlite/>

## What is Nominatim?

**Nominatim** is a tool for searching
[OpenStreetMap](https://www.openstreetmap.org/) data by name and address
([geocoding](https://wiki.openstreetmap.org/wiki/Geocoding "Geocoding"))
and generating synthetic addresses for OSM points (reverse geocoding).

## Why nominatimlite?

**nominatimlite** accesses the Nominatim API without depending on
**curl**. In some situations, **curl** may not be available or
accessible, so **nominatimlite** uses base R functions instead.

## Recommended packages

Other packages are more complete and mature than **nominatimlite** and
provide similar features:

- [**tidygeocoder**](https://jessecambon.github.io/tidygeocoder/)
  ([Cambon et al. 2021](#ref-R-tidygeocoder)): Provides an interface to
  geocoding services such as Nominatim, Google, TomTom and Mapbox.
- [**osmdata**](https://docs.ropensci.org/osmdata/) ([Padgham et al.
  2017](#ref-R-osmdata)): Downloads spatial data from OpenStreetMap with
  the [Overpass API](https://wiki.openstreetmap.org/wiki/Overpass_API).
- [**arcgeocoder**](https://dieghernan.github.io/arcgeocoder/)
  ([Hernangómez 2024](#ref-R-arcgeocoder)): Provides a lightweight
  interface for geocoding with the ArcGIS REST API service.

## Installation

<div class="pkgdown-release">

Install **nominatimlite** from
[**CRAN**](https://CRAN.R-project.org/package=nominatimlite):

``` r
install.packages("nominatimlite")
```

</div>

<div class="pkgdown-devel">

See the documentation for the development version at
<https://dieghernan.github.io/nominatimlite/dev/>.

You can install the development version of **nominatimlite** with:

``` r
pak::pak("dieghernan/nominatimlite")
```

Alternatively, you can install **nominatimlite** using the
[r-universe](https://dieghernan.r-universe.dev/nominatimlite):

``` r
# Install nominatimlite in R:
install.packages(
  "nominatimlite",
  repos = c(
    "https://dieghernan.r-universe.dev",
    "https://cloud.r-project.org"
  )
)
```

</div>

## Usage

### `sf` objects

With **nominatimlite** you can return `sf` objects:

``` r
library(nominatimlite)

# Search for Pizza Hut locations in California.

CA <- geo_lite_sf("California", points_only = FALSE)

pizzahut <- geo_lite_sf(
  "Pizza Hut, California",
  limit = 50,
  custom_query = list(countrycodes = "us")
)

library(ggplot2)

ggplot(CA) +
  geom_sf() +
  geom_sf(data = pizzahut, col = "red")
```

<img src="man/figures/README-pizzahut-1.png" style="width:100.0%"
alt="Pizza Hut restaurant locations in California extracted with nominatimlite." />

You can also return polygon and line objects when the Nominatim API
provides them, using the option `points_only = FALSE`:

``` r
# A building, returned as a polygon.
sol_poly <- geo_lite_sf("Statue of Liberty, NY, USA", points_only = FALSE)

ggplot(sol_poly) +
  geom_sf()
```

<img src="man/figures/README-statue_liberty-1.png" style="width:100.0%"
alt="Location of the Statue of Liberty extracted with nominatimlite." />

``` r
# Default, returned as a point.
dayton <- geo_lite_sf("Dayton, OH")
# A US state, returned as a polygon.
ohio_state <- geo_lite_sf("Ohio, USA", points_only = FALSE)
# A river, returned as a line.
ohio_river <- geo_lite_sf("Ohio river", points_only = FALSE)

ggplot() +
  geom_sf(data = ohio_state) +
  geom_sf(data = dayton, color = "red", pch = 4) +
  geom_sf(data = ohio_river, color = "blue")
```

<img src="man/figures/README-line-object-1.png" style="width:100.0%"
alt="Different features named Ohio extracted with nominatimlite." />

### Geocoding and reverse geocoding

*Note: examples are adapted from the **tidygeocoder** package.*

In this first example, we geocode a few addresses with `geo_lite()`:

``` r
library(tibble)

# Create a data frame with addresses.
some_addresses <- tribble(
  ~name, ~addr,
  "White House", "1600 Pennsylvania Ave NW, Washington, DC",
  "Transamerica Pyramid", "600 Montgomery St, San Francisco, CA 94111",
  "Willis Tower", "233 S Wacker Dr, Chicago, IL 60606"
)

# Geocode the addresses.
lat_longs <- geo_lite(
  some_addresses$addr,
  lat = "latitude",
  long = "longitude",
  progressbar = FALSE
)
```

This example returns only latitude, longitude and address columns from
the Nominatim API. Use `full_results = TRUE` to return all available
data from the Nominatim API.

| query | latitude | longitude | address |
|:---|---:|---:|:---|
| 1600 Pennsylvania Ave NW, Washington, DC | 38.89764 | -77.03655 | White House, 1600, Pennsylvania Avenue Northwest, Ward 2, Washington, District of Columbia, 20500, United States |
| 600 Montgomery St, San Francisco, CA 94111 | 37.79519 | -122.40279 | Transamerica Pyramid, 600, Montgomery Street, Financial District, South of Market, San Francisco, California, 94111, United States |
| 233 S Wacker Dr, Chicago, IL 60606 | 41.87874 | -87.63596 | Willis Tower, 233, South Wacker Drive, Financial District, Loop, Chicago, South Chicago Township, Cook County, Illinois, 60606, United States |

<p class="caption">

Table 1: Example: geocoding addresses.
</p>

To perform reverse geocoding, use `reverse_geo_lite()` to obtain
addresses from geographic coordinates. The arguments are similar to
`geo_lite()`, but now we provide coordinate values with the `lat` and
`long` arguments. The dataset used here is from the geocoding query
above. The single-line address is returned in a column named with the
`address` argument.

``` r
reverse <- reverse_geo_lite(
  lat = lat_longs$latitude,
  long = lat_longs$longitude,
  address = "address_found",
  progressbar = FALSE
)
```

| address_found | lat | lon |
|:---|---:|---:|
| White House, 1600, Pennsylvania Avenue Northwest, Downtown, Ward 2, Washington, District of Columbia, 20500, United States | 38.89764 | -77.03655 |
| Sky Bar, Mark Twain Place, Financial District, South of Market, San Francisco, California, 94111, United States | 37.79519 | -122.40254 |
| West Adams Street, Financial District, Loop, Chicago, South Chicago Township, Cook County, Illinois, 60675, United States | 41.87874 | -87.63589 |

<p class="caption">

Table 2: Example: reverse geocoding addresses.
</p>

For more advanced users, see the [Nominatim
documentation](https://nominatim.org/release-docs/latest/api/Search/)
for the available parameters.

## Citation

<p>

Hernangómez D (2026). <em>nominatimlite: Interface to the Nominatim
API</em>.
<a href="https://doi.org/10.32614/CRAN.package.nominatimlite">doi:10.32614/CRAN.package.nominatimlite</a>.
<a href="https://dieghernan.github.io/nominatimlite/">https://dieghernan.github.io/nominatimlite/</a>.
</p>

A BibTeX entry for LaTeX users is

    @Manual{R-nominatimlite,
      title = {{nominatimlite}: Interface to the {Nominatim} {API}},
      doi = {10.32614/CRAN.package.nominatimlite},
      author = {Diego Hernangómez},
      year = {2026},
      version = {0.6.0},
      url = {https://dieghernan.github.io/nominatimlite/},
      abstract = {Lightweight interface to the OpenStreetMap Nominatim API <https://nominatim.org/release-docs/latest/>. Extract coordinates from addresses, retrieve addresses from coordinates, look up amenities and addresses, and return results as tibble or sf objects.},
    }

## References

<div id="refs" class="references csl-bib-body hanging-indent">

<div id="ref-R-tidygeocoder" class="csl-entry">

Cambon, Jesse, Diego Hernangómez, Christopher Belanger, and Daniel
Possenriede. 2021. “<span class="nocase">tidygeocoder</span>: An R
Package for Geocoding.” *Journal of Open Source Software* 6 (65): 3544.
<https://doi.org/10.21105/joss.03544>.

</div>

<div id="ref-R-arcgeocoder" class="csl-entry">

Hernangómez, Diego. 2024. *<span class="nocase">arcgeocoder</span>:
Geocoding with the ArcGIS REST API Service*. Version 0.1.0.
<https://doi.org/10.5281/zenodo.10495365>.

</div>

<div id="ref-R-osmdata" class="csl-entry">

Padgham, Mark, Robin Lovelace, Maëlle Salmon, and Bob Rudis. 2017.
“<span class="nocase">osmdata</span>.” *Journal of Open Source Software*
2 (14): 305. <https://doi.org/10.21105/joss.00305>.

</div>

</div>

<!-- README.md is generated from README.qmd. Please edit that file -->

# nominatimlite <a href="https://dieghernan.github.io/nominatimlite/"><img src="man/figures/logo.png" alt="nominatimlite website" align="right" height="139"/></a>

<!-- badges: start -->

[![CRAN
status](https://www.r-pkg.org/badges/version/nominatimlite)](https://CRAN.R-project.org/package=nominatimlite)
[![CRAN
results](https://badges.cranchecks.info/worst/nominatimlite.svg)](https://cran.r-project.org/web/checks/check_results_nominatimlite.html)
[![Downloads](https://cranlogs.r-pkg.org/badges/nominatimlite)](https://CRAN.R-project.org/package=nominatimlite)
[![Nominatim-version](https://img.shields.io/badge/Nominatim-%E2%89%A5_4.4.0-blue)](https://nominatim.org/release-docs/develop/api/Overview/)
[![R-CMD-check](https://github.com/dieghernan/nominatimlite/actions/workflows/check-full.yaml/badge.svg)](https://github.com/dieghernan/nominatimlite/actions/workflows/check-full.yaml)
[![codecov](https://codecov.io/gh/dieghernan/nominatimlite/branch/main/graph/badge.svg?token=jSZ4RIsj91)](https://app.codecov.io/gh/dieghernan/nominatimlite)
[![CodeFactor](https://www.codefactor.io/repository/github/dieghernan/nominatimlite/badge)](https://www.codefactor.io/repository/github/dieghernan/nominatimlite)
[![r-universe](https://dieghernan.r-universe.dev/badges/nominatimlite)](https://dieghernan.r-universe.dev/nominatimlite)
[![Project Status: Active – The project has reached a stable, usable
state and is being actively
developed.](https://www.repostatus.org/badges/latest/active.svg)](https://www.repostatus.org/#active)
[![DOI](https://img.shields.io/badge/DOI-10.32614/CRAN.package.nominatimlite-blue)](https://doi.org/10.32614/CRAN.package.nominatimlite)
[![status](https://tinyverse.netlify.app/status/nominatimlite)](https://CRAN.R-project.org/package=nominatimlite)

<!-- badges: end -->

**nominatimlite** provides a lightweight interface to the [**Nominatim
API**](https://nominatim.org/release-docs/latest/). It supports
free-form and structured address searches, reverse geocoding, amenity
lookup and address lookup by **OpenStreetMap** (OSM) object identifier.
Results are returned as tibbles or **sf** objects.

The full site with examples and vignettes is available at
<https://dieghernan.github.io/nominatimlite/>.

## What is Nominatim?

**Nominatim** searches
[**OpenStreetMap**](https://www.openstreetmap.org/) data by name and
address
([geocoding](https://wiki.openstreetmap.org/wiki/Geocoding "Geocoding"))
and reverse geocodes geographic coordinates.

## Why nominatimlite?

**nominatimlite** accesses the **Nominatim API** without requiring the
**curl** package. This makes the package useful in environments where
**curl** is not available. API requests use base **R** functions
instead.

## Recommended packages

Related packages provide broader interfaces to geocoding services and
**OpenStreetMap** data:

- [**tidygeocoder**](https://jessecambon.github.io/tidygeocoder/)
  ([Cambon et al. 2021](#ref-R-tidygeocoder)) provides an interface to
  geocoding services such as **Nominatim**, **Google**, **TomTom** and
  **Mapbox**.
- [**osmdata**](https://docs.ropensci.org/osmdata/) ([Padgham et al.
  2017](#ref-R-osmdata)) downloads spatial data from **OpenStreetMap**
  with the [**Overpass**
  API](https://wiki.openstreetmap.org/wiki/Overpass_API).
- [**arcgeocoder**](https://dieghernan.github.io/arcgeocoder/)
  ([Hernangómez 2024](#ref-R-arcgeocoder)) provides a lightweight
  interface for geocoding with the **ArcGIS REST API** service.

## Installation

<div class="pkgdown-release">

Install **nominatimlite** from
[**CRAN**](https://CRAN.R-project.org/package=nominatimlite):

``` r
install.packages("nominatimlite")
```

</div>

<div class="pkgdown-devel">

See the documentation for the development version at
<https://dieghernan.github.io/nominatimlite/dev/>.

You can install the development version of **nominatimlite** with:

``` r
pak::pak("dieghernan/nominatimlite")
```

Alternatively, you can install **nominatimlite** using the
[**r-universe**](https://dieghernan.r-universe.dev/nominatimlite):

``` r
# Install nominatimlite in R:
install.packages(
  "nominatimlite",
  repos = c(
    "https://dieghernan.r-universe.dev",
    "https://cloud.r-project.org"
  )
)
```

</div>

## Usage

### sf output

Use functions with the `_sf` suffix to return results as **sf** objects:

``` r
library(nominatimlite)

# Search for Pizza Hut locations in California.

CA <- geo_lite_sf("California", points_only = FALSE)

pizzahut <- geo_lite_sf(
  "Pizza Hut, California",
  limit = 50,
  custom_query = list(countrycodes = "us")
)

library(ggplot2)

ggplot(CA) +
  geom_sf() +
  geom_sf(data = pizzahut, col = "red")
```

<img src="man/figures/README-pizzahut-1.png" style="width:100.0%"
alt="Pizza Hut locations in California returned by nominatimlite." />

Set `points_only = FALSE` to return polygon and line geometries when
they are available from the **Nominatim API**:

``` r
# A building, returned as a polygon.
sol_poly <- geo_lite_sf("Statue of Liberty, NY, USA", points_only = FALSE)

ggplot(sol_poly) +
  geom_sf()
```

<img src="man/figures/README-statue_liberty-1.png" style="width:100.0%"
alt="Statue of Liberty geometry returned by nominatimlite." />

``` r
# Default, returned as a point.
dayton <- geo_lite_sf("Dayton, OH")
# A US state, returned as a polygon.
ohio_state <- geo_lite_sf("Ohio, USA", points_only = FALSE)
# A river, returned as a line.
ohio_river <- geo_lite_sf("Ohio river", points_only = FALSE)

ggplot() +
  geom_sf(data = ohio_state) +
  geom_sf(data = dayton, color = "red", pch = 4) +
  geom_sf(data = ohio_river, color = "blue")
```

<img src="man/figures/README-line-object-1.png" style="width:100.0%"
alt="Different features named Ohio returned by nominatimlite." />

### Address search and reverse geocoding

*The examples in this section are adapted from the **tidygeocoder**
package.*

Use `geo_lite()` to search for addresses with free-form queries:

``` r
# Create a data frame with addresses.
some_addresses <- dplyr::tribble(
  ~name, ~addr,
  "White House", "1600 Pennsylvania Ave NW, Washington, DC",
  "Transamerica Pyramid", "600 Montgomery St, San Francisco, CA 94111",
  "Willis Tower", "233 S Wacker Dr, Chicago, IL 60606"
)

# Geocode the addresses.
lat_longs <- geo_lite(
  some_addresses$addr,
  lat = "latitude",
  long = "longitude",
  progressbar = FALSE
)
```

By default, `geo_lite()` returns the query, latitude, longitude and
address columns. Set `full_results = TRUE` to return all available
fields from the **Nominatim API**.

| query | latitude | longitude | address |
|:---|---:|---:|:---|
| 1600 Pennsylvania Ave NW, Washington, DC | 38.89764 | -77.03655 | White House, 1600, Pennsylvania Avenue Northwest, Ward 2, Washington, District of Columbia, 20500, United States |
| 600 Montgomery St, San Francisco, CA 94111 | 37.79519 | -122.40279 | Transamerica Pyramid, 600, Montgomery Street, Financial District, South of Market, San Francisco, California, 94111, United States |
| 233 S Wacker Dr, Chicago, IL 60606 | 41.87874 | -87.63596 | Willis Tower, 233, South Wacker Drive, Financial District, Loop, Chicago, South Chicago Township, Cook County, Illinois, 60606, United States |

<p class="caption">

Table 1: Geocoded addresses.
</p>

Use `reverse_geo_lite()` to reverse geocode latitude and longitude
coordinates. The `lat` and `long` arguments use the results from the
address search above. The `address` argument specifies the name of the
output column that contains each single-line address.

``` r
reverse <- reverse_geo_lite(
  lat = lat_longs$latitude,
  long = lat_longs$longitude,
  address = "address_found",
  progressbar = FALSE
)
```

| address_found | lat | lon |
|:---|---:|---:|
| White House, 1600, Pennsylvania Avenue Northwest, Ward 2, Washington, District of Columbia, 20500, United States | 38.89764 | -77.03655 |
| Sky Bar, 600, Montgomery Street, Financial District, South of Market, San Francisco, California, 94111, United States | 37.79519 | -122.40254 |
| 233, South Wacker Drive, Financial District, Loop, Chicago, South Chicago Township, Cook County, Illinois, 60606, United States | 41.87874 | -87.63589 |

<p class="caption">

Table 2: Reverse-geocoded addresses.
</p>

See the [**Nominatim search API**
documentation](https://nominatim.org/release-docs/latest/api/Search/)
for additional parameters to pass to `custom_query`.

## Citation

<p>

Hernangómez D (2026). <em>nominatimlite: Interface to the Nominatim
API</em>.
<a href="https://doi.org/10.32614/CRAN.package.nominatimlite">doi:10.32614/CRAN.package.nominatimlite</a>.
<a href="https://dieghernan.github.io/nominatimlite/">https://dieghernan.github.io/nominatimlite/</a>.
</p>

A BibTeX entry for LaTeX users is shown below.

    @Manual{R-nominatimlite,
      title = {{nominatimlite}: Interface to the {Nominatim} {API}},
      doi = {10.32614/CRAN.package.nominatimlite},
      author = {Diego Hernangómez},
      year = {2026},
      version = {0.7.0},
      url = {https://dieghernan.github.io/nominatimlite/},
      abstract = {Provides a lightweight interface to the Nominatim API <https://nominatim.org/release-docs/latest/>. It supports free-form and structured address searches, searches for addresses from coordinates, amenity lookup and address lookup by OpenStreetMap object identifier. It returns results as tibble data frames or sf objects.},
    }

## References

<div id="refs" class="references csl-bib-body hanging-indent">

<div id="ref-R-tidygeocoder" class="csl-entry">

Cambon, Jesse, Diego Hernangómez, Christopher Belanger, and Daniel
Possenriede. 2021. “<span class="nocase">tidygeocoder</span>: An R
Package for Geocoding.” *Journal of Open Source Software* 6 (65): 3544.
<https://doi.org/10.21105/joss.03544>.

</div>

<div id="ref-R-arcgeocoder" class="csl-entry">

Hernangómez, Diego. 2024. *<span class="nocase">arcgeocoder</span>:
Geocoding with the ArcGIS REST API Service*. Version 0.1.0.
<https://doi.org/10.5281/zenodo.10495365>.

</div>

<div id="ref-R-osmdata" class="csl-entry">

Padgham, Mark, Robin Lovelace, Maëlle Salmon, and Bob Rudis. 2017.
“<span class="nocase">osmdata</span>.” *Journal of Open Source Software*
2 (14): 305. <https://doi.org/10.21105/joss.00305>.

</div>

</div>

#' Search for addresses with free-form queries
#'
#' @description
#' Searches for addresses supplied as a character vector and returns matching
#' results as a [tibble][dplyr::tibble]. Use [geo_lite_sf()] to return an
#' [`sf`][sf::st_sf] object instead.
#'
#' This function performs the **free-form address search** described in the
#' [API endpoint](https://nominatim.org/release-docs/latest/api/Search/).
#'
#' @details
#' See <https://nominatim.org/release-docs/latest/api/Search/> for additional
#' parameters to pass to `custom_query`.
#'
#' @param address A character vector of single-line addresses, for example
#'   `"1600 Pennsylvania Ave NW, Washington"` or
#'   `c("Madrid", "Barcelona")`.
#' @param lat A character string specifying the name of the latitude column in
#'   the output. Defaults to `"lat"`.
#' @param long A character string specifying the name of the longitude column
#'   in the output. Defaults to `"lon"`.
#' @param limit A positive integer specifying the maximum number of results to
#'   return per query. Nominatim returns at most 50 results per query.
#' @param full_results A logical value indicating whether to return all
#'   available fields from the Nominatim API. If `FALSE`, only query metadata,
#'   location data and requested address columns are returned.
#' @param return_addresses A logical value indicating whether to include
#'   single-line addresses in the results.
#' @param verbose A logical value indicating whether to display detailed
#'   messages in the console.
#' @param nominatim_server A character string specifying the base URL of the
#'   Nominatim server. Defaults to
#'   `"https://nominatim.openstreetmap.org/"`.
#' @param progressbar A logical value indicating whether to display a progress
#'   bar when processing multiple queries.
#' @param custom_query A named list of additional API parameters, for example
#'   `list(countrycodes = "US")`. See **Details**.
#'
#' @returns
#' A [`tibble`][dplyr::tibble] with the results that match the query.
#'
#' @seealso
#' [reverse_geo_lite()] for reverse geocoding coordinates,
#' [geo_address_lookup()] for looking up known OpenStreetMap identifiers and
#' [geo_amenity()] for searching amenities within a bounding box.
#'
#' @family geocoding
#'
#' @export
#' @encoding UTF-8
#'
#' @examplesIf nominatim_check_access()
#' \donttest{
#' geo_lite("Madrid, Spain")
#'
#' # Search for multiple addresses.
#' geo_lite(c("Madrid", "Barcelona"))
#'
#' # Restrict the search to the United States and return all fields.
#' geo_lite(c("Madrid", "Barcelona"),
#'   custom_query = list(countrycodes = "US"),
#'   full_results = TRUE
#' )
#' }
geo_lite <- function(
  address,
  lat = "lat",
  long = "lon",
  limit = 1,
  full_results = FALSE,
  return_addresses = TRUE,
  verbose = FALSE,
  nominatim_server = "https://nominatim.openstreetmap.org/",
  progressbar = TRUE,
  custom_query = list()
) {
  limit <- cap_limit(limit)

  # Deduplicate queries before calling the API.
  init_key <- dplyr::tibble(query = address)
  key <- unique(address)

  ntot <- length(key)

  all_res <- progress_lapply(ntot, progressbar, function(x) {
    ad <- key[x]
    geo_lite_single(
      address = ad,
      lat,
      long,
      limit,
      full_results,
      return_addresses,
      verbose,
      nominatim_server = nominatim_server,
      custom_query
    )
  })

  all_res <- dplyr::bind_rows(all_res)
  all_res <- dplyr::left_join(init_key, all_res, by = "query")

  all_res
}

#' @inheritParams geo_lite
#' @noRd

geo_lite_single <- function(
  address,
  lat = "lat",
  long = "lon",
  limit = 1,
  full_results = TRUE,
  return_addresses = TRUE,
  verbose = FALSE,
  nominatim_server = "https://nominatim.openstreetmap.org/",
  custom_query = list()
) {
  url <- build_search_url(
    nominatim_server = nominatim_server,
    format = "jsonv2",
    limit = limit,
    full_results = full_results,
    custom_query = custom_query,
    query = address
  )

  # Download the API response.
  json <- api_call(url, ".json", isFALSE(verbose))

  # Keep the original query value.
  tbl_query <- dplyr::tibble(query = address)

  if (isFALSE(json)) {
    message_api_unavailable(url)
    out <- empty_tbl(tbl_query, lat, long)
    return(invisible(out))
  }

  result <- dplyr::as_tibble(jsonlite::fromJSON(json, flatten = TRUE))

  result <- rename_coordinate_cols(result, lat, long)

  # Handle empty queries.
  if (nrow(result) == 0) {
    message_no_results(address)
    out <- empty_tbl(tbl_query, lat, long)
    return(invisible(out))
  }

  result <- convert_coordinate_cols(result, lat, long)

  # Add the query to the API results.
  result_clean <- result
  result_clean$query <- address

  # Keep selected columns.
  result_out <- keep_names(
    result_clean,
    return_addresses,
    full_results,
    colstokeep = c("query", lat, long)
  )

  # Restore tibble classes.
  result_out <- dplyr::as_tibble(result_out)

  result_out
}

test_that("geo_lite() completes a live address search", {
  skip_if_nominatim_unavailable()

  out <- geo_lite("Madrid", progressbar = FALSE)

  expect_s3_class(out, "tbl")
  expect_gt(nrow(out), 0)
  expect_contains(names(out), c("query", "lat", "lon", "address"))
  expect_all_true(is.finite(out$lat))
  expect_all_true(is.finite(out$lon))
})

test_that("geo_lite_sf() completes a live spatial address search", {
  skip_if_nominatim_unavailable()

  out <- geo_lite_sf("Madrid", progressbar = FALSE)

  expect_s3_class(out, "sf")
  expect_gt(nrow(out), 0)
  expect_all_false(sf::st_is_empty(out))
  expect_equal(sf::st_crs(out)$epsg, 4326)
})

test_that("geo_lite_struct() completes a live structured search", {
  skip_if_nominatim_unavailable()

  out <- geo_lite_struct(city = "Madrid", country = "Spain")

  expect_s3_class(out, "tbl")
  expect_gt(nrow(out), 0)
  expect_contains(names(out), c("q_city", "q_country", "lat", "lon"))
  expect_all_true(is.finite(out$lat))
  expect_all_true(is.finite(out$lon))
})

test_that("geo_lite_struct_sf() completes a live spatial structured search", {
  skip_if_nominatim_unavailable()

  out <- geo_lite_struct_sf(city = "Madrid", country = "Spain")

  expect_s3_class(out, "sf")
  expect_gt(nrow(out), 0)
  expect_all_false(sf::st_is_empty(out))
  expect_equal(sf::st_crs(out)$epsg, 4326)
})

test_that("geo_address_lookup() completes a live OSM object lookup", {
  skip_if_nominatim_unavailable()

  out <- geo_address_lookup(34633854, "W")

  expect_s3_class(out, "tbl")
  expect_gt(nrow(out), 0)
  expect_contains(names(out), c("query", "lat", "lon", "address"))
  expect_all_true(is.finite(out$lat))
  expect_all_true(is.finite(out$lon))
})

test_that("geo_address_lookup_sf() completes a live spatial OSM lookup", {
  skip_if_nominatim_unavailable()

  out <- geo_address_lookup_sf(34633854, "W")

  expect_s3_class(out, "sf")
  expect_gt(nrow(out), 0)
  expect_all_false(sf::st_is_empty(out))
  expect_equal(sf::st_crs(out)$epsg, 4326)
})

test_that("geo_amenity() completes a live amenity lookup", {
  skip_if_nominatim_unavailable()

  out <- geo_amenity(
    c(-3.71, 40.41, -3.69, 40.43),
    "restaurant",
    limit = 1,
    progressbar = FALSE
  )

  expect_s3_class(out, "tbl")
  expect_gt(nrow(out), 0)
  expect_contains(names(out), c("query", "lat", "lon", "address"))
  expect_all_true(is.finite(out$lat))
  expect_all_true(is.finite(out$lon))
})

test_that("geo_amenity_sf() completes a live spatial amenity lookup", {
  skip_if_nominatim_unavailable()

  out <- geo_amenity_sf(
    c(-3.71, 40.41, -3.69, 40.43),
    "restaurant",
    limit = 1,
    progressbar = FALSE
  )

  expect_s3_class(out, "sf")
  expect_gt(nrow(out), 0)
  expect_all_false(sf::st_is_empty(out))
  expect_equal(sf::st_crs(out)$epsg, 4326)
})

test_that("reverse_geo_lite() completes a live reverse lookup", {
  skip_if_nominatim_unavailable()

  out <- reverse_geo_lite(40.4168, -3.7038, progressbar = FALSE)

  expect_s3_class(out, "tbl")
  expect_gt(nrow(out), 0)
  expect_contains(names(out), c("address", "lat", "lon"))
  expect_all_true(!is.na(out$address))
})

test_that("reverse_geo_lite_sf() completes a live spatial reverse lookup", {
  skip_if_nominatim_unavailable()

  out <- reverse_geo_lite_sf(40.4168, -3.7038, progressbar = FALSE)

  expect_s3_class(out, "sf")
  expect_gt(nrow(out), 0)
  expect_all_false(sf::st_is_empty(out))
  expect_equal(sf::st_crs(out)$epsg, 4326)
})

test_that("nominatim_check_access() confirms the live status endpoint", {
  skip_if_nominatim_unavailable()

  expect_true(nominatim_check_access())
})

#' Look up OpenStreetMap amenities and return \CRANpkg{sf} objects
#'
#' @description
#' Looks up OpenStreetMap [amenities][osm_amenities] within a bounding box of
#' the form `(xmin, ymin, xmax, ymax)`. Results are returned as an
#' [`sf`][sf::st_sf] object. Use [geo_amenity()] to return a
#' [tibble][dplyr::tibble] instead.
#'
#' @inherit geo_amenity details
#'
#' @inheritParams geo_amenity -full_results
#' @inheritParams geo_lite_sf full_results points_only
#'
#' @inherit geo_lite_sf return
#'
#' @inheritSection geo_lite_sf About geometry types
#'
#' @family amenity
#' @family spatial
#'
#' @export
#' @encoding UTF-8
#'
#' @examplesIf nominatim_check_access()
#' \donttest{
#' # Retrieve the Usera district in Madrid.
#'
#' library(ggplot2)
#' mad <- geo_lite_sf("Usera, Madrid, Spain", points_only = FALSE)
#'
#' # Search for restaurants, pubs and schools.
#'
#' rest_pub <- geo_amenity_sf(mad, c("restaurant", "pub", "school"),
#'   limit = 50
#' )
#'
#' if (!all(sf::st_is_empty(rest_pub))) {
#'   ggplot(mad) +
#'     geom_sf() +
#'     geom_sf(data = rest_pub, aes(color = query, shape = query))
#' }
#' }
geo_amenity_sf <- function(
  bbox,
  amenity,
  limit = 1,
  full_results = FALSE,
  return_addresses = TRUE,
  verbose = FALSE,
  nominatim_server = "https://nominatim.openstreetmap.org/",
  progressbar = TRUE,
  custom_query = list(),
  strict = FALSE,
  points_only = TRUE
) {
  limit <- cap_limit(limit)

  bbox <- normalize_bbox(bbox)

  # Add the viewbox restriction to the custom query.
  custom_query <- as.list(custom_query)
  custom_query$viewbox <- bbox
  custom_query$bounded <- TRUE

  # Deduplicate queries.
  key <- unique(amenity)

  ntot <- length(key)

  all_res <- progress_lapply(ntot, progressbar, function(x) {
    ad <- key[x]

    geo_lite_struct_sf(
      amenity = ad,
      limit = limit,
      full_results = full_results,
      return_addresses = return_addresses,
      verbose = verbose,
      nominatim_server = nominatim_server,
      custom_query = custom_query,
      points_only = points_only
    )
  })

  all_res <- dplyr::bind_rows(all_res)

  # Clean query columns and names.
  nm <- names(all_res)
  nm[nm == "q_amenity"] <- "query"
  names(all_res) <- nm
  all_res <- all_res[, !grepl("^q_", nm)]
  all_res <- sf_to_tbl(all_res)

  if (strict) {
    bbox_sf <- bbox_to_poly(bbox)
    int <- as.vector(sf::st_intersects(all_res, bbox_sf, sparse = FALSE))
    all_res <- all_res[int, ]
  }

  all_res
}

#' Convert a bounding box to an [`sfc`][sf::st_sfc] `POLYGON` object
#'
#' @description
#' Converts bounding box coordinates to an [`sfc`][sf::st_sfc] object with
#' `POLYGON` geometry.
#'
#' @details
#' Bounding boxes can be located using online tools such as
#' <https://boundingbox.klokantech.com/>.
#'
#' @param bbox A numeric vector of four bounding box coordinates in the form
#'   `c(xmin, ymin, xmax, ymax)`.
#' @param xmin,ymin,xmax,ymax A numeric value specifying an individual bounding
#'   box coordinate. Use these arguments as an alternative to `bbox`.
#' @inheritParams sf::st_sf crs
#'
#' @returns
#' An [`sfc`][sf::st_sfc] object with `POLYGON` geometry and the coordinate
#' reference system specified by `crs`.
#'
#' @seealso
#' [sf::st_as_sfc()] and [sf::st_sfc()].
#'
#' @family spatial
#'
#' @export
#' @encoding UTF-8
#'
#' @examplesIf nominatim_check_access() && requireNamespace("ggplot2")
#' # Convert the bounding box for Germany.
#' bbox_GER <- c(5.86631529, 47.27011137, 15.04193189, 55.09916098)
#'
#' bbox_GER_sf <- bbox_to_poly(bbox_GER)
#'
#' library(ggplot2)
#'
#' ggplot(bbox_GER_sf) +
#'   geom_sf()
#' \donttest{
#' # Extract the bounding box of an `sf` object.
#' sfobj <- geo_lite_sf("seychelles", points_only = FALSE)
#'
#' sfobj
#'
#' # Require at least one non-empty object.
#' if (!all(sf::st_is_empty(sfobj))) {
#'   bbox <- sf::st_bbox(sfobj)
#'
#'   bbox
#'
#'   bbox_sfobj <- bbox_to_poly(bbox)
#'
#'   ggplot(bbox_sfobj) +
#'     geom_sf(fill = "lightblue", alpha = 0.5) +
#'     geom_sf(data = sfobj, fill = "wheat")
#' }
#' }
bbox_to_poly <- function(
  bbox = NA,
  xmin = NA,
  ymin = NA,
  xmax = NA,
  ymax = NA,
  crs = 4326
) {
  if (!anyNA(bbox) && length(bbox) != 4) {
    stop(
      "`bbox` must contain exactly four elements, but the provided value has ",
      length(bbox),
      "."
    )
  }

  # Use explicit x and y values when `bbox` is missing.
  if (anyNA(bbox)) {
    bbox <- as.double(c(xmin, ymin, xmax, ymax))

    if (anyNA(bbox)) {
      stop(
        "Provide `bbox` or non-missing values for `xmin`, `ymin`, `xmax` ",
        "and `ymax`."
      )
    }
  }

  bbox_double <- as.double(bbox)
  names(bbox_double) <- c("xmin", "ymin", "xmax", "ymax")
  class(bbox_double) <- "bbox"

  bbox_sf <- sf::st_as_sfc(bbox_double)
  sf::st_crs(bbox_sf) <- sf::st_crs(crs)

  bbox_sf
}

test_that("bbox_to_poly() rejects incomplete bounding boxes", {
  expect_snapshot(error = TRUE, bbox_to_poly())
  expect_snapshot(error = TRUE, bbox_to_poly(1))
  expect_snapshot(error = TRUE, bbox_to_poly(xmin = 1))
})


test_that("bbox_to_poly() accepts vector and component coordinates", {
  expect_silent(bbox_to_poly(c(1, 2, 3, 4)))
  expect_silent(bbox_to_poly(xmin = 1, xmax = 2, ymin = 3, ymax = 4))
  expect_equal(
    sf::st_bbox(bbox_to_poly(xmin = 1, xmax = 2, ymin = 3, ymax = 4)) |>
      as.numeric() |>
      setNames(c("xmin", "ymin", "xmax", "ymax")),
    c(xmin = 1, ymin = 3, xmax = 2, ymax = 4)
  )

  crsa <- sf::st_crs(bbox_to_poly(c(1, 2, 3, 4)))
  crsb <- sf::st_crs(bbox_to_poly(c(1, 2, 3, 4), crs = 3857))

  expect_false(identical(crsa, crsb))
})

test_that("bbox_to_poly() returns polygons with the requested CRS", {
  obj <- bbox_to_poly(c(1, 2, 3, 4))
  expect_s3_class(obj, "sfc")
  expect_equal(as.character(sf::st_geometry_type(obj)), "POLYGON")
  expect_identical(sf::st_crs(obj), sf::st_crs(4326))

  obj <- bbox_to_poly(c(1, 2, 3, 4), crs = 3035)
  expect_s3_class(obj, "sfc")
  expect_equal(as.character(sf::st_geometry_type(obj)), "POLYGON")
  expect_identical(sf::st_crs(obj), sf::st_crs(3035))
})

#' Check access to the Nominatim API
#'
#' @description
#' Checks whether \R can access a Nominatim API server.
#'
#' @inheritParams geo_lite nominatim_server
#'
#' @returns
#' A single logical value: `TRUE` if the API is available and `FALSE` otherwise.
#'
#' @seealso
#' [geo_lite()] for submitting search requests and the
#' [Nominatim status
#' endpoint](https://nominatim.org/release-docs/latest/api/Status/)
#' for server status details.
#'
#' @keywords internal
#' @export
#' @encoding UTF-8
#'
#' @examples
#' \donttest{
#' nominatim_check_access()
#' }
nominatim_check_access <- function(
  nominatim_server = "https://nominatim.openstreetmap.org/"
) {
  if (on_cran()) {
    return(FALSE)
  }

  # Build the API address.
  url <- prepare_api_url(nominatim_server, "status?format=json")

  api_res <- api_call(url, ".json", TRUE)
  if (isFALSE(api_res)) {
    return(FALSE)
  }

  result <- dplyr::as_tibble(jsonlite::fromJSON(api_res, flatten = TRUE))

  any(result$status == 0 || result$message == "OK")
}

#' Query the Nominatim API
#'
#' @description
#' Wraps [utils::download.file()] and retries the request after an error or
#' warning. Requests are limited to approximately one query per second.
#'
#' See [Nominatim Usage
#' Policy](https://operations.osmfoundation.org/policies/nominatim/).
#'
#' @param ext A character string specifying the file extension for the cached
#'   response. Must be `".json"` or `".geojson"`.
#' @inheritParams utils::download.file url quiet
#'
#' @returns
#' A cached file path or `FALSE` when the query fails.
#'
#' @noRd
#'
api_call <- function(url, ext = c(".json", ".geojson"), quiet) {
  ext <- match.arg(ext)

  # Hash the destination file.
  destfile <- cached_filename(url, ext)
  # Return cached files.
  if (file.exists(destfile)) {
    return(destfile)
  }

  dwn_res <- download_api_file(url, destfile, quiet)

  # Always sleep to keep one call per second with an extra buffer.
  pause_api_call()

  if (!inherits(dwn_res, "try-error")) {
    return(destfile)
  }
  if (isFALSE(quiet)) {
    message("Retrying the Nominatim API query.")
  }
  pause_api_call()

  dwn_res <- download_api_file(url, destfile, quiet)

  # Return the file after a successful request.
  if (!inherits(dwn_res, "try-error")) {
    return(destfile)
  }

  unlink(destfile, force = TRUE)

  !inherits(dwn_res, "try-error")
}

download_api_file <- function(url, destfile, quiet) {
  # nocov start
  suppressWarnings(try(
    download.file(url, destfile = destfile, quiet = quiet, mode = "wb"),
    silent = TRUE
  ))
  # nocov end
}

pause_api_call <- function() {
  # nocov start
  Sys.sleep(1.2)
  # nocov end
}

#' Create a hashed filename for caching requests
#'
#' Creates a deterministic path in the session temporary directory from the
#' request URL.
#'
#' @param url A character string specifying the URL to cache.
#' @param ext A character string specifying the file extension to append to the
#'   cached file.
#'
#' @returns A path to the cached response file.
#'
#' @noRd
cached_filename <- function(url, ext = ".json") {
  tmpf <- tempfile()
  writeLines(url, tmpf)

  hash <- unname(tools::md5sum(tmpf))
  unlink(tmpf, force = TRUE)

  # Create the corresponding temporary directory and add the extension.
  tmpnomin <- file.path(tempdir(), "nominatim_cache")
  if (!dir.exists(tmpnomin)) {
    dir.create(tmpnomin, showWarnings = FALSE, recursive = TRUE)
  }

  # Return the final filename.
  fname <- file.path(tmpnomin, paste0(hash, ext))
  fname
}

#' Check whether the current session is running on CRAN
#'
#' Checks the `NOT_CRAN` environment variable and whether the session is
#' interactive.
#'
#' @returns A single logical value.
#'
#' @noRd
on_cran <- function() {
  env <- Sys.getenv("NOT_CRAN")
  if (identical(env, "")) {
    !interactive()
  } else {
    !isTRUE(as.logical(env))
  }
}

test_that("api_call() returns an existing cached response", {
  tmp <- withr::local_tempfile(fileext = ".json")
  file.create(tmp)
  local_mocked_bindings(
    cached_filename = function(url, ext) tmp,
    download_api_file = function(...) {
      testthat::fail("A cached response must not trigger a download.")
    },
    pause_api_call = function() {
      testthat::fail("A cached response must not trigger throttling.")
    }
  )

  expect_identical(
    api_call("https://example.com", ext = ".json", quiet = TRUE),
    tmp
  )
})

test_that("api_call() returns FALSE and removes files after failed retries", {
  tmp <- withr::local_tempfile(fileext = ".json")
  state <- new.env(parent = emptyenv())
  state$calls <- 0L
  state$pauses <- 0L

  local_mocked_bindings(
    cached_filename = function(url, ext) tmp,
    download_api_file = function(url, destfile, quiet) {
      state$calls <- state$calls + 1L
      file.create(destfile)
      structure("boom", class = "try-error")
    },
    pause_api_call = function() {
      state$pauses <- state$pauses + 1L
    }
  )

  expect_false(api_call("https://example.com", ext = ".json", quiet = TRUE))
  expect_equal(state$calls, 2L)
  expect_equal(state$pauses, 2L)
  expect_false(file.exists(tmp))
})

test_that("api_call() returns after the first successful request", {
  tmp <- withr::local_tempfile(fileext = ".json")
  state <- new.env(parent = emptyenv())
  state$calls <- 0L
  state$pauses <- 0L

  local_mocked_bindings(
    cached_filename = function(url, ext) tmp,
    download_api_file = function(url, destfile, quiet) {
      state$calls <- state$calls + 1L
      file.create(destfile)
      destfile
    },
    pause_api_call = function() {
      state$pauses <- state$pauses + 1L
    }
  )

  expect_identical(
    api_call("https://example.com", ext = ".json", quiet = TRUE),
    tmp
  )
  expect_equal(state$calls, 1L)
  expect_equal(state$pauses, 1L)
  expect_true(file.exists(tmp))
})

test_that("api_call() rejects unsupported cache formats before downloading", {
  expect_snapshot(
    error = TRUE,
    api_call("https://example.com", ext = ".txt", quiet = TRUE)
  )
})

test_that("cached_filename() creates stable paths for each format", {
  one <- cached_filename("https://example.com/search?q=Madrid", ".json")
  two <- cached_filename("https://example.com/search?q=Madrid", ".json")
  geojson <- cached_filename("https://example.com/search?q=Madrid", ".geojson")

  expect_identical(one, two)
  expect_match(one, "nominatim_cache", fixed = TRUE)
  expect_match(one, "\\.json$")
  expect_match(geojson, "\\.geojson$")
  expect_true(dir.exists(dirname(one)))
})

test_that("nominatim_check_access() interprets status responses", {
  local_mocked_bindings(
    on_cran = function() FALSE,
    api_call = testthat::mock_output_sequence(
      test_fixture("status-ok.json"),
      test_fixture("status-message-ok.json"),
      test_fixture("status-ko.json"),
      FALSE
    )
  )

  expect_true(nominatim_check_access())
  expect_true(nominatim_check_access())
  expect_false(nominatim_check_access())
  expect_false(nominatim_check_access())
})

test_that("on_cran() controls access checks from NOT_CRAN", {
  # Imagine we are in CRAN
  withr::local_envvar("NOT_CRAN" = "false")
  expect_true(on_cran())
  expect_false(nominatim_check_access())

  withr::local_envvar("NOT_CRAN" = "")
  expect_identical(!interactive(), on_cran())
})

test_that("api_call() reports and completes a successful retry", {
  tmp <- withr::local_tempfile(fileext = ".json")
  state <- new.env(parent = emptyenv())
  state$calls <- 0L
  state$pauses <- 0L

  local_mocked_bindings(
    cached_filename = function(url, ext) tmp,
    download_api_file = function(url, destfile, quiet) {
      state$calls <- state$calls + 1L

      if (state$calls == 1L) {
        return(structure("boom", class = "try-error"))
      }

      file.create(destfile)
      destfile
    },
    pause_api_call = function() {
      state$pauses <- state$pauses + 1L
    }
  )

  expect_message(
    res <- api_call("https://example.com", ext = ".json", quiet = FALSE),
    "Retrying the Nominatim API query."
  )

  expect_identical(res, tmp)
  expect_equal(state$calls, 2L)
  expect_equal(state$pauses, 2L)
})

test_that("api_call() retries quietly when verbose output is disabled", {
  tmp <- withr::local_tempfile(fileext = ".json")
  state <- new.env(parent = emptyenv())
  state$calls <- 0L

  local_mocked_bindings(
    cached_filename = function(url, ext) tmp,
    download_api_file = function(url, destfile, quiet) {
      state$calls <- state$calls + 1L

      if (state$calls == 1L) {
        return(structure("boom", class = "try-error"))
      }

      file.create(destfile)
      destfile
    },
    pause_api_call = function() NULL
  )

  expect_silent(
    res <- api_call("https://example.com", ext = ".json", quiet = TRUE)
  )
  expect_identical(res, tmp)
  expect_equal(state$calls, 2L)
})

# General ----
message_api_unavailable <- function(url) {
  message("Could not reach the API endpoint: ", url, ".")
}

message_no_results <- function(query = NULL) {
  if (is.null(query)) {
    message("No results were found for the query.")
    return(invisible(NULL))
  }

  message("No results were found for the query: ", query, ".")
}

cap_limit <- function(limit) {
  if (limit <= 50) {
    return(limit)
  }

  message(paste0(
    "Nominatim returns at most 50 results per query. ",
    "`limit` has been set to 50."
  ))
  min(50, limit)
}

progress_lapply <- function(n, progressbar, f) {
  progressbar <- all(progressbar, n > 1)

  if (progressbar) {
    pb <- txtProgressBar(min = 0, max = n, width = 50, style = 3)
    on.exit(close(pb))
  }

  lapply(seq_len(n), function(i) {
    if (progressbar) {
      setTxtProgressBar(pb, i)
    }
    f(i)
  })
}

cap_coordinates <- function(lat, long) {
  if (!is.numeric(lat) || !is.numeric(long)) {
    stop("`lat` and `long` must be numeric.")
  }

  if (length(lat) != length(long)) {
    stop("`lat` and `long` must have the same length.")
  }

  if (anyNA(lat) || anyNA(long)) {
    stop("`lat` and `long` must not contain missing values.")
  }

  lat_cap <- pmax(pmin(lat, 90), -90)
  if (any(lat_cap != lat)) {
    message("Latitude values outside [-90, 90] were clamped to that range.")
  }

  long_cap <- pmax(pmin(long, 180), -180)
  if (any(long_cap != long)) {
    message("Longitude values outside [-180, 180] were clamped to that range.")
  }

  list(lat = lat_cap, long = long_cap)
}

rename_coordinate_cols <- function(x, lat = "lat", long = "lon") {
  nmes <- names(x)
  nmes[nmes == "lat"] <- lat
  nmes[nmes == "lon"] <- long
  names(x) <- nmes

  x
}

convert_coordinate_cols <- function(x, lat = "lat", long = "lon") {
  x[lat] <- as.double(x[[lat]])
  x[long] <- as.double(x[[long]])

  x
}

normalize_bbox <- function(bbox) {
  if (any(inherits(bbox, "sf"), inherits(bbox, "sfc"))) {
    bbox <- sf::st_transform(bbox, 4326)
    bbox <- sf::st_bbox(bbox)
  }

  as.vector(bbox)
}

structured_query_params <- function(
  amenity = NULL,
  street = NULL,
  city = NULL,
  county = NULL,
  state = NULL,
  country = NULL,
  postalcode = NULL
) {
  pars <- list(
    amenity = amenity[1],
    street = street[1],
    city = city[1],
    county = county[1],
    state = state[1],
    country = country[1],
    postalcode = postalcode[1]
  )

  lapply(pars, function(x) {
    if (is.null(x)) {
      return(NA_character_)
    }
    as.character(x)
  })
}

structured_query_tbl <- function(pars) {
  tbl_query <- dplyr::as_tibble(pars)
  names(tbl_query) <- paste0("q_", names(tbl_query))

  tbl_query
}

compact_query_options <- function(x) {
  is_missing <- vapply(
    x,
    function(option) {
      any(is.null(option), is.na(option))
    },
    FUN.VALUE = logical(1)
  )

  x[!is_missing]
}

encode_query_value <- function(x) {
  if (is.logical(x)) {
    x <- ifelse(isTRUE(x), 1, 0)
  }
  paste0(x, collapse = ",")
}

encode_search_text <- function(x) {
  gsub(" ", "+", x, fixed = TRUE)
}

add_address_details <- function(url, full_results, always = FALSE) {
  if (isTRUE(full_results)) {
    return(paste0(url, "&addressdetails=1"))
  }

  if (always) {
    return(paste0(url, "&addressdetails=0"))
  }

  url
}

add_polygon_geojson <- function(url, points_only) {
  if (!isTRUE(points_only)) {
    return(paste0(url, "&polygon_geojson=1"))
  }

  url
}

build_search_url <- function(
  nominatim_server,
  format,
  limit,
  full_results,
  custom_query = list(),
  query = NULL,
  points_only = TRUE
) {
  if (is.null(query)) {
    url <- paste0(
      prepare_api_url(nominatim_server, "search?"),
      "format=",
      format,
      "&limit=",
      limit
    )
  } else {
    url <- paste0(
      prepare_api_url(nominatim_server, "search?q="),
      encode_search_text(query),
      "&format=",
      format,
      "&limit=",
      limit
    )
  }

  url <- add_address_details(url, full_results)
  url <- add_polygon_geojson(url, points_only)
  add_custom_query(custom_query, url)
}

build_reverse_url <- function(
  nominatim_server,
  lat,
  long,
  format,
  full_results,
  custom_query = list(),
  points_only = TRUE
) {
  url <- paste0(
    prepare_api_url(nominatim_server, "reverse?"),
    "lat=",
    lat,
    "&lon=",
    long,
    "&format=",
    format
  )

  url <- add_polygon_geojson(url, points_only)
  url <- add_address_details(url, full_results, always = TRUE)
  add_custom_query(custom_query, url)
}

build_lookup_url <- function(
  nominatim_server,
  nodes,
  full_results,
  custom_query = list()
) {
  url <- paste0(
    prepare_api_url(nominatim_server, "lookup?"),
    "osm_ids=",
    nodes,
    "&format=jsonv2"
  )

  url <- add_address_details(url, full_results)
  add_custom_query(custom_query, url)
}

reverse_query_keys <- function(lat, long) {
  coords <- cap_coordinates(lat, long)
  init_key <- dplyr::tibble(
    lat_key_int = lat,
    long_key_int = long,
    lat_cap_int = coords$lat,
    long_cap_int = coords$long
  )

  list(init = init_key, unique = dplyr::distinct(init_key))
}

run_reverse_queries <- function(key, progressbar, f) {
  all_res <- progress_lapply(nrow(key), progressbar, function(i) {
    rw <- key[i, ]
    res_single <- f(
      lat_cap = as.double(rw$lat_cap_int),
      long_cap = as.double(rw$long_cap_int)
    )

    dplyr::bind_cols(res_single, rw[, c(1, 2)])
  })

  dplyr::bind_rows(all_res)
}

add_custom_query <- function(custom_query = list(), url) {
  if (any(length(custom_query) == 0, isFALSE(is_named(custom_query)))) {
    return(url)
  }

  custom_query <- lapply(custom_query, encode_query_value)

  opts <- paste0(names(custom_query), "=", custom_query, collapse = "&")

  end_url <- paste0(url, "&", opts)

  end_url
}

is_named <- function(x) {
  nm <- names(x)

  if (is.null(nm)) {
    return(FALSE)
  }
  if (anyNA(nm)) {
    return(FALSE)
  }
  if (!all(nzchar(nm))) {
    return(FALSE)
  }

  TRUE
}

keep_names <- function(
  x,
  return_addresses,
  full_results,
  colstokeep = "query"
) {
  x$address <- x$display_name
  if ("boundingbox" %in% names(x)) {
    bbun <- lapply(x$boundingbox, function(y) {
      unl <- unlist(y)
      bb <- dplyr::tibble(boundingbox = list(as.double(unl)))
      bb
    })
    bbun <- dplyr::bind_rows(bbun)
    cln <- x[, names(x) != "boundingbox"]
    x <- dplyr::bind_cols(cln, bbun)
  }

  out_cols <- colstokeep
  if (return_addresses) {
    out_cols <- c(out_cols, "address")
  }
  if (full_results) {
    out_cols <- c(out_cols, "address", names(x))
  }

  out_cols <- unique(out_cols)
  out <- x[, out_cols]

  out
}

keep_names_rev <- function(
  x,
  address = "address",
  return_coords = FALSE,
  full_results = FALSE,
  colstokeep = address
) {
  x$xxxyyyzzz <- x$display_name
  nm <- names(x)
  nm <- gsub("xxxyyyzzz", address, nm, fixed = TRUE)
  names(x) <- nm
  out_cols <- colstokeep
  if (return_coords) {
    out_cols <- c(out_cols, "lat", "lon")
  }
  if (full_results) {
    out_cols <- c(out_cols, "lat", "lon", names(x))
  }

  out_cols <- unique(out_cols)
  out <- x[, out_cols]

  out
}

prepare_api_url <- function(
  nominatim_server = "https://nominatim.openstreetmap.org/",
  entry
) {
  api <- paste0(gsub("/$", "", nominatim_server), "/", entry)
  api
}

# Tibble helpers ----

empty_tbl <- function(x, lat, lon) {
  init_nm <- names(x)
  x <- dplyr::as_tibble(x)
  x$lat <- as.double(NA)
  x$lon <- x$lat

  names(x) <- c(init_nm, lat, lon)

  x
}

empty_tbl_rev <- function(x, address) {
  init_nm <- names(x)
  x <- dplyr::as_tibble(x)
  x$n <- as.character(NA)

  names(x) <- c(init_nm, address)

  # Put the address column first.
  x <- x[, c(address, init_nm)]

  x
}

unnest_reverse <- function(x) {
  # Unnest nested fields.

  lngths <- lengths(x)

  # Remove `NULL` fields.
  x <- x[lngths > 0]

  endobj <- dplyr::as_tibble(x[lngths == 1])

  # Add OSM address fields.
  if ("address" %in% names(x)) {
    ad <- dplyr::as_tibble(x$address)[1, ]
    names(ad) <- paste0("address.", names(ad))

    endobj <- dplyr::bind_cols(endobj, ad)
  }

  if ("extratags" %in% names(x)) {
    xtra <- dplyr::as_tibble(x$extratags)[1, ]
    names(xtra) <- paste0("extratags.", names(xtra))
    endobj <- dplyr::bind_cols(endobj, xtra)
  }

  if ("boundingbox" %in% names(x)) {
    bb <- dplyr::tibble(boundingbox = list(as.double(x$boundingbox)))
    endobj <- dplyr::bind_cols(endobj, bb)
  }

  endobj
}
# sf helpers ----
empty_sf <- function(x) {
  x <- dplyr::as_tibble(x)
  x$geometry <- "POINT EMPTY"

  out <- sf::st_as_sf(x, wkt = "geometry", crs = sf::st_crs(4326))

  out
}

sf_to_tbl <- function(x) {
  if (all(!inherits(x, "tbl"), inherits(x, "sf"))) {
    # Add the expected `sf` tibble classes when they are missing.
    template <- class(empty_sf(dplyr::tibble(a = 1)))
    class(x) <- template
  }

  # Reorder columns because geometry remains last even when not selected.
  x <- x[, setdiff(names(x), "geometry")]

  result_out <- sf::st_make_valid(x)

  result_out
}

unnest_sf <- function(x) {
  # Unnest nested fields.
  if ("address" %in% names(x)) {
    # Unnest address fields.
    add <- as.character(x$address)
    newadd <- lapply(add, function(x) {
      df <- jsonlite::fromJSON(x, simplifyVector = TRUE)
      dplyr::as_tibble(df)
    })

    newadd <- dplyr::bind_rows(newadd)
    names(newadd) <- paste0("address.", names(newadd))

    newsfobj <- x
    newsfobj <- x[, setdiff(names(x), "address")]
    x <- dplyr::bind_cols(newsfobj, newadd)
  }

  if ("extratags" %in% names(x)) {
    # Unnest extra tag fields.
    xtra <- as.character(x$extratags)

    newxtra <- lapply(xtra, function(x) {
      if (any(is.null(x), is.na(x))) {
        return(dplyr::tibble(xxx_empty_remove = NA))
      }
      df <- jsonlite::fromJSON(x, simplifyVector = TRUE)
      dplyr::as_tibble(df)
    })

    newxtra <- dplyr::bind_rows(newxtra)
    names(newxtra) <- paste0("extratags.", names(newxtra))

    newsfobj <- x
    newsfobj <- x[, setdiff(names(x), "extratags")]
    x <- dplyr::bind_cols(newsfobj, newxtra)
    x <- x[, setdiff(names(x), "extratags.xxx_empty_remove")]
  }

  x <- sf_to_tbl(x)

  x
}

unnest_sf_reverse <- function(x) {
  # Remove `NULL` fields.
  nulls_or_nas <- vapply(
    x,
    function(a_col) {
      any(is.null(a_col), is.na(a_col))
    },
    FUN.VALUE = logical(1)
  )

  x <- x[!nulls_or_nas]

  # Unnest nested fields.
  if ("address" %in% names(x)) {
    # Unnest address fields.
    add <- as.character(x$address)
    newadd <- lapply(add, function(x) {
      df <- jsonlite::fromJSON(x, simplifyVector = TRUE)
      dplyr::as_tibble(df)
    })

    newadd <- dplyr::bind_rows(newadd)[1, ]
    names(newadd) <- paste0("address.", names(newadd))

    newsfobj <- x
    newsfobj <- x[, setdiff(names(x), "address")]
    x <- dplyr::bind_cols(newsfobj, newadd)
  }

  if ("extratags" %in% names(x)) {
    # Unnest extra tag fields.
    xtra <- as.character(x$extratags)

    newxtra <- lapply(xtra, function(x) {
      df <- jsonlite::fromJSON(x, simplifyVector = TRUE)
      dplyr::as_tibble(df)
    })

    newxtra <- dplyr::bind_rows(newxtra)[1, ]
    names(newxtra) <- paste0("extratags.", names(newxtra))

    newsfobj <- x
    newsfobj <- x[, setdiff(names(x), "extratags")]
    x <- dplyr::bind_cols(newsfobj, newxtra)
  }

  x <- sf_to_tbl(x)

  x
}

test_that("add_custom_query() ignores unnamed options and encodes named ones", {
  u <- "http://test"
  t <- add_custom_query(custom_query = list(), url = u)
  expect_identical(u, t)

  # Unnamed argument
  t <- add_custom_query(custom_query = list(1, b = 2), url = u)
  expect_identical(u, t)

  # Unnamed argument
  t <- add_custom_query(custom_query = list(a = 1, 2), url = u)
  expect_identical(u, t)

  # Unnamed argument
  t <- add_custom_query(custom_query = list(3), url = u)
  expect_identical(u, t)

  # Check
  t <- add_custom_query(custom_query = list(a = 3, b = 3), url = u)
  expect_identical("http://test&a=3&b=3", t)
})

test_that("prepare_api_url() normalizes default and custom server URLs", {
  t <- prepare_api_url(entry = "an_entry")
  expect_identical(t, "https://nominatim.openstreetmap.org/an_entry")

  # Add trailing slash
  t2 <- prepare_api_url("https://nominatim.openstreetmap.org", "an_entry")
  expect_identical(t, t2)

  # Custom server
  t3 <- prepare_api_url("http://localhost:2322/nominatim-update", "custom")
  expect_identical(t3, "http://localhost:2322/nominatim-update/custom")
})

test_that("URL builders encode endpoint-specific query options", {
  expect_identical(
    build_search_url(
      "https://example.com",
      format = "json",
      limit = 10,
      full_results = TRUE,
      query = "New York",
      points_only = FALSE,
      custom_query = list(countrycodes = c("us", "ca"), extratags = TRUE)
    ),
    paste0(
      "https://example.com/search?q=New+York&format=json&limit=10",
      "&addressdetails=1&polygon_geojson=1&countrycodes=us,ca&extratags=1"
    )
  )

  expect_identical(
    build_search_url(
      "https://example.com",
      format = "geojson",
      limit = 1,
      full_results = FALSE,
      query = NULL
    ),
    "https://example.com/search?format=geojson&limit=1"
  )

  expect_identical(
    build_reverse_url(
      "https://example.com",
      lat = 42,
      long = -3,
      format = "json",
      full_results = FALSE
    ),
    "https://example.com/reverse?lat=42&lon=-3&format=json&addressdetails=0"
  )

  expect_identical(
    build_lookup_url(
      "https://example.com",
      nodes = "N1,W2",
      full_results = TRUE,
      custom_query = list(namedetails = TRUE)
    ),
    paste0(
      "https://example.com/lookup?osm_ids=N1,W2&format=jsonv2",
      "&addressdetails=1&namedetails=1"
    )
  )
})

test_that("cap helpers validate and report adjusted values", {
  expect_identical(cap_limit(10), 10)
  expect_message(limit <- cap_limit(51), "at most 50")
  expect_identical(limit, 50)

  expect_snapshot(error = TRUE, cap_coordinates("a", 1))
  expect_snapshot(error = TRUE, cap_coordinates(1, c(1, 2)))
  expect_snapshot(error = TRUE, cap_coordinates(NA_real_, 1))
  expect_snapshot(error = TRUE, cap_coordinates(1, NA_real_))

  expect_silent(integer_coords <- cap_coordinates(1L, 1L))
  expect_equal(integer_coords, list(lat = 1, long = 1))

  expect_snapshot(coords <- cap_coordinates(200, -200))
  expect_identical(coords$lat, 90)
  expect_identical(coords$long, -180)
})

test_that("sf_to_tbl() restores sf tibble classes", {
  normal_sf <- sf::st_as_sf(
    data.frame(x = 1, lon = 0, lat = 0),
    coords = c("lat", "lon"),
    crs = 4326
  )
  expect_s3_class(normal_sf, c("sf", "data.frame"), exact = TRUE)

  tbl_sf <- sf_to_tbl(normal_sf)
  expect_s3_class(tbl_sf, c("sf", "tbl_df"))
})

test_that("normalize_bbox() transforms sf and sfc inputs to coordinates", {
  bbox <- c(1, 2, 3, 4)
  bbox_sfc <- bbox_to_poly(bbox)
  bbox_sf <- sf::st_sf(id = 1, geometry = bbox_sfc)

  expect_equal(normalize_bbox(bbox_sfc), bbox)
  expect_equal(normalize_bbox(bbox_sf), bbox)
})

test_that("unnest_sf() removes placeholders for missing extra tags", {
  sfobj <- sf::st_as_sf(
    data.frame(extratags = NA_character_, lon = 0, lat = 0),
    coords = c("lon", "lat"),
    crs = 4326
  )

  out <- unnest_sf(sfobj)

  expect_s3_class(out, "sf")
  expect_named(out, "geometry")
})

test_that("is_named() distinguishes complete names from missing names", {
  # No names -> first branch
  expect_false(is_named(1:3))
  expect_false(is_named(list(1, 2, 3)))

  # NA names -> second branch
  x_na <- c(a = 1, b = 2, c = 3)
  names(x_na)[3] <- NA_character_
  expect_false(is_named(x_na))

  # Empty names -> third branch
  x_empty <- c(a = 1, b = 2, c = 3)
  names(x_empty)[2] <- ""
  expect_false(is_named(x_empty))

  # Valid names -> TRUE branch
  expect_true(is_named(c(a = 1, b = 2, c = 3)))
  expect_true(is_named(list(a = 1, b = 2, c = 3)))

  # Edge case: empty but named object
  expect_true(is_named(setNames(numeric(), character())))
})

test_that("message helpers report unavailable and empty API responses", {
  expect_message(
    unavailable <- message_api_unavailable("https://example.com/search"),
    "Could not reach the API endpoint: https://example.com/search\\."
  )
  expect_null(unavailable)

  expect_message(empty <- message_no_results(), "No results were found")
  expect_null(empty)

  expect_message(
    empty_query <- message_no_results("Madrid"),
    "No results were found for the query: Madrid\\."
  )
  expect_null(empty_query)
})

test_that("query helpers normalize structured and custom options", {
  pars <- structured_query_params(
    amenity = c("cafe", "pub"),
    city = "New York",
    country = "US"
  )

  expect_identical(
    pars,
    list(
      amenity = "cafe",
      street = NA_character_,
      city = "New York",
      county = NA_character_,
      state = NA_character_,
      country = "US",
      postalcode = NA_character_
    )
  )

  query <- structured_query_tbl(pars)
  expect_s3_class(query, "tbl")
  expect_named(
    query,
    c(
      "q_amenity",
      "q_street",
      "q_city",
      "q_county",
      "q_state",
      "q_country",
      "q_postalcode"
    )
  )

  compact <- compact_query_options(
    list(
      city = "New York",
      county = NULL,
      state = NA_character_,
      bounded = FALSE
    )
  )
  expect_identical(compact, list(city = "New York", bounded = FALSE))
  expect_identical(encode_query_value(TRUE), "1")
  expect_identical(encode_query_value(FALSE), "0")
  expect_identical(encode_query_value(c("us", "ca")), "us,ca")
  expect_identical(encode_search_text("New York City"), "New+York+City")
})

test_that("URL option helpers add only requested response details", {
  url <- "https://example.com/search?format=json"

  expect_identical(add_address_details(url, FALSE), url)
  expect_identical(
    add_address_details(url, FALSE, always = TRUE),
    paste0(url, "&addressdetails=0")
  )
  expect_identical(
    add_address_details(url, TRUE),
    paste0(url, "&addressdetails=1")
  )
  expect_identical(add_polygon_geojson(url, TRUE), url)
  expect_identical(
    add_polygon_geojson(url, FALSE),
    paste0(url, "&polygon_geojson=1")
  )
})

test_that("coordinate helpers rename and convert API columns", {
  input <- dplyr::tibble(lat = "40.5", lon = "-3.5", value = 1L)

  renamed <- rename_coordinate_cols(input, lat = "latitude", long = "longitude")
  expect_named(renamed, c("latitude", "longitude", "value"))

  converted <- convert_coordinate_cols(
    renamed,
    lat = "latitude",
    long = "longitude"
  )
  expect_type(converted$latitude, "double")
  expect_type(converted$longitude, "double")
  expect_identical(converted$latitude, 40.5)
  expect_identical(converted$longitude, -3.5)
})

test_that("reverse query helpers clamp, deduplicate and bind coordinates", {
  expect_message(
    keys <- reverse_query_keys(c(100, 100, 40), c(20, 20, -3)),
    "Latitude values outside"
  )
  expect_message(
    longitude_keys <- reverse_query_keys(40, 200),
    "Longitude values outside"
  )

  expect_equal(nrow(keys$init), 3)
  expect_equal(nrow(keys$unique), 2)
  expect_identical(keys$unique$lat_cap_int, c(90, 40))
  expect_identical(keys$unique$long_cap_int, c(20, -3))
  expect_identical(longitude_keys$unique$long_cap_int, 180)

  out <- run_reverse_queries(
    keys$unique,
    progressbar = FALSE,
    function(lat_cap, long_cap) {
      dplyr::tibble(result = paste(lat_cap, long_cap))
    }
  )

  expect_named(
    out,
    c("result", "lat_key_int", "long_key_int")
  )
  expect_identical(out$result, c("90 20", "40 -3"))
})

test_that("progress_lapply() preserves values with and without progress", {
  quiet <- progress_lapply(1, FALSE, \(i) i * 2)
  expect_identical(quiet, list(2))

  expect_output(
    visible <- progress_lapply(2, TRUE, \(i) i * 2),
    "100%",
    fixed = TRUE
  )
  expect_identical(visible, list(2, 4))
})

test_that("keep_names() selects compact fields and unnests bounding boxes", {
  input <- dplyr::tibble(
    query = "Madrid",
    lat = "40.4",
    lon = "-3.7",
    display_name = "Madrid, Spain",
    boundingbox = list(c("40", "41", "-4", "-3")),
    place_id = 1L
  )

  compact <- keep_names(
    input,
    return_addresses = TRUE,
    full_results = FALSE,
    colstokeep = c("query", "lat", "lon")
  )
  expect_named(compact, c("query", "lat", "lon", "address"))
  expect_identical(compact$address, "Madrid, Spain")

  full <- keep_names(
    input,
    return_addresses = FALSE,
    full_results = TRUE,
    colstokeep = c("query", "lat", "lon")
  )
  expect_contains(names(full), c("address", "boundingbox", "place_id"))
  expect_identical(full$boundingbox[[1]], c(40, 41, -4, -3))
})

test_that("keep_names_rev() controls reverse-geocoding output fields", {
  input <- dplyr::tibble(
    display_name = "Madrid, Spain",
    lat = 40.4,
    lon = -3.7,
    place_id = 1L
  )

  compact <- keep_names_rev(input, address = "place")
  expect_named(compact, "place")
  expect_identical(compact$place, "Madrid, Spain")

  coordinates <- keep_names_rev(input, return_coords = TRUE)
  expect_named(coordinates, c("address", "lat", "lon"))

  full <- keep_names_rev(input, full_results = TRUE)
  expect_contains(names(full), c("address", "lat", "lon", "place_id"))
})

test_that("empty output helpers preserve schemas and spatial metadata", {
  query <- dplyr::tibble(query = "missing")

  tabular <- empty_tbl(query, lat = "latitude", lon = "longitude")
  expect_named(tabular, c("query", "latitude", "longitude"))
  expect_type(tabular$latitude, "double")
  expect_type(tabular$longitude, "double")
  expect_all_true(is.na(tabular$latitude))
  expect_all_true(is.na(tabular$longitude))

  reverse <- empty_tbl_rev(dplyr::tibble(lat = 1, lon = 2), "place")
  expect_named(reverse, c("place", "lat", "lon"))
  expect_type(reverse$place, "character")
  expect_all_true(is.na(reverse$place))

  spatial <- empty_sf(query)
  expect_s3_class(spatial, "sf")
  expect_all_true(sf::st_is_empty(spatial))
  expect_identical(sf::st_crs(spatial), sf::st_crs(4326))
})

test_that("unnest_reverse() expands nested API fields", {
  input <- list(
    place_id = 1L,
    display_name = "Madrid, Spain",
    address = list(city = "Madrid", country = "Spain"),
    extratags = list(wikidata = "Q2807"),
    boundingbox = c("40", "41", "-4", "-3"),
    unused = NULL
  )

  out <- unnest_reverse(input)

  expect_s3_class(out, "tbl")
  expect_contains(
    names(out),
    c(
      "place_id",
      "display_name",
      "address.city",
      "address.country",
      "extratags.wikidata",
      "boundingbox"
    )
  )
  expect_no_match(names(out), "unused", fixed = TRUE)
  expect_identical(out$boundingbox[[1]], c(40, 41, -4, -3))
})

test_that("sf unnest helpers expand nested fields and remove missing columns", {
  input <- sf::st_as_sf(
    dplyr::tibble(
      address = '{"city":"Madrid","country":"Spain"}',
      extratags = '{"wikidata":"Q2807"}',
      missing = NA_character_,
      lon = -3.7,
      lat = 40.4
    ),
    coords = c("lon", "lat"),
    crs = 4326
  )

  search <- unnest_sf(input)
  expect_s3_class(search, "sf")
  expect_contains(
    names(search),
    c("address.city", "address.country", "extratags.wikidata")
  )

  reverse <- unnest_sf_reverse(input)
  expect_s3_class(reverse, "sf")
  expect_contains(
    names(reverse),
    c("address.city", "address.country", "extratags.wikidata")
  )
  expect_no_match(names(reverse), "missing", fixed = TRUE)
})

#' Reverse geocode coordinates and return \CRANpkg{sf} objects
#'
#' @description
#' Reverse geocodes latitude and longitude coordinates and returns matching
#' results as an [`sf`][sf::st_sf] object. Latitude values must be in
#' \eqn{\left[-90, 90 \right]} and longitude values in
#' \eqn{\left[-180, 180 \right]}. Use [reverse_geo_lite()] to return a
#' [tibble][dplyr::tibble] instead.
#'
#' @inherit reverse_geo_lite details
#'
#' @inheritParams reverse_geo_lite -full_results
#' @inheritParams geo_lite_sf full_results points_only
#'
#' @inherit geo_lite_sf return
#'
#' @inheritSection reverse_geo_lite About zooming
#' @inheritSection geo_lite_sf About geometry types
#'
#' @family reverse
#' @family spatial
#'
#' @export
#' @encoding UTF-8
#'
#' @examplesIf nominatim_check_access()
#' \donttest{
#' library(ggplot2)
#'
#' # Define the Colosseum coordinates.
#' col_lon <- 12.49309
#' col_lat <- 41.89026
#'
#' # Return the Colosseum as a polygon.
#' col_sf <- reverse_geo_lite_sf(
#'   lat = col_lat,
#'   long = col_lon,
#'   points_only = FALSE
#' )
#'
#' dplyr::glimpse(col_sf)
#'
#' if (!all(sf::st_is_empty(col_sf))) {
#'   ggplot(col_sf) +
#'     geom_sf()
#' }
#'
#' # Return the city of Rome by using the same coordinates with zoom 10.
#'
#' rome_sf <- reverse_geo_lite_sf(
#'   lat = col_lat,
#'   long = col_lon,
#'   custom_query = list(zoom = 10),
#'   points_only = FALSE
#' )
#'
#' dplyr::glimpse(rome_sf)
#'
#' if (!all(sf::st_is_empty(rome_sf))) {
#'   ggplot(rome_sf) +
#'     geom_sf()
#' }
#' }
reverse_geo_lite_sf <- function(
  lat,
  long,
  address = "address",
  full_results = FALSE,
  return_coords = TRUE,
  verbose = FALSE,
  nominatim_server = "https://nominatim.openstreetmap.org/",
  progressbar = TRUE,
  custom_query = list(),
  points_only = TRUE
) {
  keys <- reverse_query_keys(lat, long)
  all_res <- run_reverse_queries(
    keys$unique,
    progressbar,
    function(lat_cap, long_cap) {
      reverse_geo_lite_sf_single(
        lat_cap = lat_cap,
        long_cap = long_cap,
        address = address,
        full_results = full_results,
        return_coords = return_coords,
        verbose = verbose,
        custom_query = custom_query,
        points_only = points_only,
        nominatim_server = nominatim_server
      )
    }
  )

  # Restore duplicate inputs in `sf` output.
  if (!identical(nrow(keys$init), nrow(all_res))) {
    # Join with row indexes.
    tmplt <- sf::st_drop_geometry(all_res)[, c("lat_key_int", "long_key_int")]
    tmplt$rindex <- seq_len(nrow(tmplt))
    getrows <- dplyr::left_join(
      keys$init,
      tmplt,
      by = c("lat_key_int", "long_key_int")
    )

    # Restore the original row order.
    all_res <- all_res[as.double(getrows$rindex), ]
  }

  # Remove internal join keys.
  kpnms <- setdiff(names(all_res), c("lat_key_int", "long_key_int"))

  all_res <- all_res[, kpnms]

  all_res <- sf_to_tbl(all_res)

  all_res
}
#' @inheritParams reverse_geo_lite_sf
#' @noRd
reverse_geo_lite_sf_single <- function(
  lat_cap,
  long_cap,
  address = "address",
  full_results = TRUE,
  return_coords = TRUE,
  verbose = TRUE,
  nominatim_server = "https://nominatim.openstreetmap.org/",
  custom_query = list(),
  points_only = FALSE
) {
  url <- build_reverse_url(
    nominatim_server = nominatim_server,
    lat = lat_cap,
    long = long_cap,
    format = "geojson",
    full_results = full_results,
    custom_query = custom_query,
    points_only = points_only
  )

  # Download the API response.
  json <- api_call(url, ".geojson", isFALSE(verbose))

  tbl_query <- dplyr::tibble(lat = lat_cap, lon = long_cap)

  if (isFALSE(json)) {
    message_api_unavailable(url)
    out <- empty_sf(empty_tbl_rev(tbl_query, address))
    return(invisible(out))
  }

  # Handle empty queries.
  result_init <- jsonlite::fromJSON(json, flatten = TRUE)
  if ("error" %in% names(result_init)) {
    message_no_results(paste0("lat = ", lat_cap, ", long = ", long_cap))
    out <- empty_sf(empty_tbl_rev(tbl_query, address))
    return(invisible(out))
  }

  # Read the `sf` object.
  sfobj <- sf::read_sf(json, stringsAsFactors = FALSE)

  # Unnest nested fields.
  sfobj <- unnest_sf_reverse(sfobj)

  # Add latitude and longitude.
  sf_clean <- dplyr::bind_cols(sfobj, tbl_query)

  # Keep selected columns.
  result_out <- keep_names_rev(sf_clean, address, return_coords, full_results)

  # Restore tibble classes.
  result_out <- sf_to_tbl(result_out)

  result_out
}

list(
  rd_family_title = list(
    amenity = "Amenity lookup functions and data:",
    geocoding = "Address search functions:",
    lookup = "Address lookup functions:",
    reverse = "Reverse geocoding functions:",
    spatial = "Spatial helper and output functions:"
  )
)

#' Search for addresses with free-form queries and return \CRANpkg{sf} objects
#'
#' @description
#' Searches for addresses supplied as a character vector and returns matching
#' results as an [`sf`][sf::st_sf] object. Use [geo_lite()] to return a
#' [tibble][dplyr::tibble] instead.
#'
#' This function performs the **free-form address search** described in the
#' [API endpoint](https://nominatim.org/release-docs/latest/api/Search/).
#'
#' @inherit geo_lite details
#'
#' @param full_results A logical value indicating whether to return all
#'   available fields from the Nominatim API. If `FALSE`, only query metadata,
#'   geometry and requested address columns are returned.
#' @param points_only A logical value indicating whether to return only point
#'   geometries. If `FALSE`, the API may return other geometry types. See
#'   **About geometry types**.
#' @inheritParams geo_lite
#'
#' @returns
#' An [`sf`][sf::st_sf] object with the results that match the query.
#'
#' @section About geometry types:
#'
#' The `points_only` argument controls whether the results contain only points.
#' All Nominatim results have at least a point geometry.
#'
#' When `points_only = FALSE`, the geometry type depends on the matching
#' feature. Administrative areas and major buildings are returned as polygons,
#' rivers and roads are returned as lines and amenities may still be returned
#' as points.
#'
#' This function is vectorized, allowing multiple addresses to be searched.
#' With `points_only = FALSE`, multiple geometry types may be returned.
#'
#' @family geocoding
#' @family spatial
#'
#' @export
#' @encoding UTF-8
#'
#' @examplesIf nominatim_check_access()
#' \donttest{
#' # Return point geometries.
#' library(ggplot2)
#'
#' string <- "Statue of Liberty, NY, USA"
#' sol <- geo_lite_sf(string)
#'
#' if (!all(sf::st_is_empty(sol))) {
#'   ggplot(sol) +
#'     geom_sf()
#' }
#'
#' sol_poly <- geo_lite_sf(string, points_only = FALSE)
#'
#' if (!all(sf::st_is_empty(sol_poly))) {
#'   ggplot(sol_poly) +
#'     geom_sf() +
#'     geom_sf(data = sol, color = "red")
#' }
#' # Return multiple matches.
#'
#' madrid <- geo_lite_sf("Comunidad de Madrid, Spain",
#'   limit = 2,
#'   points_only = FALSE, full_results = TRUE
#' )
#'
#' if (!all(sf::st_is_empty(madrid))) {
#'   ggplot(madrid) +
#'     geom_sf(fill = NA)
#' }
#' }
geo_lite_sf <- function(
  address,
  limit = 1,
  return_addresses = TRUE,
  full_results = FALSE,
  verbose = FALSE,
  progressbar = TRUE,
  nominatim_server = "https://nominatim.openstreetmap.org/",
  custom_query = list(),
  points_only = TRUE
) {
  limit <- cap_limit(limit)

  # Deduplicate queries before calling the API.
  init_key <- dplyr::tibble(query = address)
  key <- unique(address)

  ntot <- length(key)

  # Run one request per unique query.
  all_res <- progress_lapply(ntot, progressbar, function(x) {
    ad <- key[x]
    geo_lite_sf_single(
      address = ad,
      limit,
      return_addresses,
      full_results,
      verbose,
      custom_query,
      points_only,
      nominatim_server = nominatim_server
    )
  })

  all_res <- dplyr::bind_rows(all_res)

  all_res <- sf_to_tbl(all_res)

  # Restore duplicate inputs in `sf` output.
  if (!identical(as.character(init_key$query), key)) {
    # Join with row indexes.
    template <- sf::st_drop_geometry(all_res)[, "query"]
    template$rindex <- seq_len(nrow(template))
    getrows <- dplyr::left_join(init_key, template, by = "query")

    # Restore the original row order.
    all_res <- all_res[as.double(getrows$rindex), ]
    all_res <- sf_to_tbl(all_res)
  }

  all_res
}

#' @inheritParams geo_lite
#' @noRd

geo_lite_sf_single <- function(
  address,
  limit = 1,
  return_addresses = TRUE,
  full_results = FALSE,
  verbose = FALSE,
  nominatim_server = "https://nominatim.openstreetmap.org/",
  custom_query = list(),
  points_only = TRUE
) {
  url <- build_search_url(
    nominatim_server = nominatim_server,
    format = "geojson",
    limit = limit,
    full_results = full_results,
    custom_query = custom_query,
    query = address,
    points_only = points_only
  )

  # Download the API response.
  json <- api_call(url, ".geojson", isFALSE(verbose))

  # Keep the original query value.
  tbl_query <- dplyr::tibble(query = address)

  if (isFALSE(json)) {
    message_api_unavailable(url)
    out <- empty_sf(tbl_query)
    return(invisible(out))
  }

  # Read the `sf` object.
  sfobj <- sf::read_sf(json, stringsAsFactors = FALSE)

  # Handle empty queries.
  if (length(names(sfobj)) == 1) {
    message_no_results(address)
    out <- empty_sf(tbl_query)
    return(invisible(out))
  }

  # Unnest nested fields.
  sfobj <- unnest_sf(sfobj)

  # Add the query to the API results.
  sf_clean <- sfobj
  sf_clean$query <- address

  # Keep selected columns.
  result_out <- keep_names(
    sf_clean,
    return_addresses,
    full_results,
    colstokeep = "query"
  )

  # Restore tibble classes.
  result_out <- sf_to_tbl(result_out)

  result_out
}

#' Search for addresses with structured queries
#'
#' @description
#' Searches for addresses already split into components and returns matching
#' results as a [tibble][dplyr::tibble]. Use [geo_lite_struct_sf()] to return
#' an [`sf`][sf::st_sf] object instead.
#'
#' This function performs the **structured address search** described in the
#' [API endpoint](https://nominatim.org/release-docs/latest/api/Search/). To
#' perform a free-form search, use [geo_lite()].
#'
#' @details
#' A structured address search accepts an address already split into components.
#' Each argument represents an address field. All components are optional, so
#' provide only those relevant to the address you want to find.
#'
#' See <https://nominatim.org/release-docs/latest/api/Search/> for additional
#' parameters to pass to `custom_query`.
#'
#' @param amenity A character string specifying the name or type of amenity. See
#'   [geo_amenity()].
#' @param street A character string specifying the house number and street name.
#' @param city A character string specifying the city.
#' @param county A character string specifying the county.
#' @param state A character string specifying the state.
#' @param country A character string specifying the country.
#' @param postalcode A character string specifying the postal code.
#' @inheritParams geo_lite
#'
#' @inherit geo_lite return
#'
#' @family geocoding
#'
#' @export
#' @encoding UTF-8
#'
#' @examplesIf nominatim_check_access()
#' \donttest{
#' pl_mayor <- geo_lite_struct(
#'   street = "Plaza Mayor", country = "Spain",
#'   limit = 50, full_results = TRUE
#' )
#'
#' dplyr::glimpse(pl_mayor)
#' }
geo_lite_struct <- function(
  amenity = NULL,
  street = NULL,
  city = NULL,
  county = NULL,
  state = NULL,
  country = NULL,
  postalcode = NULL,
  lat = "lat",
  long = "lon",
  limit = 1,
  full_results = FALSE,
  return_addresses = TRUE,
  verbose = FALSE,
  nominatim_server = "https://nominatim.openstreetmap.org/",
  custom_query = list()
) {
  limit <- cap_limit(limit)

  # Keep the first value of each parameter. This function is not vectorized.
  pars <- structured_query_params(
    amenity = amenity,
    street = street,
    city = city,
    county = county,
    state = state,
    country = country,
    postalcode = postalcode
  )
  tbl_query <- structured_query_tbl(pars)

  if (all(is.na(pars))) {
    message("No search parameters were provided.")
    out <- empty_tbl(tbl_query, lat, long)
    return(invisible(out))
  }

  pars <- lapply(pars, encode_search_text)
  custom_query <- compact_query_options(c(pars, custom_query))
  url <- build_search_url(
    nominatim_server = nominatim_server,
    format = "jsonv2",
    limit = limit,
    full_results = full_results,
    custom_query = custom_query
  )

  # Download the API response.
  json <- api_call(url, ".json", isFALSE(verbose))

  if (isFALSE(json)) {
    message_api_unavailable(url)
    out <- empty_tbl(tbl_query, lat, long)
    return(invisible(out))
  }

  result <- dplyr::as_tibble(jsonlite::fromJSON(json, flatten = TRUE))

  result <- rename_coordinate_cols(result, lat, long)

  # Handle empty queries.
  if (nrow(result) == 0) {
    message_no_results()
    out <- empty_tbl(tbl_query, lat, long)
    return(invisible(out))
  }

  result <- convert_coordinate_cols(result, lat, long)

  # Add the structured query to the API results.
  result_clean <- dplyr::bind_cols(tbl_query[rep(1, nrow(result)), ], result)

  # Keep selected columns.
  result_out <- keep_names(
    result_clean,
    return_addresses,
    full_results,
    colstokeep = c(names(tbl_query), lat, long)
  )

  # Restore tibble classes.
  result_out <- dplyr::as_tibble(result_out)

  result_out
}

#' Look up OpenStreetMap amenities
#'
#' @description
#' Looks up OpenStreetMap [amenities][osm_amenities] within a bounding box of
#' the form `(xmin, ymin, xmax, ymax)`. Results are returned as a
#' [tibble][dplyr::tibble]. Use [geo_amenity_sf()] to return an
#' [`sf`][sf::st_sf] object instead.
#'
#' @details
#' Bounding boxes can be located using online tools such as
#' <https://boundingbox.klokantech.com/>.
#'
#' For a full list of valid amenities, see
#' <https://wiki.openstreetmap.org/wiki/Key:amenity> and [osm_amenities].
#'
#' See <https://nominatim.org/release-docs/latest/api/Search/> for additional
#' parameters to pass to `custom_query`.
#'
#' @param bbox A numeric vector, an [`sf`][sf::st_sf] object or an
#'   [`sfc`][sf::st_sfc] object specifying a bounding box (viewbox) used to
#'   limit the search. Numeric vectors must contain **longitude** (`x`) and
#'   **latitude** (`y`) in the form `(xmin, ymin, xmax, ymax)`. See **Details**.
#' @param amenity A character vector of amenities to look up,
#'   for example `c("pub", "restaurant")`. See [osm_amenities].
#' @param strict A logical value indicating whether to keep only results inside
#'   `bbox`. If `FALSE` (the default), Nominatim may return results outside the
#'   bounding box.
#' @inheritParams geo_lite
#'
#' @inherit geo_lite return
#'
#' @seealso
#' [bbox_to_poly()] for converting bounding box coordinates to a
#' \CRANpkg{sf} polygon.
#'
#' @family amenity
#'
#' @export
#' @encoding UTF-8
#'
#' @examplesIf nominatim_check_access()
#' \donttest{
#' # Define a bounding box around Times Square, New York.
#' bbox <- c(
#'   -73.9894467311, 40.75573629,
#'   -73.9830630737, 40.75789245
#' )
#'
#' geo_amenity(
#'   bbox = bbox,
#'   amenity = "restaurant"
#' )
#'
#' # Search for multiple amenities.
#' geo_amenity(
#'   bbox = bbox,
#'   amenity = c("restaurant", "pub")
#' )
#'
#' # Increase `limit` and use strict filtering.
#' geo_amenity(
#'   bbox = bbox,
#'   amenity = c("restaurant", "pub"),
#'   limit = 10,
#'   strict = TRUE
#' )
#' }
geo_amenity <- function(
  bbox,
  amenity,
  lat = "lat",
  long = "lon",
  limit = 1,
  full_results = FALSE,
  return_addresses = TRUE,
  verbose = FALSE,
  nominatim_server = "https://nominatim.openstreetmap.org/",
  progressbar = TRUE,
  custom_query = list(),
  strict = FALSE
) {
  limit <- cap_limit(limit)

  bbox <- normalize_bbox(bbox)

  # Add the viewbox restriction to the custom query.
  custom_query <- as.list(custom_query)
  custom_query$viewbox <- bbox
  custom_query$bounded <- TRUE

  # Deduplicate queries.
  key <- unique(amenity)

  ntot <- length(key)

  all_res <- progress_lapply(ntot, progressbar, function(x) {
    ad <- key[x]

    geo_lite_struct(
      amenity = ad,
      lat = lat,
      long = long,
      limit = limit,
      full_results = full_results,
      return_addresses = return_addresses,
      verbose = verbose,
      nominatim_server = nominatim_server,
      custom_query = custom_query
    )
  })

  all_res <- dplyr::bind_rows(all_res)

  # Clean query columns and names.
  nm <- names(all_res)
  nm[nm == "q_amenity"] <- "query"
  names(all_res) <- nm
  all_res <- all_res[, !grepl("^q_", nm)]

  if (strict) {
    bbox_sf <- bbox_to_poly(bbox)
    all_res_sf <- sf::st_as_sf(all_res, coords = c("lon", "lat"), crs = 4326)

    int <- as.vector(sf::st_intersects(all_res_sf, bbox_sf, sparse = FALSE))
    all_res <- all_res[int, ]
  }

  all_res
}

#' Look up OpenStreetMap objects and return \CRANpkg{sf} objects
#'
#' @description
#' Looks up addresses and other details for one or more OpenStreetMap (OSM)
#' objects, such as nodes, ways or relations. Results are returned as an
#' [`sf`][sf::st_sf] object. Use [geo_address_lookup()] to return a
#' [tibble][dplyr::tibble] instead.
#'
#' @inherit geo_address_lookup details
#'
#' @inheritParams geo_address_lookup -full_results
#' @inheritParams geo_lite_sf full_results points_only
#'
#' @inherit geo_lite_sf return
#'
#' @inheritSection geo_lite_sf About geometry types
#'
#' @family lookup
#' @family spatial
#'
#' @export
#' @encoding UTF-8
#'
#' @examplesIf nominatim_check_access()
#' \donttest{
#' # Look up Notre-Dame Cathedral in Paris.
#'
#' NotreDame <- geo_address_lookup_sf(osm_ids = 201611261, type = "W")
#'
#' # Require at least one non-empty object.
#' if (!all(sf::st_is_empty(NotreDame))) {
#'   library(ggplot2)
#'
#'   ggplot(NotreDame) +
#'     geom_sf()
#' }
#'
#' NotreDame_poly <- geo_address_lookup_sf(201611261,
#'   type = "W",
#'   points_only = FALSE
#' )
#'
#' if (!all(sf::st_is_empty(NotreDame_poly))) {
#'   ggplot(NotreDame_poly) +
#'     geom_sf()
#' }
#'
#' # Look up multiple OSM objects.
#'
#' several <- geo_address_lookup_sf(c(146656, 240109189), type = c("R", "N"))
#' several
#' }
geo_address_lookup_sf <- function(
  osm_ids,
  type = c("N", "W", "R"),
  full_results = FALSE,
  return_addresses = TRUE,
  verbose = FALSE,
  nominatim_server = "https://nominatim.openstreetmap.org/",
  custom_query = list(),
  points_only = TRUE
) {
  # Build the API address.
  api <- prepare_api_url(nominatim_server, "lookup?")

  # Prepare OSM object identifiers.
  osm_ids <- as.numeric(osm_ids)
  osm_ids <- floor(abs(osm_ids))
  type <- as.character(type)
  nodes <- paste0(type, osm_ids, collapse = ",")

  # Compose the lookup URL.
  url <- paste0(api, "osm_ids=", nodes, "&format=geojson")

  if (!isTRUE(points_only)) {
    url <- paste0(url, "&polygon_geojson=1")
  }
  if (full_results) {
    url <- paste0(url, "&addressdetails=1")
  }

  # Add custom query options.
  url <- add_custom_query(custom_query, url)

  # Download the API response.
  json <- api_call(url, ".geojson", quiet = isFALSE(verbose))

  # Keep the original query values.
  tbl_query <- dplyr::tibble(query = paste0(type, osm_ids))

  # Handle missing responses.
  if (isFALSE(json)) {
    message_api_unavailable(url)
    out <- empty_sf(tbl_query)
    return(invisible(out))
  }

  # Read the `sf` object.
  sfobj <- sf::read_sf(json, stringsAsFactors = FALSE)

  # Handle empty queries.
  if (length(names(sfobj)) == 1) {
    message_no_results(nodes)
    out <- empty_sf(tbl_query)
    return(invisible(out))
  }

  # Unnest address fields.
  sfobj <- unnest_sf(sfobj)

  # Recreate `tbl_query` with normalized OSM IDs.
  tbl_query <- dplyr::tibble(query = paste0(type, osm_ids), osm_id = osm_ids)

  # Keep only matched results.
  sf_clean <- dplyr::inner_join(sfobj, tbl_query, by = "osm_id")

  # Warn about lost rows.
  if (all(nrow(sf_clean) < nrow(tbl_query), verbose)) {
    missing_ids <- setdiff(tbl_query$query, sf_clean$query)
    warning(paste0(
      "No results were found for the following OSM IDs: ",
      paste(missing_ids, collapse = ", "),
      ". The output contains only matched IDs."
    ))
  }

  # Keep selected columns.
  result_out <- keep_names(
    sf_clean,
    return_addresses,
    full_results,
    colstokeep = "query"
  )

  # Restore tibble classes.
  result_out <- sf_to_tbl(result_out)

  result_out
}

#' OpenStreetMap amenities
#'
#' @description
#' A dataset of amenity values available on OpenStreetMap.
#'
#' @format
#' A [tibble][dplyr::tibble] with
#' `r prettyNum(nrow(nominatimlite::osm_amenities), big.mark = ",")` rows and
#' three columns:
#' \describe{
#'   \item{category}{Amenity category.}
#'   \item{amenity}{Amenity value.}
#'   \item{comment}{Brief description of the amenity type.}
#' }
#'
#' @source <https://wiki.openstreetmap.org/wiki/Key:amenity>
#'
#' @note The data were extracted on **July 11, 2026**. See
#' `inst/COPYRIGHTS` for copyright and license details.
#'
#' @family amenity
#' @keywords datasets
#'
#' @docType data
#' @name osm_amenities
#' @encoding UTF-8
#'
#' @examples
#' data("osm_amenities")
#'
#' osm_amenities
NULL

#' Look up OpenStreetMap objects
#'
#' @description
#' Looks up addresses and other details for one or more OpenStreetMap (OSM)
#' objects, such as nodes, ways or relations. Results are returned as a
#' [tibble][dplyr::tibble]. Use [geo_address_lookup_sf()] to return an
#' [`sf`][sf::st_sf] object instead.
#'
#' @details
#' See <https://nominatim.org/release-docs/latest/api/Lookup/> for additional
#' parameters to pass to `custom_query`.
#'
#' @param osm_ids A numeric vector of OSM identifiers, for example
#'   `c(12345, 67890)`.
#' @param type A character vector containing the OSM object type associated
#'   with each value in `osm_ids`. Possible values are node (`"N"`), way
#'   (`"W"`) and relation (`"R"`). A single value is recycled.
#'
#' @inheritParams geo_lite
#' @inherit geo_lite return
#'
#' @seealso
#' [geo_lite()] for finding objects when their OSM identifiers are unknown.
#'
#' @family lookup
#'
#' @export
#' @encoding UTF-8
#'
#' @examplesIf nominatim_check_access()
#' \donttest{
#' ids <- geo_address_lookup(osm_ids = c(46240148, 34633854), type = "W")
#'
#' ids
#'
#' several <- geo_address_lookup(c(146656, 240109189), type = c("R", "N"))
#' several
#' }
geo_address_lookup <- function(
  osm_ids,
  type = c("N", "W", "R"),
  lat = "lat",
  long = "lon",
  full_results = FALSE,
  return_addresses = TRUE,
  verbose = FALSE,
  nominatim_server = "https://nominatim.openstreetmap.org/",
  custom_query = list()
) {
  # Prepare OSM object identifiers.
  osm_ids <- as.numeric(osm_ids)
  osm_ids <- floor(abs(osm_ids))
  type <- as.character(type)
  nodes <- paste0(type, osm_ids, collapse = ",")

  url <- build_lookup_url(
    nominatim_server = nominatim_server,
    nodes = nodes,
    full_results = full_results,
    custom_query = custom_query
  )

  # Download the API response.
  json <- api_call(url, ".json", isFALSE(verbose))

  # Keep the original query values.
  tbl_query <- dplyr::tibble(query = paste0(type, osm_ids))

  # Handle missing responses.
  if (isFALSE(json)) {
    message_api_unavailable(url)
    out <- empty_tbl(tbl_query, lat, long)
    return(invisible(out))
  }
  result <- dplyr::as_tibble(jsonlite::fromJSON(json, flatten = TRUE))

  result <- rename_coordinate_cols(result, lat, long)

  # Handle empty queries.
  if (nrow(result) == 0) {
    message_no_results(nodes)
    out <- empty_tbl(tbl_query, lat, long)
    return(invisible(out))
  }

  result <- convert_coordinate_cols(result, lat, long)

  # Recreate `tbl_query` with normalized OSM IDs.
  tbl_query <- dplyr::tibble(query = paste0(type, osm_ids), osm_id = osm_ids)

  # Keep only matched results.
  result_clean <- dplyr::inner_join(result, tbl_query, by = "osm_id")

  # Warn about lost rows.
  if (all(nrow(result_clean) < nrow(tbl_query), verbose)) {
    missing_ids <- setdiff(tbl_query$query, result_clean$query)
    warning(paste0(
      "No results were found for the following OSM IDs: ",
      paste(missing_ids, collapse = ", "),
      ". The output contains only matched IDs."
    ))
  }

  # Keep selected columns.
  result_out <- keep_names(
    result_clean,
    return_addresses,
    full_results,
    colstokeep = c("query", lat, long)
  )

  # Restore tibble classes.
  result_out <- dplyr::as_tibble(result_out)

  result_out
}

#' Search for addresses with structured queries and return \CRANpkg{sf} objects
#'
#' @description
#' Searches for addresses already split into components and returns matching
#' results as an [`sf`][sf::st_sf] object. Use [geo_lite_struct()] to return a
#' [tibble][dplyr::tibble] instead.
#'
#' This function performs the **structured address search** described in the
#' [API endpoint](https://nominatim.org/release-docs/latest/api/Search/). To
#' perform a free-form search, use [geo_lite_sf()].
#'
#' @inherit geo_lite_struct details
#'
#' @inheritParams geo_lite_struct -full_results
#' @inheritParams geo_lite_sf full_results points_only
#'
#' @inherit geo_lite_sf return
#'
#' @inheritSection geo_lite_sf About geometry types
#'
#' @family geocoding
#' @family spatial
#'
#' @export
#' @encoding UTF-8
#'
#' @examplesIf nominatim_check_access()
#' \donttest{
#' # Search with a structured address.
#'
#' pl_mayor <- geo_lite_struct_sf(
#'   street = "Plaza Mayor",
#'   county = "Comunidad de Madrid",
#'   country = "Spain", limit = 50,
#'   full_results = TRUE, verbose = TRUE
#' )
#'
#' # Retrieve an administrative boundary.
#' ccaa <- geo_lite_sf("Comunidad de Madrid, Spain", points_only = FALSE)
#'
#' library(ggplot2)
#'
#' if (any(!sf::st_is_empty(pl_mayor), !sf::st_is_empty(ccaa))) {
#'   ggplot(ccaa) +
#'     geom_sf() +
#'     geom_sf(data = pl_mayor, aes(shape = addresstype, color = addresstype))
#' }
#' }
geo_lite_struct_sf <- function(
  amenity = NULL,
  street = NULL,
  city = NULL,
  county = NULL,
  state = NULL,
  country = NULL,
  postalcode = NULL,
  limit = 1,
  full_results = FALSE,
  return_addresses = TRUE,
  verbose = FALSE,
  nominatim_server = "https://nominatim.openstreetmap.org/",
  custom_query = list(),
  points_only = TRUE
) {
  limit <- cap_limit(limit)

  # Keep the first value of each parameter. This function is not vectorized.
  pars <- structured_query_params(
    amenity = amenity,
    street = street,
    city = city,
    county = county,
    state = state,
    country = country,
    postalcode = postalcode
  )
  tbl_query <- structured_query_tbl(pars)

  if (all(is.na(pars))) {
    message("No search parameters were provided.")
    out <- empty_sf(tbl_query)
    return(invisible(out))
  }

  pars <- lapply(pars, encode_search_text)
  custom_query <- compact_query_options(c(pars, custom_query))
  url <- build_search_url(
    nominatim_server = nominatim_server,
    format = "geojson",
    limit = limit,
    full_results = full_results,
    custom_query = custom_query,
    points_only = points_only
  )

  # Download the API response.
  json <- api_call(url, ".geojson", isFALSE(verbose))

  if (isFALSE(json)) {
    message_api_unavailable(url)
    out <- empty_sf(tbl_query)
    return(invisible(out))
  }

  # Read the `sf` object.
  sfobj <- sf::read_sf(json, stringsAsFactors = FALSE)

  # Handle empty queries.
  if (length(names(sfobj)) == 1) {
    message_no_results()
    out <- empty_sf(tbl_query)
    return(invisible(out))
  }

  # Unnest nested fields.
  sfobj <- unnest_sf(sfobj)

  # Prepare the output with query metadata.
  sf_clean <- sfobj

  # Preserve the naming order.
  sf_clean <- dplyr::bind_cols(sf_clean, tbl_query[rep(1, nrow(sf_clean)), ])

  # Keep selected columns.
  result_out <- keep_names(
    sf_clean,
    return_addresses,
    full_results,
    colstokeep = names(tbl_query)
  )

  # Restore tibble classes.
  result_out <- sf_to_tbl(result_out)

  result_out
}

skip_if_nominatim_unavailable <- function() {
  testthat::skip_on_cran()
  testthat::skip_if_offline(host = "nominatim.openstreetmap.org")

  if (is.null(.nominatim_check_access$value)) {
    .nominatim_check_access$value <- nominatim_check_access()
  }

  if (.nominatim_check_access$value) {
    return(invisible(TRUE))
  }

  testthat::skip("Nominatim API is not reachable.")

  invisible()
}

skip_nominatim_ci <- function() {
  testthat::skip_on_ci()
  skip_if_nominatim_unavailable()

  invisible()
}

.nominatim_check_access <- new.env(parent = emptyenv())
.nominatim_check_access$value <- NULL

test_fixture <- function(...) {
  testthat::test_path("fixtures", ...)
}

mock_geo_tbl <- function(query, lat = "lat", lon = "lon") {
  out <- dplyr::tibble(
    query = query,
    lat = seq_along(query),
    lon = seq_along(query) * -1,
    address = paste("address", query)
  )
  names(out)[2:3] <- c(lat, lon)
  out
}

mock_reverse_tbl <- function(lat, lon, address = "address") {
  out <- dplyr::tibble(
    address = paste("address", lat, lon),
    lat = lat,
    lon = lon
  )
  names(out)[1] <- address
  out
}

mock_geo_sf <- function(query) {
  sf::st_as_sf(
    dplyr::tibble(
      query = query,
      address = paste("address", query),
      lon = seq_along(query) * -1,
      lat = seq_along(query)
    ),
    coords = c("lon", "lat"),
    crs = 4326
  )
}

mock_reverse_sf <- function(lat, lon, address = "address") {
  out <- sf::st_as_sf(
    dplyr::tibble(address = paste("address", lat, lon), lat = lat, lon = lon),
    coords = c("lon", "lat"),
    crs = 4326,
    remove = FALSE
  )
  names(out)[1] <- address
  out
}

test_that("reverse_geo_lite_sf() rejects invalid coordinate inputs", {
  expect_snapshot(error = TRUE, reverse_geo_lite_sf(0, c(2, 3)))
  expect_snapshot(error = TRUE, reverse_geo_lite_sf("a", "a"))
})

test_that("reverse_geo_lite_sf() reports coordinates clamped to bounds", {
  skip_nominatim_ci()

  expect_snapshot(obj <- reverse_geo_lite_sf(0, 200))
  expect_equal(nrow(obj), 1)
  expect_equal(obj$lon, 180)
  expect_equal(obj$address, NA_character_)
  expect_s3_class(obj, "sf")
  expect_s3_class(obj, "tbl")
  expect_true(sf::st_is_empty(obj))
  expect_identical(sf::st_crs(obj), sf::st_crs(4326))

  expect_snapshot(obj <- reverse_geo_lite_sf(200, 200))
  expect_equal(nrow(obj), 1)
  expect_equal(obj$lat, 90)
  expect_equal(obj$address, NA_character_)
  expect_s3_class(obj, "sf")
  expect_s3_class(obj, "tbl")
  expect_true(sf::st_is_empty(obj))
  expect_identical(sf::st_crs(obj), sf::st_crs(4326))
})


test_that("reverse_geo_lite_sf() returns empty geometry without a match", {
  local_mocked_bindings(api_call = function(...) {
    test_fixture("reverse-error.json")
  })

  expect_snapshot(obj <- reverse_geo_lite_sf(89.999999, 179.9999))

  expect_equal(nrow(obj), 1)
  expect_equal(obj$lat, 89.999999)
  expect_equal(obj$lon, 179.9999)
  expect_s3_class(obj, "tbl")
  expect_s3_class(obj, "sf")
  expect_named(obj, c("address", "lat", "lon", "geometry"))
  expect_equal(obj$address, NA_character_)

  expect_snapshot(
    obj_renamed <- reverse_geo_lite_sf(89.999999, 179.9999, address = "adddata")
  )

  expect_named(obj_renamed, c("adddata", "lat", "lon", "geometry"))

  names(obj_renamed) <- names(obj)

  expect_identical(sf::st_drop_geometry(obj), sf::st_drop_geometry(obj_renamed))
  expect_identical(sf::st_geometry(obj), sf::st_geometry(obj_renamed))
})


test_that("reverse_geo_lite_sf() returns point and polygon geometries", {
  skip_nominatim_ci()

  obj <- reverse_geo_lite_sf(42, 3)
  expect_s3_class(obj, "sf")
  expect_s3_class(obj, "tbl")
  expect_equal(nrow(obj), 1)
  expect_identical(as.character(sf::st_geometry_type(obj)), "POINT")

  # Polygon

  expect_message(
    test <- reverse_geo_lite_sf(
      c(42.34, 89.9999, 39.6777),
      c(-3.474, -179.9999, -4.8383),
      points_only = FALSE,
      custom_query = list(zoom = 5)
    ),
    "No results were found"
  )

  expect_match(
    as.character(sf::st_geometry_type(test)),
    "POLYGON|GEOMETRYCOLLECTION"
  )
  expect_s3_class(test, "sf")
  expect_s3_class(test, "tbl")
  expect_equal(nrow(test), 3)

  expect_identical(sf::st_is_empty(test), c(FALSE, TRUE, FALSE))
})

test_that("reverse_geo_lite_sf() parses a successful GeoJSON response", {
  local_mocked_bindings(api_call = function(...) {
    test_fixture("reverse-one.geojson")
  })

  obj <- reverse_geo_lite_sf(40.4168, -3.7038, full_results = TRUE)

  expect_s3_class(obj, "sf")
  expect_s3_class(obj, "tbl")
  expect_equal(nrow(obj), 1)
  expect_equal(obj$lat, 40.4168)
  expect_equal(obj$lon, -3.7038)
  expect_identical(as.character(sf::st_geometry_type(obj)), "POINT")
  expect_contains(names(obj), c("address.city", "extratags.wikidata"))
})

test_that("reverse_geo_lite_sf() returns sf results and forwards zoom", {
  skip_nominatim_ci()

  obj <- reverse_geo_lite_sf(40.4207414, -3.6687109)
  expect_s3_class(obj, "tbl")
  expect_s3_class(obj, "sf")
  expect_equal(nrow(obj), 1)

  expect_named(obj, c("address", "lat", "lon", "geometry"))

  # Same with different zoom
  obj_zoom <- reverse_geo_lite_sf(
    40.4207414,
    -3.6687109,
    custom_query = list(zoom = 3)
  )

  expect_s3_class(obj_zoom, "tbl")
  expect_s3_class(obj_zoom, "sf")

  expect_equal(nrow(obj_zoom), 1)
  expect_false(identical(obj, obj_zoom))
})

test_that("reverse_geo_lite_sf() handles batches and renamed columns", {
  skip_nominatim_ci()

  # Several coordinates
  sev <- reverse_geo_lite_sf(
    lat = c(40.75728, 55.95335),
    long = c(-73.98586, -3.188375)
  )

  expect_equal(nrow(sev), 2)
  expect_s3_class(sev, "tbl")
  expect_s3_class(sev, "sf")

  # Check opts
  obj <- reverse_geo_lite_sf(40.4207414, -3.6687109, address = "addrs")

  expect_s3_class(obj, "tbl")
  expect_s3_class(obj, "sf")
  expect_equal(nrow(obj), 1)

  expect_named(obj, c("addrs", "lat", "lon", "geometry"))

  # Check opts
  obj <- reverse_geo_lite_sf(
    40.4207414,
    -3.6687109,
    address = "addrs",
    return_coords = FALSE
  )

  expect_s3_class(obj, "tbl")
  expect_s3_class(obj, "sf")
  expect_named(obj, c("addrs", "geometry"))
})

test_that("reverse_geo_lite_sf() controls coordinate and full-result columns", {
  skip_nominatim_ci()

  obj <- reverse_geo_lite_sf(
    40.4207414,
    -3.6687109,
    address = "addrs",
    return_coords = FALSE,
    full_results = TRUE
  )

  expect_s3_class(obj, "tbl")
  expect_s3_class(obj, "sf")
  expect_identical(names(obj)[1:3], c("addrs", "lat", "lon"))
  expect_gt(ncol(obj), 5)

  obj2 <- reverse_geo_lite_sf(
    40.4207414,
    -3.6687109,
    address = "addrs",
    return_coords = TRUE,
    full_results = TRUE
  )

  expect_identical(obj, obj2)
})


test_that("reverse_geo_lite_sf() unnests full results without list columns", {
  skip_nominatim_ci()

  # Several coordinates
  sev <- reverse_geo_lite_sf(
    lat = c(40.75728, 55.95335),
    long = c(-73.98586, -3.188375),
    full_results = TRUE,
    custom_query = list(extratags = TRUE)
  )

  expect_s3_class(sev, "tbl")
  expect_equal(nrow(sev), 2)

  # Classes of all cols

  colclass <- vapply(sf::st_drop_geometry(sev), class, FUN.VALUE = character(1))

  # Not lists
  expect_no_match(colclass, "list", fixed = TRUE)
})

test_that("reverse_geo_lite_sf() deduplicates coordinates and restores order", {
  local_mocked_bindings(reverse_geo_lite_sf_single = function(
    lat_cap,
    long_cap,
    address,
    ...
  ) {
    mock_reverse_sf(lat_cap, long_cap, address)
  })

  # Dupes

  lats <- rep(c(40.75728, 55.95335), 50)
  longs <- rep(c(-73.98586, -3.188375), 50)

  expect_silent(dup <- reverse_geo_lite_sf(lats, longs, progressbar = FALSE))

  expect_s3_class(dup, "sf")
  expect_s3_class(dup, "tbl")
  expect_equal(nrow(dup), 100)

  nms <- unique(as.character(dup$address))
  expect_length(nms, 2)
  expect_equal(as.character(dup$address), rep(nms, 50))

  # Check deduping
  dedup <- dplyr::distinct(dup)

  expect_equal(nrow(dedup), 2)
  expect_equal(as.character(dedup$address), nms)
})

test_that("reverse_geo_lite_sf() shows progress for coordinate batches", {
  local_mocked_bindings(reverse_geo_lite_sf_single = function(
    lat_cap,
    long_cap,
    address,
    ...
  ) {
    mock_reverse_sf(lat_cap, long_cap, address)
  })

  lat <- c(40.75728, 55.95335)
  long <- c(-73.98586, -3.188375)

  # No pbar
  expect_silent(reverse_geo_lite_sf(lat[1], long[1]))
  expect_silent(reverse_geo_lite_sf(lat[1], long[1], progressbar = TRUE))

  # Get a pbar
  expect_output(reverse_geo_lite_sf(lat, long))

  # Not
  expect_silent(reverse_geo_lite_sf(lat, long, progressbar = FALSE))
})
test_that("reverse_geo_lite_sf() returns empty geometry after API failure", {
  local_mocked_bindings(api_call = function(...) FALSE)

  expect_snapshot(
    several <- reverse_geo_lite_sf(40.75728, -73.98, full_results = TRUE)
  )

  expect_all_equal(sf::st_is_empty(several), TRUE)
})

test_that("geo_lite() returns a typed row when no address matches", {
  local_mocked_bindings(api_call = function(...) {
    test_fixture("search-empty.json")
  })

  expect_snapshot(obj <- geo_lite("xbzbzbzoa aiaia"))

  expect_equal(nrow(obj), 1)
  expect_identical(obj$query, "xbzbzbzoa aiaia")
  expect_s3_class(obj, "tbl")
  expect_named(obj, c("query", "lat", "lon"))
  expect_equal(
    vapply(obj, class, FUN.VALUE = character(1)),
    c(query = "character", lat = "numeric", lon = "numeric")
  )
  expect_equal(obj$lat, NA_real_)
  expect_equal(obj$lon, NA_real_)

  expect_snapshot(
    obj_renamed <- geo_lite("xbzbzbzoa aiaia", lat = "lata", long = "longa")
  )

  expect_named(obj_renamed, c("query", "lata", "longa"))

  names(obj_renamed) <- names(obj)

  expect_identical(obj, obj_renamed)
})

test_that("geo_lite() returns a non-spatial tibble", {
  skip_nominatim_ci()

  obj <- geo_lite("Madrid")

  expect_s3_class(obj, "tbl")
  expect_false(inherits(obj, "sf"))
  # this is _not_ a _sf function
})

test_that("geo_lite() parses a successful JSON response", {
  local_mocked_bindings(api_call = function(...) {
    test_fixture("search-one.json")
  })

  obj <- geo_lite("Madrid", full_results = TRUE, return_addresses = FALSE)

  expect_s3_class(obj, "tbl")
  expect_equal(nrow(obj), 1)
  expect_identical(obj$query, "Madrid")
  expect_equal(obj$lat, 40.4168)
  expect_equal(obj$lon, -3.7038)
  expect_contains(names(obj), c("address", "boundingbox"))
  expect_type(obj$boundingbox, "list")
})


test_that("geo_lite() caps limits and controls output columns", {
  skip_nominatim_ci()

  expect_message(
    obj <- geo_lite(c("Madrid", "Barcelona"), limit = 51),
    "at most 50"
  )

  expect_named(obj, c("query", "lat", "lon", "address"))

  obj <- geo_lite(
    "Madrid",
    long = "ong",
    lat = "at",
    full_results = FALSE,
    return_addresses = FALSE
  )
  expect_named(obj, c("query", "at", "ong"))

  obj <- geo_lite(
    "Madrid",
    long = "ong",
    lat = "at",
    full_results = FALSE,
    return_addresses = TRUE
  )

  expect_named(obj, c("query", "at", "ong", "address"))

  obj <- geo_lite(
    "Madrid",
    long = "ong",
    lat = "at",
    full_results = TRUE,
    return_addresses = FALSE
  )

  expect_identical(names(obj)[1:4], c("query", "at", "ong", "address"))
  expect_gt(ncol(obj), 4)
})

test_that("geo_lite() forwards custom query options", {
  skip_nominatim_ci()

  expect_gt(
    nrow(geo_lite(
      "Catedral",
      limit = 10,
      custom_query = list(countrycode = "ES")
    )),
    4
  )

  expect_equal(
    nrow(geo_lite("Madrid", custom_query = list(countrycode = "es"))),
    1
  )

  expect_equal(
    nrow(geo_lite("Madrid", custom_query = list(extratags = TRUE))),
    1
  )
})

test_that("geo_lite() queries unique addresses and restores input order", {
  local_mocked_bindings(geo_lite_single = function(address, lat, long, ...) {
    mock_geo_tbl(address, lat, long)
  })

  # Dupes
  expect_silent(
    dup <- geo_lite(
      rep(c("Pentagon", "Barcelona"), 50),
      limit = 1,
      progressbar = FALSE,
      verbose = FALSE
    )
  )

  expect_equal(nrow(dup), 100)
  expect_equal(as.character(dup$query), rep(c("Pentagon", "Barcelona"), 50))

  # Check deduping
  dedup <- dplyr::distinct(dup)

  expect_equal(nrow(dedup), 2)
  expect_equal(as.character(dedup$query), rep(c("Pentagon", "Barcelona"), 1))
})


test_that("geo_lite() displays progress only for multiple enabled queries", {
  local_mocked_bindings(geo_lite_single = function(address, lat, long, ...) {
    mock_geo_tbl(address, lat, long)
  })

  # No pbar
  expect_silent(geo_lite("Madrid"))
  expect_silent(geo_lite("Madrid", progressbar = TRUE))

  # Get a pbar
  expect_output(geo_lite(c("Madrid", "Barcelona")))

  # Not
  expect_silent(geo_lite(c("Madrid", "Barcelona"), progressbar = FALSE))
})
test_that("geo_lite() returns a typed row when the API is unavailable", {
  local_mocked_bindings(api_call = function(...) FALSE)

  expect_snapshot(several <- geo_lite("Madrid", full_results = TRUE))

  expect_equal(
    several[, 2:3],
    dplyr::tibble(
      lat = rep(NA_real_, nrow(several)),
      lon = rep(NA_real_, nrow(several))
    )
  )
})

test_that("geo_address_lookup_sf() returns empty geometry without a match", {
  local_mocked_bindings(api_call = function(...) test_fixture("empty.geojson"))

  expect_snapshot(obj <- geo_address_lookup_sf(34633854, "N"))

  expect_equal(nrow(obj), 1)
  expect_identical(obj$query, "N34633854")
  expect_s3_class(obj, "sf")
  expect_s3_class(obj, "tbl")
  expect_true(sf::st_is_empty(obj))
  expect_identical(sf::st_crs(obj), sf::st_crs(4326))
})


test_that("geo_address_lookup_sf() returns an sf tibble", {
  skip_nominatim_ci()

  obj <- geo_address_lookup_sf(34633854, "W")
  expect_s3_class(obj, "sf")
  expect_s3_class(obj, "tbl")
})

test_that("geo_address_lookup_sf() parses a successful GeoJSON response", {
  local_mocked_bindings(api_call = function(...) {
    test_fixture("lookup-one.geojson")
  })

  obj <- geo_address_lookup_sf(10, "N", full_results = TRUE)

  expect_s3_class(obj, "sf")
  expect_s3_class(obj, "tbl")
  expect_equal(nrow(obj), 1)
  expect_identical(obj$query, "N10")
  expect_identical(as.character(sf::st_geometry_type(obj)), "POINT")
  expect_contains(names(obj), c("address.city", "extratags.wikidata"))
})

test_that("geo_address_lookup_sf() normalizes OSM IDs before building query", {
  state <- new.env(parent = emptyenv())
  state$urls <- character()

  local_mocked_bindings(api_call = function(url, ...) {
    state$urls <- c(state$urls, url)
    test_fixture("lookup-one.geojson")
  })

  out_string <- geo_address_lookup_sf("10", "N")
  out_decimal <- geo_address_lookup_sf(10.9, "N")
  out_negative <- geo_address_lookup_sf(-10.9, "N")

  expect_identical(out_string$query, "N10")
  expect_identical(out_decimal$query, "N10")
  expect_identical(out_negative$query, "N10")
  expect_length(state$urls, 3)
  expect_equal(
    state$urls,
    rep(
      "https://nominatim.openstreetmap.org/lookup?osm_ids=N10&format=geojson",
      3
    )
  )
})

test_that("geo_address_lookup_sf() adds geometry and custom options to URL", {
  state <- new.env(parent = emptyenv())
  state$url_seen <- NULL

  local_mocked_bindings(api_call = function(url, ...) {
    state$url_seen <- url
    test_fixture("lookup-one.geojson")
  })

  out <- geo_address_lookup_sf(
    10,
    "N",
    full_results = TRUE,
    points_only = FALSE,
    custom_query = list(extratags = TRUE, countrycodes = c("es", "fr"))
  )

  expect_identical(out$query, "N10")
  expect_match(state$url_seen, "lookup\\?osm_ids=N10&format=geojson")
  expect_match(state$url_seen, "polygon_geojson=1", fixed = TRUE)
  expect_match(state$url_seen, "addressdetails=1", fixed = TRUE)
  expect_match(state$url_seen, "extratags=1", fixed = TRUE)
  expect_match(state$url_seen, "countrycodes=es,fr", fixed = TRUE)
})

test_that("geo_address_lookup_sf() warns when some OSM IDs are missing", {
  local_mocked_bindings(api_call = function(...) {
    test_fixture("lookup-one.geojson")
  })

  expect_warning(
    out <- geo_address_lookup_sf(c(10, 999), "N", verbose = TRUE),
    "following OSM IDs: N999"
  )

  expect_identical(out$query, "N10")
})

test_that("geo_address_lookup_sf() controls compact and full output columns", {
  skip_nominatim_ci()

  obj <- geo_address_lookup_sf(34633854, "W")

  expect_equal(ncol(obj), 3)
  expect_gt(ncol(geo_address_lookup_sf(34633854, "W", full_results = TRUE)), 3)
})

test_that("geo_address_lookup_sf() forwards geometry and custom options", {
  skip_nominatim_ci()

  expect_equal(
    nrow(geo_address_lookup_sf(
      34633854,
      "W",
      full_results = TRUE,
      custom_query = list(extratags = TRUE)
    )),
    1
  )
  expect_equal(
    nrow(geo_address_lookup_sf(
      34633854,
      "W",
      points_only = FALSE,
      custom_query = list(countrycode = "us")
    )),
    1
  )
  expect_identical(
    as.character(sf::st_geometry_type(geo_address_lookup_sf(
      34633854,
      "W",
      points_only = TRUE,
      custom_query = list(countrycode = "us")
    ))),
    "POINT"
  )
  expect_identical(
    as.character(sf::st_geometry_type(geo_address_lookup_sf(
      34633854,
      "W",
      points_only = FALSE,
      custom_query = list(countrycode = "us")
    ))),
    "POLYGON"
  )
})


test_that("geo_address_lookup_sf() drops unmatched IDs from multiple lookups", {
  skip_nominatim_ci()

  # Ok
  vector_ids <- c(343921, 240109189)
  vector_type <- c("R", "N")
  several <- geo_address_lookup_sf(vector_ids, vector_type)
  expect_equal(nrow(several), 2)
  expect_identical(names(several)[1], "query")

  expect_identical(as.vector(several$query), paste0(vector_type, vector_ids))

  # KO

  vector_ids <- c(146, 240109189)
  vector_type <- c("J", "N")

  expect_snapshot(
    several <- geo_address_lookup_sf(vector_ids, vector_type, verbose = TRUE)
  )
  expect_equal(nrow(several), 1)
  expect_identical(names(several)[1], "query")

  expect_identical(as.vector(several$query), paste0(vector_type, vector_ids)[2])
})


test_that("geo_address_lookup_sf() returns unique column names", {
  skip_nominatim_ci()

  # Ok
  vector_ids <- c(343921, 240109189)
  vector_type <- c("R", "N")
  several <- geo_address_lookup_sf(vector_ids, vector_type, full_results = TRUE)

  expect_named(several, unique(names(several)))

  # Do I have dups by any chance?
  expect_no_match(names(several), "\\.[0-9]$")
})

test_that("geo_address_lookup_sf() returns empty geometry after API failure", {
  local_mocked_bindings(api_call = function(...) FALSE)

  vector_ids <- c(343921, 240109189)
  vector_type <- c("R", "N")
  expect_snapshot(
    several <- geo_address_lookup_sf(
      vector_ids,
      vector_type,
      full_results = TRUE
    )
  )

  expect_all_equal(sf::st_is_empty(several), TRUE)
})


test_that("geo_address_lookup_sf() preserves normalized long OSM identifiers", {
  skip_nominatim_ci()

  vector_ids <- "9743343761"

  several <- geo_address_lookup_sf(vector_ids)

  # IDs should have the right string
  comp <- unique(gsub("[^0-9]", "", several$query))

  expect_identical(vector_ids, comp)

  # With decimals
  vector_ids2 <- 9743343761.34
  several <- geo_address_lookup_sf(vector_ids2)
  comp <- unique(gsub("[^0-9]", "", several$query))

  expect_identical(vector_ids, comp)

  # With negatives
  vector_ids3 <- -1 * vector_ids2
  several <- geo_address_lookup_sf(vector_ids3)
  comp <- unique(gsub("[^0-9]", "", several$query))

  expect_identical(vector_ids, comp)
})

test_that("geo_amenity_sf() displays progress for multiple enabled queries", {
  local_mocked_bindings(geo_lite_struct_sf = function(amenity, ...) {
    mock_geo_sf(amenity)
  })

  bbox <- c(-73.9894467311, 40.75573629, -73.9830630737, 40.75789245)

  # No pbar
  expect_silent(geo_amenity_sf(bbox, "restaurant"))
  expect_silent(geo_amenity_sf(bbox, "restaurant", progressbar = TRUE))

  # Get a pbar
  expect_output(geo_amenity_sf(bbox, c("pub", "restaurant")))

  # Not
  expect_silent(geo_amenity_sf(
    bbox,
    c("pub", "restaurant"),
    progressbar = FALSE
  ))
})

test_that("geo_amenity_sf() forwards options and deduplicates amenities", {
  bbox <- c(-2, 40, -1, 41)
  state <- new.env(parent = emptyenv())
  state$calls <- list()

  local_mocked_bindings(geo_lite_struct_sf = function(
    amenity,
    limit,
    full_results,
    return_addresses,
    verbose,
    nominatim_server,
    custom_query,
    points_only
  ) {
    state$calls[[length(state$calls) + 1L]] <- list(
      amenity = amenity,
      limit = limit,
      full_results = full_results,
      return_addresses = return_addresses,
      verbose = verbose,
      nominatim_server = nominatim_server,
      custom_query = custom_query,
      points_only = points_only
    )
    mock_geo_sf(amenity)
  })

  out <- geo_amenity_sf(
    bbox,
    c("pub", "restaurant", "pub"),
    limit = 3,
    full_results = TRUE,
    return_addresses = FALSE,
    verbose = TRUE,
    nominatim_server = "https://example.com/nominatim",
    progressbar = FALSE,
    custom_query = list(countrycode = "es"),
    points_only = FALSE
  )

  expect_s3_class(out, "sf")
  expect_named(out, c("query", "address", "geometry"))
  expect_equal(out$query, c("pub", "restaurant"))
  expect_length(state$calls, 2)
  expect_equal(
    vapply(state$calls, `[[`, character(1), "amenity"),
    c("pub", "restaurant")
  )
  expect_equal(state$calls[[1]]$custom_query$viewbox, bbox)
  expect_true(state$calls[[1]]$custom_query$bounded)
  expect_identical(state$calls[[1]]$custom_query$countrycode, "es")
  expect_identical(state$calls[[1]]$limit, 3)
  expect_true(state$calls[[1]]$full_results)
  expect_false(state$calls[[1]]$return_addresses)
  expect_false(state$calls[[1]]$points_only)
})

test_that("geo_amenity_sf() strict mode keeps only rows inside the bbox", {
  local_mocked_bindings(geo_lite_struct_sf = function(amenity, ...) {
    sf::st_as_sf(
      dplyr::tibble(
        q_amenity = amenity,
        address = c("inside", "outside"),
        lon = c(0.5, 2),
        lat = c(0.5, 2)
      ),
      coords = c("lon", "lat"),
      crs = 4326
    )
  })

  out <- geo_amenity_sf(
    c(0, 0, 1, 1),
    "pub",
    strict = TRUE,
    progressbar = FALSE
  )

  expect_s3_class(out, "sf")
  expect_named(out, c("query", "address", "geometry"))
  expect_equal(nrow(out), 1)
  expect_equal(out$query, "pub")
  expect_equal(out$address, "inside")
})

test_that("geo_amenity_sf() caps limits and controls output columns", {
  skip_nominatim_ci()

  expect_message(
    obj <- geo_amenity_sf(
      bbox = c(-1.1446, 41.5022, -0.4854, 41.8795),
      c("pub", "restaurant"),
      limit = 51
    ),
    "at most 50"
  )

  expect_named(obj, c("query", "address", "geometry"))

  obj <- geo_amenity_sf(
    bbox = c(-1.1446, 41.5022, -0.4854, 41.8795),
    "pub",
    full_results = FALSE,
    return_addresses = FALSE
  )
  expect_named(obj, c("query", "geometry"))

  obj <- geo_amenity_sf(
    bbox = c(-1.1446, 41.5022, -0.4854, 41.8795),
    "pub",
    full_results = FALSE,
    return_addresses = TRUE
  )

  expect_named(obj, c("query", "address", "geometry"))

  obj <- geo_amenity_sf(
    bbox = c(-1.1446, 41.5022, -0.4854, 41.8795),
    "pub",
    full_results = TRUE,
    return_addresses = FALSE
  )

  expect_identical(names(obj)[1:2], c("query", "address"))
  expect_gt(ncol(obj), 3)
})

test_that("geo_amenity_sf() forwards custom query options", {
  skip_nominatim_ci()

  expect_gt(
    nrow(geo_amenity_sf(
      bbox = c(-1.1446, 41.5022, -0.4854, 41.8795),
      "pub",
      limit = 10,
      custom_query = list(countrycode = "es")
    )),
    4
  )
  expect_equal(
    nrow(geo_amenity_sf(
      bbox = c(-1.1446, 41.5022, -0.4854, 41.8795),
      "pub",
      custom_query = list(countrycode = "es")
    )),
    1
  )
  expect_equal(
    nrow(geo_amenity_sf(
      bbox = c(-1.1446, 41.5022, -0.4854, 41.8795),
      "pub",
      custom_query = list(extratags = 1)
    )),
    1
  )
})

test_that("geo_amenity_sf() accepts numeric, sfc and sf bounding boxes", {
  skip_nominatim_ci()

  expect_lt(
    nrow(geo_amenity_sf(
      bbox = c(-1.1446, 41.5022, -0.4854, 41.8795),
      "pub",
      limit = 1,
      strict = TRUE
    )),
    2
  )

  bbox_sfc <- bbox_to_poly(c(-1.1446, 41.5022, -0.4854, 41.8795))
  expect_s3_class(bbox_sfc, "sfc")

  expect_silent(
    a <- geo_amenity_sf(bbox = bbox_sfc, "pub", limit = 1, strict = TRUE)
  )
  expect_s3_class(a, "sf")

  bbox_sf <- sf::st_sf(x = 1, bbox_sfc)
  expect_s3_class(bbox_sf, "sf")

  bbox_sf <- sf::st_transform(bbox_sf, 3857)

  expect_silent(
    a <- geo_amenity_sf(bbox = bbox_sf, "pub", limit = 1, strict = TRUE)
  )
  expect_s3_class(a, "sf")
})

test_that("geo_lite_struct() handles missing and unmatched components", {
  local_mocked_bindings(api_call = function(...) {
    test_fixture("search-empty.json")
  })

  expect_snapshot(obj <- geo_lite_struct())

  expect_snapshot(obj <- geo_lite_struct(amenity = "xbzbzbzoa aiaia"))

  expect_equal(nrow(obj), 1)
  expect_identical(obj$q_amenity, "xbzbzbzoa aiaia")
  expect_s3_class(obj, "tbl")
  expect_named(
    obj,
    c(
      "q_amenity",
      "q_street",
      "q_city",
      "q_county",
      "q_state",
      "q_country",
      "q_postalcode",
      "lat",
      "lon"
    )
  )

  expect_equal(obj$lat, NA_real_)
  expect_equal(obj$lon, NA_real_)

  expect_snapshot(
    obj_renamed <- geo_lite_struct(
      "xbzbzbzoa aiaia",
      lat = "lata",
      long = "longa"
    )
  )

  expect_named(
    obj_renamed,
    c(
      "q_amenity",
      "q_street",
      "q_city",
      "q_county",
      "q_state",
      "q_country",
      "q_postalcode",
      "lata",
      "longa"
    )
  )

  names(obj_renamed) <- names(obj)

  expect_identical(obj, obj_renamed)
})

test_that("geo_lite_struct() returns a non-spatial tibble", {
  skip_nominatim_ci()

  obj <- geo_lite_struct(city = "Madrid")

  expect_s3_class(obj, "tbl")
  expect_false(inherits(obj, "sf"))
  # this is _not_ a _sf function
})

test_that("geo_lite_struct() parses a successful JSON response", {
  local_mocked_bindings(api_call = function(...) {
    test_fixture("search-one.json")
  })

  obj <- geo_lite_struct(city = "Madrid", full_results = TRUE)

  expect_s3_class(obj, "tbl")
  expect_equal(nrow(obj), 1)
  expect_identical(obj$q_city, "Madrid")
  expect_equal(obj$lat, 40.4168)
  expect_equal(obj$lon, -3.7038)
  expect_contains(names(obj), c("address", "boundingbox"))
  expect_type(obj$boundingbox, "list")
})


test_that("geo_lite_struct() caps limits and controls output columns", {
  skip_nominatim_ci()

  expect_message(
    obj <- geo_lite_struct(city = c("Madrid", "Barcelona"), limit = 51),
    "at most 50"
  )

  expect_identical(rev(names(obj))[1:3], rev(c("lat", "lon", "address")))

  obj <- geo_lite_struct(
    city = "Madrid",
    long = "ong",
    lat = "at",
    full_results = FALSE,
    return_addresses = FALSE
  )
  expect_identical(rev(names(obj))[1:2], rev(c("at", "ong")))

  obj <- geo_lite_struct(
    city = "Madrid",
    long = "ong",
    lat = "at",
    full_results = FALSE,
    return_addresses = TRUE
  )

  expect_identical(rev(names(obj))[1:3], rev(c("at", "ong", "address")))

  obj <- geo_lite_struct(
    city = "Madrid",
    long = "ong",
    lat = "at",
    full_results = TRUE,
    return_addresses = FALSE
  )

  expect_gt(ncol(obj), 10)
})

test_that("geo_lite_struct() forwards custom query options", {
  skip_nominatim_ci()

  expect_gt(nrow(geo_lite_struct("Catedral", country = "ES", limit = 10)), 4)

  expect_equal(
    nrow(geo_lite_struct("Madrid", custom_query = list(countrycode = "es"))),
    1
  )

  expect_equal(
    nrow(geo_lite_struct("Madrid", custom_query = list(extratags = TRUE))),
    1
  )
})
test_that("geo_lite_struct() returns a typed row when the API is unavailable", {
  local_mocked_bindings(api_call = function(...) FALSE)

  expect_snapshot(several <- geo_lite_struct("Madrid", full_results = TRUE))

  expect_equal(
    several[, c("lat", "lon")],
    dplyr::tibble(
      lat = rep(NA_real_, nrow(several)),
      lon = rep(NA_real_, nrow(several))
    )
  )
})

test_that("geo_address_lookup() returns a typed row when no object matches", {
  local_mocked_bindings(api_call = function(...) {
    test_fixture("search-empty.json")
  })

  expect_snapshot(obj <- geo_address_lookup(34633854, "N"))

  expect_equal(nrow(obj), 1)
  expect_identical(obj$query, "N34633854")
  expect_s3_class(obj, "tbl")
  expect_named(obj, c("query", "lat", "lon"))

  objclass <- vapply(obj, class, FUN.VALUE = character(1))

  expect_equal(
    objclass,
    c(query = "character", lat = "numeric", lon = "numeric")
  )
  expect_equal(obj$lat, NA_real_)
  expect_equal(obj$lon, NA_real_)

  expect_snapshot(
    obj_renamed <- geo_address_lookup(
      34633854,
      "N",
      lat = "lata",
      long = "longa"
    )
  )

  expect_named(obj_renamed, c("query", "lata", "longa"))

  names(obj_renamed) <- names(obj)

  expect_identical(obj, obj_renamed)
})

test_that("geo_address_lookup() returns a non-spatial tibble", {
  skip_nominatim_ci()

  out <- geo_address_lookup(34633854, "W")

  expect_s3_class(out, "tbl")
  expect_false(inherits(out, "sf"))
})

test_that("geo_address_lookup() parses a successful JSON response", {
  local_mocked_bindings(api_call = function(...) {
    test_fixture("lookup-one.json")
  })

  obj <- geo_address_lookup(10, "N", full_results = TRUE)

  expect_s3_class(obj, "tbl")
  expect_equal(nrow(obj), 1)
  expect_identical(obj$query, "N10")
  expect_equal(obj$lat, 40.4168)
  expect_equal(obj$lon, -3.7038)
  expect_contains(names(obj), c("address", "boundingbox"))
  expect_type(obj$boundingbox, "list")
})

test_that("geo_address_lookup() normalizes OSM IDs before building query", {
  state <- new.env(parent = emptyenv())
  state$urls <- character()

  local_mocked_bindings(api_call = function(url, ...) {
    state$urls <- c(state$urls, url)
    test_fixture("lookup-one.json")
  })

  out_string <- geo_address_lookup("10", "N")
  out_decimal <- geo_address_lookup(10.9, "N")
  out_negative <- geo_address_lookup(-10.9, "N")

  expect_identical(out_string$query, "N10")
  expect_identical(out_decimal$query, "N10")
  expect_identical(out_negative$query, "N10")
  expect_length(state$urls, 3)
  expect_equal(
    state$urls,
    rep(
      "https://nominatim.openstreetmap.org/lookup?osm_ids=N10&format=jsonv2",
      3
    )
  )
})

test_that("geo_address_lookup() adds full results and custom options to URL", {
  state <- new.env(parent = emptyenv())
  state$url_seen <- NULL

  local_mocked_bindings(api_call = function(url, ...) {
    state$url_seen <- url
    test_fixture("lookup-one.json")
  })

  out <- geo_address_lookup(
    10,
    "N",
    full_results = TRUE,
    custom_query = list(extratags = TRUE, countrycodes = c("es", "fr"))
  )

  expect_identical(out$query, "N10")
  expect_match(state$url_seen, "lookup\\?osm_ids=N10&format=jsonv2")
  expect_match(state$url_seen, "addressdetails=1", fixed = TRUE)
  expect_match(state$url_seen, "extratags=1", fixed = TRUE)
  expect_match(state$url_seen, "countrycodes=es,fr", fixed = TRUE)
})

test_that("geo_address_lookup() warns when some OSM IDs are missing", {
  local_mocked_bindings(api_call = function(...) {
    test_fixture("lookup-one.json")
  })

  expect_warning(
    out <- geo_address_lookup(c(10, 999), "N", verbose = TRUE),
    "following OSM IDs: N999"
  )

  expect_identical(out$query, "N10")
})

test_that("geo_address_lookup() controls coordinate and address columns", {
  skip_nominatim_ci()

  obj <- geo_address_lookup(32965412, "W")

  expect_named(obj, c("query", "lat", "lon", "address"))

  obj <- geo_address_lookup(
    34633854,
    "W",
    long = "ong",
    lat = "at",
    full_results = FALSE,
    return_addresses = FALSE
  )

  expect_named(obj, c("query", "at", "ong"))

  obj <- geo_address_lookup(
    34633854,
    "W",
    long = "ong",
    lat = "at",
    full_results = FALSE,
    return_addresses = TRUE
  )
  expect_named(obj, c("query", "at", "ong", "address"))

  obj <- geo_address_lookup(
    34633854,
    "W",
    long = "ong",
    lat = "at",
    full_results = TRUE,
    return_addresses = TRUE
  )

  expect_identical(names(obj)[1:4], c("query", "at", "ong", "address"))
  expect_gt(ncol(obj), 4)
})

test_that("geo_address_lookup() forwards custom query options", {
  skip_nominatim_ci()

  expect_equal(
    nrow(geo_address_lookup(
      34633854,
      "W",
      full_results = TRUE,
      custom_query = list(extratags = TRUE)
    )),
    1
  )

  expect_equal(
    nrow(geo_address_lookup(
      34633854,
      "W",
      custom_query = list(countrycode = "us")
    )),
    1
  )
})


test_that("geo_address_lookup() matches multiple IDs and drops missing ones", {
  skip_nominatim_ci()

  # Ok
  vector_ids <- c(343921, 240109189)
  vector_type <- c("R", "N")
  several <- geo_address_lookup(vector_ids, vector_type)
  expect_equal(nrow(several), 2)
  expect_named(several, c("query", "lat", "lon", "address"))

  expect_identical(as.vector(several$query), paste0(vector_type, vector_ids))

  # KO

  vector_ids <- c(146, 240109189)
  vector_type <- c("J", "N")

  expect_snapshot(
    several <- geo_address_lookup(vector_ids, vector_type, verbose = TRUE)
  )

  expect_equal(nrow(several), 1)
  expect_named(several, c("query", "lat", "lon", "address"))

  expect_identical(as.vector(several$query), paste0(vector_type, vector_ids)[2])
})

test_that("geo_address_lookup() returns typed rows after API failure", {
  local_mocked_bindings(api_call = function(...) FALSE)

  vector_ids <- c(343921, 240109189)
  vector_type <- c("R", "N")
  expect_snapshot(
    several <- geo_address_lookup(vector_ids, vector_type, full_results = TRUE)
  )

  expect_equal(
    several[, 2:3],
    dplyr::tibble(
      lat = rep(NA_real_, nrow(several)),
      lon = rep(NA_real_, nrow(several))
    )
  )
})


test_that("geo_address_lookup() preserves normalized long OSM identifiers", {
  skip_nominatim_ci()

  vector_ids <- "9743343761"

  several <- geo_address_lookup(vector_ids)

  # IDs should have the right string
  comp <- unique(gsub("[^0-9]", "", several$query))

  expect_identical(vector_ids, comp)

  # With decimals
  vector_ids2 <- 9743343761.34
  several <- geo_address_lookup(vector_ids2)
  comp <- unique(gsub("[^0-9]", "", several$query))

  expect_identical(vector_ids, comp)

  # With negatives
  vector_ids3 <- -1 * vector_ids2
  several <- geo_address_lookup(vector_ids3)
  comp <- unique(gsub("[^0-9]", "", several$query))

  expect_identical(vector_ids, comp)
})

test_that("reverse_geo_lite() rejects nonnumeric and mismatched coordinates", {
  expect_snapshot(error = TRUE, reverse_geo_lite(0, c(2, 3)))
  expect_snapshot(error = TRUE, reverse_geo_lite("a", "a"))
})

test_that("reverse_geo_lite() reports and clamps coordinates outside bounds", {
  skip_nominatim_ci()

  expect_snapshot(out <- reverse_geo_lite(0, 200))
  chk <- dplyr::tibble(lat = 0, lon = 180)
  expect_identical(out[, c("lat", "lon")], chk)

  expect_snapshot(out <- reverse_geo_lite(200, 0))
  chk <- dplyr::tibble(lat = 90, lon = 0)
  expect_identical(out[, c("lat", "lon")], chk)
})


test_that("reverse_geo_lite() returns a typed row when no address matches", {
  local_mocked_bindings(api_call = function(...) {
    test_fixture("reverse-error.json")
  })

  expect_snapshot(obj <- reverse_geo_lite(89.999999, 179.9999))

  expect_equal(nrow(obj), 1)
  expect_equal(obj$lat, 89.999999)
  expect_equal(obj$lon, 179.9999)
  expect_s3_class(obj, "tbl")
  expect_named(obj, c("address", "lat", "lon"))
  expect_equal(
    vapply(obj, class, FUN.VALUE = character(1)),
    c(address = "character", lat = "numeric", lon = "numeric")
  )
  expect_equal(obj$address, NA_character_)

  expect_snapshot(
    obj_renamed <- reverse_geo_lite(89.999999, 179.9999, address = "adddata")
  )

  expect_named(obj_renamed, c("adddata", "lat", "lon"))

  names(obj_renamed) <- names(obj)

  expect_identical(obj, obj_renamed)
})


test_that("reverse_geo_lite() returns a non-spatial tibble", {
  skip_nominatim_ci()

  obj <- reverse_geo_lite(40.4168, -3.7038)
  expect_s3_class(obj, "tbl")
  expect_false(inherits(obj, "sf"))
  # this is _not_ a _sf function
})

test_that("reverse_geo_lite() parses a successful JSON response", {
  local_mocked_bindings(api_call = function(...) {
    test_fixture("reverse-one.json")
  })

  obj <- reverse_geo_lite(40.4168, -3.7038, full_results = TRUE)

  expect_s3_class(obj, "tbl")
  expect_equal(nrow(obj), 1)
  expect_equal(obj$lat, 40.4168)
  expect_equal(obj$lon, -3.7038)
  expect_contains(names(obj), c("address", "boundingbox"))
  expect_type(obj$boundingbox, "list")
})

test_that("reverse_geo_lite() returns tabular results and forwards zoom", {
  skip_nominatim_ci()

  obj <- reverse_geo_lite(40.4207414, -3.6687109)
  expect_s3_class(obj, "tbl")
  expect_equal(nrow(obj), 1)

  expect_named(obj, c("address", "lat", "lon"))

  # Same with different zoom
  obj_zoom <- reverse_geo_lite(
    40.4207414,
    -3.6687109,
    custom_query = list(zoom = 3)
  )

  expect_s3_class(obj_zoom, "tbl")
  expect_equal(nrow(obj_zoom), 1)
  expect_false(identical(obj, obj_zoom))
})

test_that("reverse_geo_lite() handles multiple coordinates and address names", {
  skip_nominatim_ci()

  # Several coordinates
  sev <- reverse_geo_lite(
    lat = c(40.75728, 55.95335),
    long = c(-73.98586, -3.188375)
  )

  expect_equal(nrow(sev), 2)
  expect_s3_class(sev, "tbl")

  # Check opts
  obj <- reverse_geo_lite(40.4207414, -3.6687109, address = "addrs")

  expect_s3_class(obj, "tbl")
  expect_equal(nrow(obj), 1)

  expect_named(obj, c("addrs", "lat", "lon"))

  # Check opts
  obj <- reverse_geo_lite(
    40.4207414,
    -3.6687109,
    address = "addrs",
    return_coords = FALSE
  )

  expect_s3_class(obj, "tbl")
  expect_named(obj, "addrs")
})

test_that("reverse_geo_lite() controls coordinate and full-result columns", {
  skip_nominatim_ci()

  obj <- reverse_geo_lite(
    40.4207414,
    -3.6687109,
    address = "addrs",
    return_coords = FALSE,
    full_results = TRUE
  )

  expect_s3_class(obj, "tbl")
  expect_identical(names(obj)[1:3], c("addrs", "lat", "lon"))
  expect_gt(ncol(obj), 5)

  obj2 <- reverse_geo_lite(
    40.4207414,
    -3.6687109,
    address = "addrs",
    return_coords = TRUE,
    full_results = TRUE
  )

  expect_identical(obj, obj2)
})


test_that("reverse_geo_lite() unnests bounding boxes in full results", {
  skip_nominatim_ci()

  # Several coordinates
  sev <- reverse_geo_lite(
    lat = c(40.75728, 55.95335),
    long = c(-73.98586, -3.188375),
    full_results = TRUE,
    custom_query = list(extratags = TRUE)
  )

  expect_s3_class(sev, "tbl")
  expect_equal(nrow(sev), 2)

  # Classes of all cols

  colclass <- vapply(sev, class, FUN.VALUE = character(1))
  expect_contains(names(colclass), "boundingbox")
  expect_identical(colclass[["boundingbox"]], "list")

  # Rest of columns not list
  expect_no_match(
    colclass["boundingbox" != names(colclass)],
    "list",
    fixed = TRUE
  )

  # Extract
  bb_l <- sev[["boundingbox"]]
  expect_type(bb_l, "list")
  expect_length(bb_l, 2)

  # Each object
  for (i in seq_along(bb_l)) {
    lc <- bb_l[[i]]
    expect_length(lc, 4)
    expect_type(lc, "double")
  }
})

test_that("reverse_geo_lite() queries unique coordinates and restores order", {
  local_mocked_bindings(reverse_geo_lite_single = function(
    lat_cap,
    long_cap,
    address,
    ...
  ) {
    mock_reverse_tbl(lat_cap, long_cap, address)
  })

  # Dupes

  lats <- rep(c(40.75728, 55.95335), 50)
  longs <- rep(c(-73.98586, -3.188375), 50)

  expect_silent(dup <- reverse_geo_lite(lats, longs, progressbar = FALSE))

  expect_equal(nrow(dup), 100)

  nms <- unique(as.character(dup$address))
  expect_length(nms, 2)
  expect_equal(as.character(dup$address), rep(nms, 50))

  # Check deduping
  dedup <- dplyr::distinct(dup)

  expect_equal(nrow(dedup), 2)
  expect_equal(as.character(dedup$address), nms)
})

test_that("reverse_geo_lite() displays progress for multiple enabled queries", {
  local_mocked_bindings(reverse_geo_lite_single = function(
    lat_cap,
    long_cap,
    address,
    ...
  ) {
    mock_reverse_tbl(lat_cap, long_cap, address)
  })

  lat <- c(40.75728, 55.95335)
  long <- c(-73.98586, -3.188375)

  # No pbar
  expect_silent(reverse_geo_lite(lat[1], long[1]))
  expect_silent(reverse_geo_lite(lat[1], long[1], progressbar = TRUE))

  # Get a pbar
  expect_output(reverse_geo_lite(lat, long))

  # Not
  expect_silent(reverse_geo_lite(lat, long, progressbar = FALSE))
})
test_that("reverse_geo_lite() returns a typed row after API failure", {
  local_mocked_bindings(api_call = function(...) FALSE)

  expect_snapshot(
    several <- reverse_geo_lite(40.75728, -73.98, full_results = TRUE)
  )

  expect_equal(several$address, NA_character_)
})

test_that("geo_lite_struct_sf() handles missing and unmatched components", {
  local_mocked_bindings(api_call = function(...) test_fixture("empty.geojson"))

  expect_snapshot(obj <- geo_lite_struct_sf())
  expect_s3_class(obj, "sf")
  expect_true(sf::st_is_empty(obj))

  expect_snapshot(obj <- geo_lite_struct_sf("xbzbzbzoa aiaia"))
  expect_s3_class(obj, "sf")
  expect_true(sf::st_is_empty(obj))

  expect_equal(nrow(obj), 1)
  expect_identical(obj$q_amenity, "xbzbzbzoa aiaia")
  expect_s3_class(obj, "sf")
  expect_s3_class(obj, "tbl")
  expect_true(sf::st_is_empty(obj))
  expect_identical(sf::st_crs(obj), sf::st_crs(4326))
})


test_that("geo_lite_struct_sf() returns point and polygon geometries", {
  skip_nominatim_ci()

  obj <- geo_lite_struct_sf(city = "Madrid")

  expect_s3_class(obj, "sf")
  expect_s3_class(obj, "tbl")
  expect_equal(nrow(obj), 1)
  expect_identical(as.character(obj$q_city), "Madrid")
  expect_identical(as.character(sf::st_geometry_type(obj)), "POINT")

  # Polygon

  expect_snapshot(
    test <- geo_lite_struct_sf(
      city = "Madrid",
      points_only = FALSE,
      limit = 100
    )
  )

  expect_contains(as.character(sf::st_geometry_type(test)), "POLYGON")
  expect_s3_class(test, "sf")
  expect_s3_class(test, "tbl")
  expect_gt(nrow(test), 2)
  expect_identical(unique(as.character(test$q_city)), "Madrid")
})

test_that("geo_lite_struct_sf() parses a successful GeoJSON response", {
  local_mocked_bindings(api_call = function(...) {
    test_fixture("search-one.geojson")
  })

  obj <- geo_lite_struct_sf(city = "Madrid", full_results = TRUE)

  expect_s3_class(obj, "sf")
  expect_s3_class(obj, "tbl")
  expect_equal(nrow(obj), 1)
  expect_identical(obj$q_city, "Madrid")
  expect_identical(as.character(sf::st_geometry_type(obj)), "POINT")
  expect_contains(names(obj), c("address.city", "extratags.wikidata"))
})

test_that("geo_lite_struct_sf() caps limits and controls output columns", {
  skip_nominatim_ci()

  expect_message(
    obj <- geo_lite_struct_sf(city = c("Madrid", "Barcelona"), limit = 51),
    "at most 50"
  )

  expect_s3_class(obj, "sf")
  expect_s3_class(obj, "tbl")
  expect_identical(rev(names(obj))[1:2], rev(c("address", "geometry")))

  obj <- geo_lite_struct_sf(
    "Madrid",
    full_results = FALSE,
    return_addresses = FALSE
  )

  expect_s3_class(obj, "sf")
  expect_s3_class(obj, "tbl")
  expect_no_match(names(obj), "^address$")

  obj <- geo_lite_struct_sf(
    city = "Madrid",
    full_results = FALSE,
    return_addresses = TRUE
  )

  expect_s3_class(obj, "sf")
  expect_s3_class(obj, "tbl")
  expect_identical(rev(names(obj))[1:2], rev(c("address", "geometry")))

  obj <- geo_lite_struct_sf(
    city = "Madrid",
    full_results = TRUE,
    return_addresses = FALSE
  )
  expect_s3_class(obj, "sf")
  expect_s3_class(obj, "tbl")
  expect_gt(ncol(obj), 4)
})

test_that("geo_lite_struct_sf() forwards custom query options", {
  skip_nominatim_ci()

  expect_gt(
    nrow(geo_lite_struct_sf(
      "restaurant",
      limit = 50,
      custom_query = list(countrycode = "es")
    )),
    4
  )

  expect_equal(
    nrow(geo_lite_struct_sf("school", custom_query = list(countrycode = "es"))),
    1
  )

  expect_equal(
    nrow(geo_lite_struct_sf("hospital", custom_query = list(extratags = TRUE))),
    1
  )
})


test_that("geo_lite_struct_sf() returns empty geometry after API failure", {
  local_mocked_bindings(api_call = function(...) FALSE)

  expect_snapshot(several <- geo_lite_struct_sf("madrid", full_results = TRUE))

  expect_all_equal(sf::st_is_empty(several), TRUE)
})

test_that("geo_amenity() displays progress only for multiple enabled queries", {
  local_mocked_bindings(geo_lite_struct = function(
    amenity,
    lat = "lat",
    long = "lon",
    ...
  ) {
    mock_geo_tbl(amenity, lat, long)
  })

  bbox <- c(-73.9894467311, 40.75573629, -73.9830630737, 40.75789245)

  # No pbar
  expect_silent(geo_amenity(bbox, "restaurant"))
  expect_silent(geo_amenity(bbox, "restaurant", progressbar = TRUE))

  # Get a pbar
  expect_output(geo_amenity(bbox, c("pub", "restaurant")))

  # Not
  expect_silent(geo_amenity(bbox, c("pub", "restaurant"), progressbar = FALSE))
})

test_that("geo_amenity() forwards options and deduplicates amenities", {
  bbox <- c(-2, 40, -1, 41)
  state <- new.env(parent = emptyenv())
  state$calls <- list()

  local_mocked_bindings(geo_lite_struct = function(
    amenity,
    lat,
    long,
    limit,
    full_results,
    return_addresses,
    verbose,
    nominatim_server,
    custom_query
  ) {
    state$calls[[length(state$calls) + 1L]] <- list(
      amenity = amenity,
      lat = lat,
      long = long,
      limit = limit,
      full_results = full_results,
      return_addresses = return_addresses,
      verbose = verbose,
      nominatim_server = nominatim_server,
      custom_query = custom_query
    )
    mock_geo_tbl(amenity, lat, long)
  })

  out <- geo_amenity(
    bbox,
    c("pub", "restaurant", "pub"),
    lat = "latitude",
    long = "longitude",
    limit = 3,
    full_results = TRUE,
    return_addresses = FALSE,
    verbose = TRUE,
    nominatim_server = "https://example.com/nominatim",
    progressbar = FALSE,
    custom_query = list(countrycode = "es")
  )

  expect_named(out, c("query", "latitude", "longitude", "address"))
  expect_equal(out$query, c("pub", "restaurant"))
  expect_length(state$calls, 2)
  expect_equal(
    vapply(state$calls, `[[`, character(1), "amenity"),
    c("pub", "restaurant")
  )
  expect_equal(state$calls[[1]]$custom_query$viewbox, bbox)
  expect_true(state$calls[[1]]$custom_query$bounded)
  expect_identical(state$calls[[1]]$custom_query$countrycode, "es")
  expect_identical(state$calls[[1]]$limit, 3)
  expect_true(state$calls[[1]]$full_results)
  expect_false(state$calls[[1]]$return_addresses)
})

test_that("geo_amenity() strict mode keeps only rows inside the bbox", {
  local_mocked_bindings(geo_lite_struct = function(amenity, lat, long, ...) {
    dplyr::tibble(
      q_amenity = amenity,
      lat = c(0.5, 2),
      lon = c(0.5, 2),
      address = c("inside", "outside")
    )
  })

  out <- geo_amenity(c(0, 0, 1, 1), "pub", strict = TRUE, progressbar = FALSE)

  expect_named(out, c("query", "lat", "lon", "address"))
  expect_equal(nrow(out), 1)
  expect_equal(out$query, "pub")
  expect_equal(out$address, "inside")
  expect_equal(out$lat, 0.5)
  expect_equal(out$lon, 0.5)
})

test_that("geo_amenity() caps limits and controls output columns", {
  skip_nominatim_ci()

  expect_message(
    obj <- geo_amenity(
      bbox = c(-1.1446, 41.5022, -0.4854, 41.8795),
      c("pub", "restaurant"),
      limit = 51
    ),
    "at most 50"
  )

  expect_named(obj, c("query", "lat", "lon", "address"))

  obj <- geo_amenity(
    bbox = c(-1.1446, 41.5022, -0.4854, 41.8795),
    "pub",
    long = "ong",
    lat = "at",
    full_results = FALSE,
    return_addresses = FALSE
  )
  expect_named(obj, c("query", "at", "ong"))

  obj <- geo_amenity(
    bbox = c(-1.1446, 41.5022, -0.4854, 41.8795),
    "pub",
    long = "ong",
    lat = "at",
    full_results = FALSE,
    return_addresses = TRUE
  )

  expect_named(obj, c("query", "at", "ong", "address"))

  obj <- geo_amenity(
    bbox = c(-1.1446, 41.5022, -0.4854, 41.8795),
    "pub",
    long = "ong",
    lat = "at",
    full_results = TRUE,
    return_addresses = FALSE
  )

  expect_identical(names(obj)[1:4], c("query", "at", "ong", "address"))
  expect_gt(ncol(obj), 4)
})

test_that("geo_amenity() forwards custom query options", {
  skip_nominatim_ci()

  expect_gt(
    nrow(geo_amenity(
      bbox = c(-1.1446, 41.5022, -0.4854, 41.8795),
      "pub",
      limit = 10,
      custom_query = list(countrycode = "es")
    )),
    4
  )
  expect_equal(
    nrow(geo_amenity(
      bbox = c(-1.1446, 41.5022, -0.4854, 41.8795),
      "pub",
      custom_query = list(countrycode = "es")
    )),
    1
  )
  expect_equal(
    nrow(geo_amenity(
      bbox = c(-1.1446, 41.5022, -0.4854, 41.8795),
      "pub",
      custom_query = list(extratags = 1)
    )),
    1
  )
})

test_that("geo_amenity() applies strict mode to numeric, sfc and sf bboxes", {
  skip_nominatim_ci()

  expect_lt(
    nrow(geo_amenity(
      bbox = c(-1.1446, 41.5022, -0.4854, 41.8795),
      "pub",
      limit = 1,
      strict = TRUE
    )),
    2
  )

  bbox_sfc <- bbox_to_poly(c(-1.1446, 41.5022, -0.4854, 41.8795))
  expect_s3_class(bbox_sfc, "sfc")

  expect_silent(
    a <- geo_amenity(bbox = bbox_sfc, "pub", limit = 1, strict = TRUE)
  )
  expect_s3_class(a, "tbl")

  bbox_sf <- sf::st_sf(x = 1, bbox_sfc)
  expect_s3_class(bbox_sf, "sf")

  bbox_sf <- sf::st_transform(bbox_sf, 3857)

  expect_silent(
    a <- geo_amenity(bbox = bbox_sf, "pub", limit = 1, strict = TRUE)
  )
  expect_s3_class(a, "tbl")
})

test_that("geo_lite_sf() returns an empty geometry when no address matches", {
  local_mocked_bindings(api_call = function(...) test_fixture("empty.geojson"))

  expect_snapshot(obj <- geo_lite_sf("xbzbzbzoa aiaia"))

  expect_equal(nrow(obj), 1)
  expect_identical(obj$query, "xbzbzbzoa aiaia")
  expect_s3_class(obj, "sf")
  expect_s3_class(obj, "tbl")
  expect_true(sf::st_is_empty(obj))
  expect_identical(sf::st_crs(obj), sf::st_crs(4326))
})


test_that("geo_lite_sf() returns point and polygon geometries", {
  skip_nominatim_ci()

  obj <- geo_lite_sf(c("Madrid", "Barcelona"))

  expect_s3_class(obj, "sf")
  expect_s3_class(obj, "tbl")
  expect_equal(nrow(obj), 2)
  expect_identical(as.character(obj$query), c("Madrid", "Barcelona"))
  expect_equal(as.character(sf::st_geometry_type(obj)), c("POINT", "POINT"))

  # Polygon

  expect_message(
    test <- geo_lite_sf(
      c("Madrid", "ga hann xx kaa pa", "Barcelona"),
      points_only = FALSE
    ),
    "No results were found for the query"
  )

  expect_contains(as.character(sf::st_geometry_type(test)), "POLYGON")
  expect_s3_class(test, "sf")
  expect_s3_class(test, "tbl")
  expect_equal(nrow(test), 3)
  expect_identical(
    as.character(test$query),
    c("Madrid", "ga hann xx kaa pa", "Barcelona")
  )
  expect_identical(sf::st_is_empty(test), c(FALSE, TRUE, FALSE))
})

test_that("geo_lite_sf() parses a successful GeoJSON response", {
  local_mocked_bindings(api_call = function(...) {
    test_fixture("search-one.geojson")
  })

  obj <- geo_lite_sf("Madrid", full_results = TRUE, return_addresses = FALSE)

  expect_s3_class(obj, "sf")
  expect_s3_class(obj, "tbl")
  expect_equal(nrow(obj), 1)
  expect_identical(obj$query, "Madrid")
  expect_identical(as.character(sf::st_geometry_type(obj)), "POINT")
  expect_contains(names(obj), c("address.city", "extratags.wikidata"))
})

test_that("geo_lite_sf() caps limits and controls output columns", {
  skip_nominatim_ci()

  expect_message(
    obj <- geo_lite_sf(c("Madrid", "Barcelona"), limit = 51),
    "at most 50"
  )

  expect_s3_class(obj, "sf")
  expect_s3_class(obj, "tbl")
  expect_named(obj, c("query", "address", "geometry"))

  obj <- geo_lite_sf("Madrid", full_results = FALSE, return_addresses = FALSE)

  expect_s3_class(obj, "sf")
  expect_s3_class(obj, "tbl")
  expect_named(obj, c("query", "geometry"))

  obj <- geo_lite_sf("Madrid", full_results = FALSE, return_addresses = TRUE)

  expect_s3_class(obj, "sf")
  expect_s3_class(obj, "tbl")
  expect_named(obj, c("query", "address", "geometry"))

  obj <- geo_lite_sf("Madrid", full_results = TRUE, return_addresses = FALSE)
  expect_s3_class(obj, "sf")
  expect_s3_class(obj, "tbl")
  expect_identical(names(obj)[1:2], c("query", "address"))
  expect_gt(ncol(obj), 4)
})

test_that("geo_lite_sf() forwards custom query options", {
  skip_nominatim_ci()

  expect_gt(
    nrow(geo_lite_sf(
      "Catedral",
      limit = 10,
      custom_query = list(countrycode = "es")
    )),
    4
  )

  expect_equal(
    nrow(geo_lite_sf("Madrid", custom_query = list(countrycode = "es"))),
    1
  )

  expect_equal(
    nrow(geo_lite_sf("Madrid", custom_query = list(extratags = TRUE))),
    1
  )
})

test_that("geo_lite_sf() queries unique addresses and restores input order", {
  local_mocked_bindings(geo_lite_sf_single = function(address, ...) {
    mock_geo_sf(address)
  })

  # Dupes
  dup <- geo_lite_sf(rep(c("Madrid", "Barcelona"), 50), limit = 1)

  expect_s3_class(dup, "sf")
  expect_s3_class(dup, "tbl")

  expect_equal(nrow(dup), 100)
  expect_equal(as.character(dup$query), rep(c("Madrid", "Barcelona"), 50))

  # Check deduping
  dedup <- dplyr::distinct(dup)

  expect_equal(nrow(dedup), 2)
  expect_equal(as.character(dedup$query), rep(c("Madrid", "Barcelona"), 1))
})

test_that("geo_lite_sf() returns unique column names", {
  skip_nominatim_ci()

  # Ok
  several <- geo_lite_sf(
    c("Murcia", "Segovia"),
    limit = 20,
    full_results = TRUE
  )

  expect_named(several, unique(names(several)))

  # Do I have dups by any chance?
  expect_no_match(names(several), "\\.[0-9]$")
})

test_that("geo_lite_sf() displays progress only for multiple enabled queries", {
  local_mocked_bindings(geo_lite_sf_single = function(address, ...) {
    mock_geo_sf(address)
  })

  # No pbar
  expect_silent(geo_lite_sf("Madrid"))
  expect_silent(geo_lite_sf("Madrid", progressbar = TRUE))

  # Get a pbar
  expect_output(geo_lite_sf(c("Madrid", "Barcelona")))

  # Not
  expect_silent(geo_lite_sf(c("Madrid", "Barcelona"), progressbar = FALSE))
})
test_that("geo_lite_sf() returns empty geometry when the API is unavailable", {
  local_mocked_bindings(api_call = function(...) FALSE)

  expect_snapshot(several <- geo_lite_sf("madrid", full_results = TRUE))

  expect_all_equal(sf::st_is_empty(several), TRUE)
})

#' Reverse geocode coordinates
#'
#' @description
#' Reverse geocodes latitude and longitude coordinates and returns matching
#' results as a [tibble][dplyr::tibble]. Latitude values must be in
#' \eqn{\left[-90, 90 \right]} and longitudes in
#' \eqn{\left[-180, 180 \right]}. Use [reverse_geo_lite_sf()] to return an
#' [`sf`][sf::st_sf] object instead.
#'
#' @details
#' See <https://nominatim.org/release-docs/latest/api/Reverse/> for additional
#' parameters to pass to `custom_query`.
#'
#' @param lat A numeric vector of latitude values in the range
#'   \eqn{\left[-90, 90 \right]}.
#' @param long A numeric vector of longitude values in the range
#'   \eqn{\left[-180, 180 \right]}.
#' @param address A character string specifying the name of the address column
#'   in the output. Defaults to `"address"`.
#' @param return_coords A logical value indicating whether to return the input
#'   coordinates with the results.
#' @param custom_query A named list of API-specific parameters, for example
#'   `list(zoom = 3)`. See **Details**.
#' @inheritParams geo_lite
#'
#' @inherit geo_lite return
#'
#' @section About zooming:
#'
#' Set `custom_query = list(zoom = 3)` to adjust the output. Selected zoom
#' levels correspond to these address details:
#'
#' ```{r, echo=FALSE}
#'
#' t <- dplyr::tribble(
#'  ~zoom, ~address_detail,
#'  "`3`", "country",
#'  "`5`", "state",
#'  "`8`", "county",
#'  "`10`", "city",
#'  "`14`", "suburb",
#'  "`16`", "major streets",
#'  "`17`", "major and minor streets",
#'  "`18`", "building"
#'  )
#'
#' knitr::kable(t, col.names = paste0("**", names(t), "**"))
#'
#' ```
#'
#' @seealso
#' [geo_lite()] for geocoding addresses.
#'
#' @family reverse
#'
#' @export
#' @encoding UTF-8
#'
#' @examplesIf nominatim_check_access()
#' \donttest{
#' reverse_geo_lite(lat = 40.75728, long = -73.98586)
#'
#' # Reverse geocode multiple coordinate pairs.
#' reverse_geo_lite(lat = c(40.75728, 55.95335), long = c(-73.98586, -3.188375))
#'
#' # Set the zoom to the country level.
#' sev <- reverse_geo_lite(
#'   lat = c(40.75728, 55.95335), long = c(-73.98586, -3.188375),
#'   custom_query = list(zoom = 0, extratags = TRUE),
#'   verbose = TRUE, full_results = TRUE
#' )
#'
#' dplyr::glimpse(sev)
#' }
reverse_geo_lite <- function(
  lat,
  long,
  address = "address",
  full_results = FALSE,
  return_coords = TRUE,
  verbose = FALSE,
  nominatim_server = "https://nominatim.openstreetmap.org/",
  progressbar = TRUE,
  custom_query = list()
) {
  keys <- reverse_query_keys(lat, long)
  all_res <- run_reverse_queries(
    keys$unique,
    progressbar,
    function(lat_cap, long_cap) {
      reverse_geo_lite_single(
        lat_cap = lat_cap,
        long_cap = long_cap,
        address = address,
        full_results = full_results,
        return_coords = return_coords,
        verbose = verbose,
        custom_query = custom_query,
        nominatim_server = nominatim_server
      )
    }
  )

  all_res <- dplyr::left_join(
    keys$init[, c(1, 2)],
    all_res,
    by = c("lat_key_int", "long_key_int")
  )

  # Remove internal join keys.
  all_res <- all_res[, -c(1, 2)]
  all_res
}

#' @inheritParams reverse_geo_lite
#' @noRd
reverse_geo_lite_single <- function(
  lat_cap,
  long_cap,
  address = "address",
  full_results = FALSE,
  return_coords = TRUE,
  verbose = TRUE,
  nominatim_server = "https://nominatim.openstreetmap.org/",
  custom_query = list()
) {
  url <- build_reverse_url(
    nominatim_server = nominatim_server,
    lat = lat_cap,
    long = long_cap,
    format = "jsonv2",
    full_results = full_results,
    custom_query = custom_query
  )

  # Download the API response.
  json <- api_call(url, ".json", isFALSE(verbose))

  tbl_query <- dplyr::tibble(lat = lat_cap, lon = long_cap)

  if (isFALSE(json)) {
    message_api_unavailable(url)
    out <- empty_tbl_rev(tbl_query, address)
    return(invisible(out))
  }

  result_init <- jsonlite::fromJSON(json, flatten = TRUE)

  # Handle empty queries.
  if ("error" %in% names(result_init)) {
    message_no_results(paste0("lat = ", lat_cap, ", long = ", long_cap))
    out <- empty_tbl_rev(tbl_query, address)
    return(invisible(out))
  }

  # Unnest nested fields.
  result <- unnest_reverse(result_init)

  result$lat <- as.double(result$lat)
  result$lon <- as.double(result$lon)

  # Keep selected columns.
  result_out <- keep_names_rev(
    result,
    address = address,
    return_coords = return_coords,
    full_results = full_results
  )

  result_out
}

test_that("osm_amenities has the documented tabular schema", {
  expect_s3_class(osm_amenities, c("tbl_df", "data.frame"))
  expect_named(osm_amenities, c("category", "amenity", "comment"))
  expect_type(osm_amenities$category, "character")
  expect_type(osm_amenities$amenity, "character")
  expect_type(osm_amenities$comment, "character")
})

test_that("osm_amenities contains non-missing unique amenity keys", {
  expect_gt(nrow(osm_amenities), 0)
  expect_false(anyNA(osm_amenities$category))
  expect_false(anyNA(osm_amenities$amenity))
  expect_identical(anyDuplicated(osm_amenities$amenity), 0L)
  expect_no_match(osm_amenities$category, "^\\s*$")
  expect_no_match(osm_amenities$amenity, "^\\s*$")
})

---
title: "Get started with nominatimlite"
description: Quick examples showing what **nominatimlite** can do for you.
vignette: >
  %\VignetteIndexEntry{Get started with nominatimlite}
  %\VignetteEngine{quarto::html}
  %\VignetteEncoding{UTF-8}
bibliography: REFERENCES.bib
link-citations: true
tbl-cap-location: bottom
---

<!-- nominatimlite.qmd is generated from nominatimlite.qmd.orig. Please edit that file -->



The goal of **nominatimlite** is to provide a lightweight interface for
geocoding addresses with the [Nominatim
API](https://nominatim.org/release-docs/latest/). It also allows you to return
results as `sf` objects using the **sf** package.

The full site with examples and vignettes is available at
<https://dieghernan.github.io/nominatimlite/>

## What is Nominatim?

**Nominatim** is a tool for searching
[OpenStreetMap](https://www.openstreetmap.org/) data by name and address
([geocoding](https://wiki.openstreetmap.org/wiki/Geocoding "Geocoding")) and
generating synthetic addresses for OSM points (reverse geocoding).

## Why nominatimlite?

**nominatimlite** accesses the Nominatim API without depending on **curl**. In
some situations, **curl** may not be available or accessible, so
**nominatimlite** uses base R functions instead.

## Recommended packages

Other packages are more complete and mature than **nominatimlite** and provide
similar features:

- [**tidygeocoder**](https://jessecambon.github.io/tidygeocoder/)
  [@R-tidygeocoder]: Provides an interface to geocoding services such as
  Nominatim, Google, TomTom and Mapbox.
- [**osmdata**](https://docs.ropensci.org/osmdata/) [@R-osmdata]: Downloads
  spatial data from OpenStreetMap with the [Overpass
  API](https://wiki.openstreetmap.org/wiki/Overpass_API).
- [**arcgeocoder**](https://dieghernan.github.io/arcgeocoder/) [@R-arcgeocoder]:
  Provides a lightweight interface for geocoding with the ArcGIS REST API
  service.

## Usage

### `sf` objects

With **nominatimlite** you can return `sf` objects:


``` r
library(nominatimlite)

# Search for Pizza Hut locations in California.

CA <- geo_lite_sf("California", points_only = FALSE)

pizzahut <- geo_lite_sf(
  "Pizza Hut, California",
  limit = 50,
  custom_query = list(countrycodes = "us")
)

library(ggplot2)

ggplot(CA) +
  geom_sf() +
  geom_sf(data = pizzahut, col = "red")
```

::: {#fig-phut}
![](../man/figures/README-pizzahut-1.png){width="100%"}

Locations of Pizza Hut in California
:::

You can also return polygon and line objects when the Nominatim API provides
them, using the option `points_only = FALSE`:


``` r
sol_poly <- geo_lite_sf("Statue of Liberty, NY, USA", points_only = FALSE)

ggplot(sol_poly) +
  geom_sf()
```

::: {#fig-sol}
![](../man/figures/README-statue_liberty-1.png){width="100%"}

Statue of Liberty
:::

### Geocoding and reverse geocoding

*Note: examples are adapted from the **tidygeocoder** package.*

In this first example, we geocode a few addresses with `geo_lite()`:


``` r
library(tibble)

# Create a data frame with addresses.
some_addresses <- tribble(
  ~name, ~addr,
  "White House", "1600 Pennsylvania Ave NW, Washington, DC",
  "Transamerica Pyramid", "600 Montgomery St, San Francisco, CA 94111",
  "Willis Tower", "233 S Wacker Dr, Chicago, IL 60606"
)

# Geocode the addresses.
lat_longs <- geo_lite(
  some_addresses$addr,
  lat = "latitude",
  long = "longitude",
  progressbar = FALSE
)
```

This example returns only latitude, longitude and address columns from the
Nominatim API. Use `full_results = TRUE` to return all available data from the
Nominatim API.

::: {#tbl-geo}


|query                                      | latitude|  longitude|address                                                                                                                                       |
|:------------------------------------------|--------:|----------:|:---------------------------------------------------------------------------------------------------------------------------------------------|
|1600 Pennsylvania Ave NW, Washington, DC   | 38.89764|  -77.03655|White House, 1600, Pennsylvania Avenue Northwest, Downtown, Ward 2, Washington, District of Columbia, 20500, United States                    |
|600 Montgomery St, San Francisco, CA 94111 | 37.79519| -122.40279|Transamerica Pyramid, 600, Montgomery Street, Financial District, South of Market, San Francisco, California, 94111, United States            |
|233 S Wacker Dr, Chicago, IL 60606         | 41.87874|  -87.63596|Willis Tower, 233, South Wacker Drive, Financial District, Loop, Chicago, South Chicago Township, Cook County, Illinois, 60606, United States |



Example: geocoding addresses
:::

To perform reverse geocoding, use `reverse_geo_lite()` to obtain addresses from
geographic coordinates. The arguments are similar to `geo_lite()`, but now we
provide coordinate values with the `lat` and `long` arguments. The dataset used
here is from the geocoding query above. The single-line address is returned in a
column named with the `address` argument.


``` r
reverse <- reverse_geo_lite(
  lat = lat_longs$latitude,
  long = lat_longs$longitude,
  address = "address_found",
  progressbar = FALSE
)
```

::: {#tbl-rev}


|address_found                                                                                                             |      lat|        lon|
|:-------------------------------------------------------------------------------------------------------------------------|--------:|----------:|
|White House, 1600, Pennsylvania Avenue Northwest, Ward 2, Washington, District of Columbia, 20500, United States          | 38.89764|  -77.03655|
|Sky Bar, Mark Twain Place, Financial District, South of Market, San Francisco, California, 94111, United States           | 37.79519| -122.40254|
|West Adams Street, Financial District, Loop, Chicago, South Chicago Township, Cook County, Illinois, 60606, United States | 41.87874|  -87.63589|



Example: reverse geocoding addresses
:::

For more advanced users, see the [Nominatim
documentation](https://nominatim.org/release-docs/latest/api/Search/) for the
available parameters.

## References

---
title: "Get started with nominatimlite"
description: Search for addresses and work with Nominatim API results.
vignette: >
  %\VignetteIndexEntry{Get started with nominatimlite}
  %\VignetteEngine{quarto::html}
  %\VignetteEncoding{UTF-8}
bibliography: REFERENCES.bib
link-citations: true
tbl-cap-location: bottom
---

<!-- nominatimlite.qmd is generated from nominatimlite.qmd.orig. Please edit that file -->



**nominatimlite** provides a lightweight interface to the [**Nominatim
API**](https://nominatim.org/release-docs/latest/). It supports free-form and
structured address searches, reverse geocoding, amenity lookup and address
lookup by **OpenStreetMap** (OSM) object identifier. Results are returned as
tibbles or **sf** objects.

The full site with examples and vignettes is available at
<https://dieghernan.github.io/nominatimlite/>.

## What is Nominatim?

**Nominatim** searches [**OpenStreetMap**](https://www.openstreetmap.org/) data by
name and address
([geocoding](https://wiki.openstreetmap.org/wiki/Geocoding "Geocoding")) and
reverse geocodes geographic coordinates.

## Why nominatimlite?

**nominatimlite** accesses the **Nominatim API** without requiring the
**curl** package. This makes the package useful in environments where
**curl** is not available. API requests use base **R** functions instead.

## Recommended packages

Related packages provide broader interfaces to geocoding services and
**OpenStreetMap** data:

- [**tidygeocoder**](https://jessecambon.github.io/tidygeocoder/)
  [@R-tidygeocoder] provides an interface to geocoding services such as
  **Nominatim**, **Google**, **TomTom** and **Mapbox**.
- [**osmdata**](https://docs.ropensci.org/osmdata/) [@R-osmdata] downloads
  spatial data from **OpenStreetMap** with the [**Overpass**
  API](https://wiki.openstreetmap.org/wiki/Overpass_API).
- [**arcgeocoder**](https://dieghernan.github.io/arcgeocoder/) [@R-arcgeocoder]
  provides a lightweight interface for geocoding with the **ArcGIS REST API**
  service.

## Usage

### sf output

Use functions with the `_sf` suffix to return results as **sf** objects:


``` r
library(nominatimlite)

# Search for Pizza Hut locations in California.

CA <- geo_lite_sf("California", points_only = FALSE)

pizzahut <- geo_lite_sf(
  "Pizza Hut, California",
  limit = 50,
  custom_query = list(countrycodes = "us")
)

library(ggplot2)

ggplot(CA) +
  geom_sf() +
  geom_sf(data = pizzahut, col = "red")
```

::: {#fig-phut}
![Map of Pizza Hut locations in
California.](../man/figures/README-pizzahut-1.png){width="100%"}

Pizza Hut locations in California.
:::

Set `points_only = FALSE` to return polygon and line geometries when they are
available from the **Nominatim API**:


``` r
sol_poly <- geo_lite_sf("Statue of Liberty, NY, USA", points_only = FALSE)

ggplot(sol_poly) +
  geom_sf()
```

::: {#fig-sol}
![Map of the Statue of Liberty
geometry.](../man/figures/README-statue_liberty-1.png){width="100%"}

Statue of Liberty.
:::

### Address search and reverse geocoding

*The examples in this section are adapted from the **tidygeocoder** package.*

Use `geo_lite()` to search for addresses with free-form queries:


``` r
# Create a data frame with addresses.
some_addresses <- dplyr::tribble(
  ~name, ~addr,
  "White House", "1600 Pennsylvania Ave NW, Washington, DC",
  "Transamerica Pyramid", "600 Montgomery St, San Francisco, CA 94111",
  "Willis Tower", "233 S Wacker Dr, Chicago, IL 60606"
)

# Geocode the addresses.
lat_longs <- geo_lite(
  some_addresses$addr,
  lat = "latitude",
  long = "longitude",
  progressbar = FALSE
)
```

By default, `geo_lite()` returns the query, latitude, longitude and address
columns. Set `full_results = TRUE` to return all available fields from the
**Nominatim API**.

::: {#tbl-geo}


|query                                      | latitude|  longitude|address                                                                                                                                       |
|:------------------------------------------|--------:|----------:|:---------------------------------------------------------------------------------------------------------------------------------------------|
|1600 Pennsylvania Ave NW, Washington, DC   | 38.89764|  -77.03655|White House, 1600, Pennsylvania Avenue Northwest, Ward 2, Washington, District of Columbia, 20500, United States                              |
|600 Montgomery St, San Francisco, CA 94111 | 37.79519| -122.40279|Transamerica Pyramid, 600, Montgomery Street, Financial District, South of Market, San Francisco, California, 94111, United States            |
|233 S Wacker Dr, Chicago, IL 60606         | 41.87874|  -87.63596|Willis Tower, 233, South Wacker Drive, Financial District, Loop, Chicago, South Chicago Township, Cook County, Illinois, 60606, United States |



Geocoded addresses.
:::

Use `reverse_geo_lite()` to reverse geocode latitude and longitude coordinates.
The `lat` and `long` arguments use the results from the address search above.
The `address` argument specifies the name of the output column that contains
each single-line address.


``` r
reverse <- reverse_geo_lite(
  lat = lat_longs$latitude,
  long = lat_longs$longitude,
  address = "address_found",
  progressbar = FALSE
)
```

::: {#tbl-rev}


|address_found                                                                                                                   |      lat|        lon|
|:-------------------------------------------------------------------------------------------------------------------------------|--------:|----------:|
|White House, 1600, Pennsylvania Avenue Northwest, Ward 2, Washington, District of Columbia, 20500, United States                | 38.89764|  -77.03655|
|Sky Bar, 600, Montgomery Street, Financial District, South of Market, San Francisco, California, 94111, United States           | 37.79519| -122.40254|
|233, South Wacker Drive, Financial District, Loop, Chicago, South Chicago Township, Cook County, Illinois, 60606, United States | 41.87874|  -87.63589|



Reverse-geocoded addresses.
:::

See the [**Nominatim search API**
documentation](https://nominatim.org/release-docs/latest/api/Search/) for
additional parameters to pass to `custom_query`.

## References

# api_call() rejects unsupported cache formats before downloading

    Code
      api_call("https://example.com", ext = ".txt", quiet = TRUE)
    Condition
      Error in `match.arg()`:
      ! 'arg' should be one of ".json", ".geojson"

# geo_lite_struct_sf() handles missing and unmatched components

    Code
      obj <- geo_lite_struct_sf()
    Message
      No search parameters were provided.

---

    Code
      obj <- geo_lite_struct_sf("xbzbzbzoa aiaia")
    Message
      No results were found for the query.

# geo_lite_struct_sf() returns point and polygon geometries

    Code
      test <- geo_lite_struct_sf(city = "Madrid", points_only = FALSE, limit = 100)
    Message
      Nominatim returns at most 50 results per query. `limit` has been set to 50.

# geo_lite_struct_sf() returns empty geometry after API failure

    Code
      several <- geo_lite_struct_sf("madrid", full_results = TRUE)
    Message
      Could not reach the API endpoint: https://nominatim.openstreetmap.org/search?format=geojson&limit=1&addressdetails=1&amenity=madrid.

# cap helpers validate and report adjusted values

    Code
      cap_coordinates("a", 1)
    Condition
      Error in `cap_coordinates()`:
      ! `lat` and `long` must be numeric.

---

    Code
      cap_coordinates(1, c(1, 2))
    Condition
      Error in `cap_coordinates()`:
      ! `lat` and `long` must have the same length.

---

    Code
      cap_coordinates(NA_real_, 1)
    Condition
      Error in `cap_coordinates()`:
      ! `lat` and `long` must not contain missing values.

---

    Code
      cap_coordinates(1, NA_real_)
    Condition
      Error in `cap_coordinates()`:
      ! `lat` and `long` must not contain missing values.

---

    Code
      coords <- cap_coordinates(200, -200)
    Message
      Latitude values outside [-90, 90] were clamped to that range.
      Longitude values outside [-180, 180] were clamped to that range.

# geo_lite_sf() returns an empty geometry when no address matches

    Code
      obj <- geo_lite_sf("xbzbzbzoa aiaia")
    Message
      No results were found for the query: xbzbzbzoa aiaia.

# geo_lite_sf() returns empty geometry when the API is unavailable

    Code
      several <- geo_lite_sf("madrid", full_results = TRUE)
    Message
      Could not reach the API endpoint: https://nominatim.openstreetmap.org/search?q=madrid&format=geojson&limit=1&addressdetails=1.

# geo_lite_struct() handles missing and unmatched components

    Code
      obj <- geo_lite_struct()
    Message
      No search parameters were provided.

---

    Code
      obj <- geo_lite_struct(amenity = "xbzbzbzoa aiaia")
    Message
      No results were found for the query.

---

    Code
      obj_renamed <- geo_lite_struct("xbzbzbzoa aiaia", lat = "lata", long = "longa")
    Message
      No results were found for the query.

# geo_lite_struct() returns a typed row when the API is unavailable

    Code
      several <- geo_lite_struct("Madrid", full_results = TRUE)
    Message
      Could not reach the API endpoint: https://nominatim.openstreetmap.org/search?format=jsonv2&limit=1&addressdetails=1&amenity=Madrid.

# reverse_geo_lite_sf() rejects invalid coordinate inputs

    Code
      reverse_geo_lite_sf(0, c(2, 3))
    Condition
      Error in `cap_coordinates()`:
      ! `lat` and `long` must have the same length.

---

    Code
      reverse_geo_lite_sf("a", "a")
    Condition
      Error in `cap_coordinates()`:
      ! `lat` and `long` must be numeric.

# reverse_geo_lite_sf() reports coordinates clamped to bounds

    Code
      obj <- reverse_geo_lite_sf(0, 200)
    Message
      Longitude values outside [-180, 180] were clamped to that range.
      No results were found for the query: lat = 0, long = 180.

---

    Code
      obj <- reverse_geo_lite_sf(200, 200)
    Message
      Latitude values outside [-90, 90] were clamped to that range.
      Longitude values outside [-180, 180] were clamped to that range.
      No results were found for the query: lat = 90, long = 180.

# reverse_geo_lite_sf() returns empty geometry without a match

    Code
      obj <- reverse_geo_lite_sf(89.999999, 179.9999)
    Message
      No results were found for the query: lat = 89.999999, long = 179.9999.

---

    Code
      obj_renamed <- reverse_geo_lite_sf(89.999999, 179.9999, address = "adddata")
    Message
      No results were found for the query: lat = 89.999999, long = 179.9999.

# reverse_geo_lite_sf() returns empty geometry after API failure

    Code
      several <- reverse_geo_lite_sf(40.75728, -73.98, full_results = TRUE)
    Message
      Could not reach the API endpoint: https://nominatim.openstreetmap.org/reverse?lat=40.75728&lon=-73.98&format=geojson&addressdetails=1.

# geo_lite() returns a typed row when no address matches

    Code
      obj <- geo_lite("xbzbzbzoa aiaia")
    Message
      No results were found for the query: xbzbzbzoa aiaia.

---

    Code
      obj_renamed <- geo_lite("xbzbzbzoa aiaia", lat = "lata", long = "longa")
    Message
      No results were found for the query: xbzbzbzoa aiaia.

# geo_lite() returns a typed row when the API is unavailable

    Code
      several <- geo_lite("Madrid", full_results = TRUE)
    Message
      Could not reach the API endpoint: https://nominatim.openstreetmap.org/search?q=Madrid&format=jsonv2&limit=1&addressdetails=1.

# bbox_to_poly() rejects incomplete bounding boxes

    Code
      bbox_to_poly()
    Condition
      Error in `bbox_to_poly()`:
      ! Provide `bbox` or non-missing values for `xmin`, `ymin`, `xmax` and `ymax`.

---

    Code
      bbox_to_poly(1)
    Condition
      Error in `bbox_to_poly()`:
      ! `bbox` must contain exactly four elements, but the provided value has 1.

---

    Code
      bbox_to_poly(xmin = 1)
    Condition
      Error in `bbox_to_poly()`:
      ! Provide `bbox` or non-missing values for `xmin`, `ymin`, `xmax` and `ymax`.

# geo_address_lookup() returns a typed row when no object matches

    Code
      obj <- geo_address_lookup(34633854, "N")
    Message
      No results were found for the query: N34633854.

---

    Code
      obj_renamed <- geo_address_lookup(34633854, "N", lat = "lata", long = "longa")
    Message
      No results were found for the query: N34633854.

# geo_address_lookup() matches multiple IDs and drops missing ones

    Code
      several <- geo_address_lookup(vector_ids, vector_type, verbose = TRUE)
    Condition
      Warning in `geo_address_lookup()`:
      No results were found for the following OSM IDs: J146. The output contains only matched IDs.

# geo_address_lookup() returns typed rows after API failure

    Code
      several <- geo_address_lookup(vector_ids, vector_type, full_results = TRUE)
    Message
      Could not reach the API endpoint: https://nominatim.openstreetmap.org/lookup?osm_ids=R343921,N240109189&format=jsonv2&addressdetails=1.

# geo_address_lookup_sf() returns empty geometry without a match

    Code
      obj <- geo_address_lookup_sf(34633854, "N")
    Message
      No results were found for the query: N34633854.

# geo_address_lookup_sf() drops unmatched IDs from multiple lookups

    Code
      several <- geo_address_lookup_sf(vector_ids, vector_type, verbose = TRUE)
    Condition
      Warning in `geo_address_lookup_sf()`:
      No results were found for the following OSM IDs: J146. The output contains only matched IDs.

# geo_address_lookup_sf() returns empty geometry after API failure

    Code
      several <- geo_address_lookup_sf(vector_ids, vector_type, full_results = TRUE)
    Message
      Could not reach the API endpoint: https://nominatim.openstreetmap.org/lookup?osm_ids=R343921,N240109189&format=geojson&addressdetails=1.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geo-address-lookup.R
\encoding{UTF-8}
\name{geo_address_lookup}
\alias{geo_address_lookup}
\title{Look up OpenStreetMap objects}
\usage{
geo_address_lookup(
  osm_ids,
  type = c("N", "W", "R"),
  lat = "lat",
  long = "lon",
  full_results = FALSE,
  return_addresses = TRUE,
  verbose = FALSE,
  nominatim_server = "https://nominatim.openstreetmap.org/",
  custom_query = list()
)
}
\arguments{
\item{osm_ids}{A numeric vector of OSM identifiers, for example
\code{c(12345, 67890)}.}

\item{type}{A character vector containing the OSM object type associated
with each value in \code{osm_ids}. Possible values are node (\code{"N"}), way
(\code{"W"}) and relation (\code{"R"}). A single value is recycled.}

\item{lat}{A character string specifying the name of the latitude column in
the output. Defaults to \code{"lat"}.}

\item{long}{A character string specifying the name of the longitude column
in the output. Defaults to \code{"lon"}.}

\item{full_results}{A logical value indicating whether to return all
available fields from the Nominatim API. If \code{FALSE}, only query metadata,
location data and requested address columns are returned.}

\item{return_addresses}{A logical value indicating whether to include
single-line addresses in the results.}

\item{verbose}{A logical value indicating whether to display detailed
messages in the console.}

\item{nominatim_server}{A character string specifying the base URL of the
Nominatim server. Defaults to
\code{"https://nominatim.openstreetmap.org/"}.}

\item{custom_query}{A named list of additional API parameters, for example
\code{list(countrycodes = "US")}. See \strong{Details}.}
}
\value{
A \code{\link[dplyr:tibble]{tibble}} with the results that match the query.
}
\description{
Looks up addresses and other details for one or more OpenStreetMap (OSM)
objects, such as nodes, ways or relations. Results are returned as a
\link[dplyr:tibble]{tibble}. Use \code{\link[=geo_address_lookup_sf]{geo_address_lookup_sf()}} to return an
\code{\link[sf:st_sf]{sf}} object instead.
}
\details{
See \url{https://nominatim.org/release-docs/latest/api/Lookup/} for additional
parameters to pass to \code{custom_query}.
}
\examples{
\dontshow{if (nominatim_check_access()) withAutoprint(\{ # examplesIf}
\donttest{
ids <- geo_address_lookup(osm_ids = c(46240148, 34633854), type = "W")

ids

several <- geo_address_lookup(c(146656, 240109189), type = c("R", "N"))
several
}
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=geo_lite]{geo_lite()}} for finding objects when their OSM identifiers are unknown.

Address lookup functions:
\code{\link[=geo_address_lookup_sf]{geo_address_lookup_sf()}}
}
\concept{lookup}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geo-amenity-sf.R
\encoding{UTF-8}
\name{geo_amenity_sf}
\alias{geo_amenity_sf}
\title{Look up OpenStreetMap amenities and return \CRANpkg{sf} objects}
\usage{
geo_amenity_sf(
  bbox,
  amenity,
  limit = 1,
  full_results = FALSE,
  return_addresses = TRUE,
  verbose = FALSE,
  nominatim_server = "https://nominatim.openstreetmap.org/",
  progressbar = TRUE,
  custom_query = list(),
  strict = FALSE,
  points_only = TRUE
)
}
\arguments{
\item{bbox}{A numeric vector, an \code{\link[sf:st_sf]{sf}} object or an
\code{\link[sf:st_sfc]{sfc}} object specifying a bounding box (viewbox) used to
limit the search. Numeric vectors must contain \strong{longitude} (\code{x}) and
\strong{latitude} (\code{y}) in the form \verb{(xmin, ymin, xmax, ymax)}. See \strong{Details}.}

\item{amenity}{A character vector of amenities to look up,
for example \code{c("pub", "restaurant")}. See \link{osm_amenities}.}

\item{limit}{A positive integer specifying the maximum number of results to
return per query. Nominatim returns at most 50 results per query.}

\item{full_results}{A logical value indicating whether to return all
available fields from the Nominatim API. If \code{FALSE}, only query metadata,
geometry and requested address columns are returned.}

\item{return_addresses}{A logical value indicating whether to include
single-line addresses in the results.}

\item{verbose}{A logical value indicating whether to display detailed
messages in the console.}

\item{nominatim_server}{A character string specifying the base URL of the
Nominatim server. Defaults to
\code{"https://nominatim.openstreetmap.org/"}.}

\item{progressbar}{A logical value indicating whether to display a progress
bar when processing multiple queries.}

\item{custom_query}{A named list of additional API parameters, for example
\code{list(countrycodes = "US")}. See \strong{Details}.}

\item{strict}{A logical value indicating whether to keep only results inside
\code{bbox}. If \code{FALSE} (the default), Nominatim may return results outside the
bounding box.}

\item{points_only}{A logical value indicating whether to return only point
geometries. If \code{FALSE}, the API may return other geometry types. See
\strong{About geometry types}.}
}
\value{
An \code{\link[sf:st_sf]{sf}} object with the results that match the query.
}
\description{
Looks up OpenStreetMap \link[=osm_amenities]{amenities} within a bounding box of
the form \verb{(xmin, ymin, xmax, ymax)}. Results are returned as an
\code{\link[sf:st_sf]{sf}} object. Use \code{\link[=geo_amenity]{geo_amenity()}} to return a
\link[dplyr:tibble]{tibble} instead.
}
\details{
Bounding boxes can be located using online tools such as
\url{https://boundingbox.klokantech.com/}.

For a full list of valid amenities, see
\url{https://wiki.openstreetmap.org/wiki/Key:amenity} and \link{osm_amenities}.

See \url{https://nominatim.org/release-docs/latest/api/Search/} for additional
parameters to pass to \code{custom_query}.
}
\section{About geometry types}{


The \code{points_only} argument controls whether the results contain only points.
All Nominatim results have at least a point geometry.

When \code{points_only = FALSE}, the geometry type depends on the matching
feature. Administrative areas and major buildings are returned as polygons,
rivers and roads are returned as lines and amenities may still be returned
as points.

This function is vectorized, allowing multiple addresses to be searched.
With \code{points_only = FALSE}, multiple geometry types may be returned.
}

\examples{
\dontshow{if (nominatim_check_access()) withAutoprint(\{ # examplesIf}
\donttest{
# Retrieve the Usera district in Madrid.

library(ggplot2)
mad <- geo_lite_sf("Usera, Madrid, Spain", points_only = FALSE)

# Search for restaurants, pubs and schools.

rest_pub <- geo_amenity_sf(mad, c("restaurant", "pub", "school"),
  limit = 50
)

if (!all(sf::st_is_empty(rest_pub))) {
  ggplot(mad) +
    geom_sf() +
    geom_sf(data = rest_pub, aes(color = query, shape = query))
}
}
\dontshow{\}) # examplesIf}
}
\seealso{
Amenity lookup functions and data:
\code{\link[=geo_amenity]{geo_amenity()}},
\code{\link{osm_amenities}}

Spatial helper and output functions:
\code{\link[=bbox_to_poly]{bbox_to_poly()}},
\code{\link[=geo_address_lookup_sf]{geo_address_lookup_sf()}},
\code{\link[=geo_lite_sf]{geo_lite_sf()}},
\code{\link[=geo_lite_struct_sf]{geo_lite_struct_sf()}},
\code{\link[=reverse_geo_lite_sf]{reverse_geo_lite_sf()}}
}
\concept{amenity}
\concept{spatial}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\encoding{UTF-8}
\name{osm_amenities}
\alias{osm_amenities}
\title{OpenStreetMap amenities}
\format{
A \link[dplyr:tibble]{tibble} with
140 rows and
three columns:
\describe{
\item{category}{Amenity category.}
\item{amenity}{Amenity value.}
\item{comment}{Brief description of the amenity type.}
}
}
\source{
\url{https://wiki.openstreetmap.org/wiki/Key:amenity}
}
\description{
A dataset of amenity values available on OpenStreetMap.
}
\note{
The data were extracted on \strong{July 11, 2026}. See
\code{inst/COPYRIGHTS} for copyright and license details.
}
\examples{
data("osm_amenities")

osm_amenities
}
\seealso{
Amenity lookup functions and data:
\code{\link[=geo_amenity]{geo_amenity()}},
\code{\link[=geo_amenity_sf]{geo_amenity_sf()}}
}
\concept{amenity}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geo-lite-sf.R
\encoding{UTF-8}
\name{geo_lite_sf}
\alias{geo_lite_sf}
\title{Search for addresses with free-form queries and return \CRANpkg{sf} objects}
\usage{
geo_lite_sf(
  address,
  limit = 1,
  return_addresses = TRUE,
  full_results = FALSE,
  verbose = FALSE,
  progressbar = TRUE,
  nominatim_server = "https://nominatim.openstreetmap.org/",
  custom_query = list(),
  points_only = TRUE
)
}
\arguments{
\item{address}{A character vector of single-line addresses, for example
\code{"1600 Pennsylvania Ave NW, Washington"} or
\code{c("Madrid", "Barcelona")}.}

\item{limit}{A positive integer specifying the maximum number of results to
return per query. Nominatim returns at most 50 results per query.}

\item{return_addresses}{A logical value indicating whether to include
single-line addresses in the results.}

\item{full_results}{A logical value indicating whether to return all
available fields from the Nominatim API. If \code{FALSE}, only query metadata,
geometry and requested address columns are returned.}

\item{verbose}{A logical value indicating whether to display detailed
messages in the console.}

\item{progressbar}{A logical value indicating whether to display a progress
bar when processing multiple queries.}

\item{nominatim_server}{A character string specifying the base URL of the
Nominatim server. Defaults to
\code{"https://nominatim.openstreetmap.org/"}.}

\item{custom_query}{A named list of additional API parameters, for example
\code{list(countrycodes = "US")}. See \strong{Details}.}

\item{points_only}{A logical value indicating whether to return only point
geometries. If \code{FALSE}, the API may return other geometry types. See
\strong{About geometry types}.}
}
\value{
An \code{\link[sf:st_sf]{sf}} object with the results that match the query.
}
\description{
Searches for addresses supplied as a character vector and returns matching
results as an \code{\link[sf:st_sf]{sf}} object. Use \code{\link[=geo_lite]{geo_lite()}} to return a
\link[dplyr:tibble]{tibble} instead.

This function performs the \strong{free-form address search} described in the
\href{https://nominatim.org/release-docs/latest/api/Search/}{API endpoint}.
}
\details{
See \url{https://nominatim.org/release-docs/latest/api/Search/} for additional
parameters to pass to \code{custom_query}.
}
\section{About geometry types}{


The \code{points_only} argument controls whether the results contain only points.
All Nominatim results have at least a point geometry.

When \code{points_only = FALSE}, the geometry type depends on the matching
feature. Administrative areas and major buildings are returned as polygons,
rivers and roads are returned as lines and amenities may still be returned
as points.

This function is vectorized, allowing multiple addresses to be searched.
With \code{points_only = FALSE}, multiple geometry types may be returned.
}

\examples{
\dontshow{if (nominatim_check_access()) withAutoprint(\{ # examplesIf}
\donttest{
# Return point geometries.
library(ggplot2)

string <- "Statue of Liberty, NY, USA"
sol <- geo_lite_sf(string)

if (!all(sf::st_is_empty(sol))) {
  ggplot(sol) +
    geom_sf()
}

sol_poly <- geo_lite_sf(string, points_only = FALSE)

if (!all(sf::st_is_empty(sol_poly))) {
  ggplot(sol_poly) +
    geom_sf() +
    geom_sf(data = sol, color = "red")
}
# Return multiple matches.

madrid <- geo_lite_sf("Comunidad de Madrid, Spain",
  limit = 2,
  points_only = FALSE, full_results = TRUE
)

if (!all(sf::st_is_empty(madrid))) {
  ggplot(madrid) +
    geom_sf(fill = NA)
}
}
\dontshow{\}) # examplesIf}
}
\seealso{
Address search functions:
\code{\link[=geo_lite]{geo_lite()}},
\code{\link[=geo_lite_struct]{geo_lite_struct()}},
\code{\link[=geo_lite_struct_sf]{geo_lite_struct_sf()}}

Spatial helper and output functions:
\code{\link[=bbox_to_poly]{bbox_to_poly()}},
\code{\link[=geo_address_lookup_sf]{geo_address_lookup_sf()}},
\code{\link[=geo_amenity_sf]{geo_amenity_sf()}},
\code{\link[=geo_lite_struct_sf]{geo_lite_struct_sf()}},
\code{\link[=reverse_geo_lite_sf]{reverse_geo_lite_sf()}}
}
\concept{geocoding}
\concept{spatial}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geo-lite-struct.R
\encoding{UTF-8}
\name{geo_lite_struct}
\alias{geo_lite_struct}
\title{Search for addresses with structured queries}
\usage{
geo_lite_struct(
  amenity = NULL,
  street = NULL,
  city = NULL,
  county = NULL,
  state = NULL,
  country = NULL,
  postalcode = NULL,
  lat = "lat",
  long = "lon",
  limit = 1,
  full_results = FALSE,
  return_addresses = TRUE,
  verbose = FALSE,
  nominatim_server = "https://nominatim.openstreetmap.org/",
  custom_query = list()
)
}
\arguments{
\item{amenity}{A character string specifying the name or type of amenity. See
\code{\link[=geo_amenity]{geo_amenity()}}.}

\item{street}{A character string specifying the house number and street name.}

\item{city}{A character string specifying the city.}

\item{county}{A character string specifying the county.}

\item{state}{A character string specifying the state.}

\item{country}{A character string specifying the country.}

\item{postalcode}{A character string specifying the postal code.}

\item{lat}{A character string specifying the name of the latitude column in
the output. Defaults to \code{"lat"}.}

\item{long}{A character string specifying the name of the longitude column
in the output. Defaults to \code{"lon"}.}

\item{limit}{A positive integer specifying the maximum number of results to
return per query. Nominatim returns at most 50 results per query.}

\item{full_results}{A logical value indicating whether to return all
available fields from the Nominatim API. If \code{FALSE}, only query metadata,
location data and requested address columns are returned.}

\item{return_addresses}{A logical value indicating whether to include
single-line addresses in the results.}

\item{verbose}{A logical value indicating whether to display detailed
messages in the console.}

\item{nominatim_server}{A character string specifying the base URL of the
Nominatim server. Defaults to
\code{"https://nominatim.openstreetmap.org/"}.}

\item{custom_query}{A named list of additional API parameters, for example
\code{list(countrycodes = "US")}. See \strong{Details}.}
}
\value{
A \code{\link[dplyr:tibble]{tibble}} with the results that match the query.
}
\description{
Searches for addresses already split into components and returns matching
results as a \link[dplyr:tibble]{tibble}. Use \code{\link[=geo_lite_struct_sf]{geo_lite_struct_sf()}} to return
an \code{\link[sf:st_sf]{sf}} object instead.

This function performs the \strong{structured address search} described in the
\href{https://nominatim.org/release-docs/latest/api/Search/}{API endpoint}. To
perform a free-form search, use \code{\link[=geo_lite]{geo_lite()}}.
}
\details{
A structured address search accepts an address already split into components.
Each argument represents an address field. All components are optional, so
provide only those relevant to the address you want to find.

See \url{https://nominatim.org/release-docs/latest/api/Search/} for additional
parameters to pass to \code{custom_query}.
}
\examples{
\dontshow{if (nominatim_check_access()) withAutoprint(\{ # examplesIf}
\donttest{
pl_mayor <- geo_lite_struct(
  street = "Plaza Mayor", country = "Spain",
  limit = 50, full_results = TRUE
)

dplyr::glimpse(pl_mayor)
}
\dontshow{\}) # examplesIf}
}
\seealso{
Address search functions:
\code{\link[=geo_lite]{geo_lite()}},
\code{\link[=geo_lite_sf]{geo_lite_sf()}},
\code{\link[=geo_lite_struct_sf]{geo_lite_struct_sf()}}
}
\concept{geocoding}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geo-lite.R
\encoding{UTF-8}
\name{geo_lite}
\alias{geo_lite}
\title{Search for addresses with free-form queries}
\usage{
geo_lite(
  address,
  lat = "lat",
  long = "lon",
  limit = 1,
  full_results = FALSE,
  return_addresses = TRUE,
  verbose = FALSE,
  nominatim_server = "https://nominatim.openstreetmap.org/",
  progressbar = TRUE,
  custom_query = list()
)
}
\arguments{
\item{address}{A character vector of single-line addresses, for example
\code{"1600 Pennsylvania Ave NW, Washington"} or
\code{c("Madrid", "Barcelona")}.}

\item{lat}{A character string specifying the name of the latitude column in
the output. Defaults to \code{"lat"}.}

\item{long}{A character string specifying the name of the longitude column
in the output. Defaults to \code{"lon"}.}

\item{limit}{A positive integer specifying the maximum number of results to
return per query. Nominatim returns at most 50 results per query.}

\item{full_results}{A logical value indicating whether to return all
available fields from the Nominatim API. If \code{FALSE}, only query metadata,
location data and requested address columns are returned.}

\item{return_addresses}{A logical value indicating whether to include
single-line addresses in the results.}

\item{verbose}{A logical value indicating whether to display detailed
messages in the console.}

\item{nominatim_server}{A character string specifying the base URL of the
Nominatim server. Defaults to
\code{"https://nominatim.openstreetmap.org/"}.}

\item{progressbar}{A logical value indicating whether to display a progress
bar when processing multiple queries.}

\item{custom_query}{A named list of additional API parameters, for example
\code{list(countrycodes = "US")}. See \strong{Details}.}
}
\value{
A \code{\link[dplyr:tibble]{tibble}} with the results that match the query.
}
\description{
Searches for addresses supplied as a character vector and returns matching
results as a \link[dplyr:tibble]{tibble}. Use \code{\link[=geo_lite_sf]{geo_lite_sf()}} to return an
\code{\link[sf:st_sf]{sf}} object instead.

This function performs the \strong{free-form address search} described in the
\href{https://nominatim.org/release-docs/latest/api/Search/}{API endpoint}.
}
\details{
See \url{https://nominatim.org/release-docs/latest/api/Search/} for additional
parameters to pass to \code{custom_query}.
}
\examples{
\dontshow{if (nominatim_check_access()) withAutoprint(\{ # examplesIf}
\donttest{
geo_lite("Madrid, Spain")

# Search for multiple addresses.
geo_lite(c("Madrid", "Barcelona"))

# Restrict the search to the United States and return all fields.
geo_lite(c("Madrid", "Barcelona"),
  custom_query = list(countrycodes = "US"),
  full_results = TRUE
)
}
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=reverse_geo_lite]{reverse_geo_lite()}} for reverse geocoding coordinates,
\code{\link[=geo_address_lookup]{geo_address_lookup()}} for looking up known OpenStreetMap identifiers and
\code{\link[=geo_amenity]{geo_amenity()}} for searching amenities within a bounding box.

Address search functions:
\code{\link[=geo_lite_sf]{geo_lite_sf()}},
\code{\link[=geo_lite_struct]{geo_lite_struct()}},
\code{\link[=geo_lite_struct_sf]{geo_lite_struct_sf()}}
}
\concept{geocoding}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nominatim-check-access.R
\encoding{UTF-8}
\name{nominatim_check_access}
\alias{nominatim_check_access}
\title{Check access to the Nominatim API}
\usage{
nominatim_check_access(
  nominatim_server = "https://nominatim.openstreetmap.org/"
)
}
\arguments{
\item{nominatim_server}{A character string specifying the base URL of the
Nominatim server. Defaults to
\code{"https://nominatim.openstreetmap.org/"}.}
}
\value{
A single logical value: \code{TRUE} if the API is available and \code{FALSE} otherwise.
}
\description{
Checks whether \R can access a Nominatim API server.
}
\examples{
\donttest{
nominatim_check_access()
}
}
\seealso{
\code{\link[=geo_lite]{geo_lite()}} for submitting search requests and the
\href{https://nominatim.org/release-docs/latest/api/Status/}{Nominatim status endpoint}
for server status details.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geo-address-lookup-sf.R
\encoding{UTF-8}
\name{geo_address_lookup_sf}
\alias{geo_address_lookup_sf}
\title{Look up OpenStreetMap objects and return \CRANpkg{sf} objects}
\usage{
geo_address_lookup_sf(
  osm_ids,
  type = c("N", "W", "R"),
  full_results = FALSE,
  return_addresses = TRUE,
  verbose = FALSE,
  nominatim_server = "https://nominatim.openstreetmap.org/",
  custom_query = list(),
  points_only = TRUE
)
}
\arguments{
\item{osm_ids}{A numeric vector of OSM identifiers, for example
\code{c(12345, 67890)}.}

\item{type}{A character vector containing the OSM object type associated
with each value in \code{osm_ids}. Possible values are node (\code{"N"}), way
(\code{"W"}) and relation (\code{"R"}). A single value is recycled.}

\item{full_results}{A logical value indicating whether to return all
available fields from the Nominatim API. If \code{FALSE}, only query metadata,
geometry and requested address columns are returned.}

\item{return_addresses}{A logical value indicating whether to include
single-line addresses in the results.}

\item{verbose}{A logical value indicating whether to display detailed
messages in the console.}

\item{nominatim_server}{A character string specifying the base URL of the
Nominatim server. Defaults to
\code{"https://nominatim.openstreetmap.org/"}.}

\item{custom_query}{A named list of additional API parameters, for example
\code{list(countrycodes = "US")}. See \strong{Details}.}

\item{points_only}{A logical value indicating whether to return only point
geometries. If \code{FALSE}, the API may return other geometry types. See
\strong{About geometry types}.}
}
\value{
An \code{\link[sf:st_sf]{sf}} object with the results that match the query.
}
\description{
Looks up addresses and other details for one or more OpenStreetMap (OSM)
objects, such as nodes, ways or relations. Results are returned as an
\code{\link[sf:st_sf]{sf}} object. Use \code{\link[=geo_address_lookup]{geo_address_lookup()}} to return a
\link[dplyr:tibble]{tibble} instead.
}
\details{
See \url{https://nominatim.org/release-docs/latest/api/Lookup/} for additional
parameters to pass to \code{custom_query}.
}
\section{About geometry types}{


The \code{points_only} argument controls whether the results contain only points.
All Nominatim results have at least a point geometry.

When \code{points_only = FALSE}, the geometry type depends on the matching
feature. Administrative areas and major buildings are returned as polygons,
rivers and roads are returned as lines and amenities may still be returned
as points.

This function is vectorized, allowing multiple addresses to be searched.
With \code{points_only = FALSE}, multiple geometry types may be returned.
}

\examples{
\dontshow{if (nominatim_check_access()) withAutoprint(\{ # examplesIf}
\donttest{
# Look up Notre-Dame Cathedral in Paris.

NotreDame <- geo_address_lookup_sf(osm_ids = 201611261, type = "W")

# Require at least one non-empty object.
if (!all(sf::st_is_empty(NotreDame))) {
  library(ggplot2)

  ggplot(NotreDame) +
    geom_sf()
}

NotreDame_poly <- geo_address_lookup_sf(201611261,
  type = "W",
  points_only = FALSE
)

if (!all(sf::st_is_empty(NotreDame_poly))) {
  ggplot(NotreDame_poly) +
    geom_sf()
}

# Look up multiple OSM objects.

several <- geo_address_lookup_sf(c(146656, 240109189), type = c("R", "N"))
several
}
\dontshow{\}) # examplesIf}
}
\seealso{
Address lookup functions:
\code{\link[=geo_address_lookup]{geo_address_lookup()}}

Spatial helper and output functions:
\code{\link[=bbox_to_poly]{bbox_to_poly()}},
\code{\link[=geo_amenity_sf]{geo_amenity_sf()}},
\code{\link[=geo_lite_sf]{geo_lite_sf()}},
\code{\link[=geo_lite_struct_sf]{geo_lite_struct_sf()}},
\code{\link[=reverse_geo_lite_sf]{reverse_geo_lite_sf()}}
}
\concept{lookup}
\concept{spatial}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reverse-geo-lite.R
\encoding{UTF-8}
\name{reverse_geo_lite}
\alias{reverse_geo_lite}
\title{Reverse geocode coordinates}
\usage{
reverse_geo_lite(
  lat,
  long,
  address = "address",
  full_results = FALSE,
  return_coords = TRUE,
  verbose = FALSE,
  nominatim_server = "https://nominatim.openstreetmap.org/",
  progressbar = TRUE,
  custom_query = list()
)
}
\arguments{
\item{lat}{A numeric vector of latitude values in the range
\eqn{\left[-90, 90 \right]}.}

\item{long}{A numeric vector of longitude values in the range
\eqn{\left[-180, 180 \right]}.}

\item{address}{A character string specifying the name of the address column
in the output. Defaults to \code{"address"}.}

\item{full_results}{A logical value indicating whether to return all
available fields from the Nominatim API. If \code{FALSE}, only query metadata,
location data and requested address columns are returned.}

\item{return_coords}{A logical value indicating whether to return the input
coordinates with the results.}

\item{verbose}{A logical value indicating whether to display detailed
messages in the console.}

\item{nominatim_server}{A character string specifying the base URL of the
Nominatim server. Defaults to
\code{"https://nominatim.openstreetmap.org/"}.}

\item{progressbar}{A logical value indicating whether to display a progress
bar when processing multiple queries.}

\item{custom_query}{A named list of API-specific parameters, for example
\code{list(zoom = 3)}. See \strong{Details}.}
}
\value{
A \code{\link[dplyr:tibble]{tibble}} with the results that match the query.
}
\description{
Reverse geocodes latitude and longitude coordinates and returns matching
results as a \link[dplyr:tibble]{tibble}. Latitude values must be in
\eqn{\left[-90, 90 \right]} and longitudes in
\eqn{\left[-180, 180 \right]}. Use \code{\link[=reverse_geo_lite_sf]{reverse_geo_lite_sf()}} to return an
\code{\link[sf:st_sf]{sf}} object instead.
}
\details{
See \url{https://nominatim.org/release-docs/latest/api/Reverse/} for additional
parameters to pass to \code{custom_query}.
}
\section{About zooming}{


Set \code{custom_query = list(zoom = 3)} to adjust the output. Selected zoom
levels correspond to these address details:\tabular{ll}{
   \strong{zoom} \tab \strong{address_detail} \cr
   \code{3} \tab country \cr
   \code{5} \tab state \cr
   \code{8} \tab county \cr
   \code{10} \tab city \cr
   \code{14} \tab suburb \cr
   \code{16} \tab major streets \cr
   \code{17} \tab major and minor streets \cr
   \code{18} \tab building \cr
}
}

\examples{
\dontshow{if (nominatim_check_access()) withAutoprint(\{ # examplesIf}
\donttest{
reverse_geo_lite(lat = 40.75728, long = -73.98586)

# Reverse geocode multiple coordinate pairs.
reverse_geo_lite(lat = c(40.75728, 55.95335), long = c(-73.98586, -3.188375))

# Set the zoom to the country level.
sev <- reverse_geo_lite(
  lat = c(40.75728, 55.95335), long = c(-73.98586, -3.188375),
  custom_query = list(zoom = 0, extratags = TRUE),
  verbose = TRUE, full_results = TRUE
)

dplyr::glimpse(sev)
}
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=geo_lite]{geo_lite()}} for geocoding addresses.

Reverse geocoding functions:
\code{\link[=reverse_geo_lite_sf]{reverse_geo_lite_sf()}}
}
\concept{reverse}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nominatimlite-package.R
\docType{package}
\encoding{UTF-8}
\name{nominatimlite-package}
\alias{nominatimlite}
\alias{nominatimlite-package}
\title{nominatimlite: Interface to the 'Nominatim' API}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

Provides a lightweight interface to the 'Nominatim' API \url{https://nominatim.org/release-docs/latest/}. It supports free-form and structured address searches, searches for addresses from coordinates, amenity lookup and address lookup by 'OpenStreetMap' object identifier. It returns results as 'tibble' data frames or 'sf' objects.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://dieghernan.github.io/nominatimlite/}
  \item \url{https://github.com/dieghernan/nominatimlite}
  \item Report bugs at \url{https://github.com/dieghernan/nominatimlite/issues}
}

}
\author{
\strong{Maintainer}: Diego Hernangómez \email{diego.hernangomezherrero@gmail.com} (\href{https://orcid.org/0000-0001-8457-4658}{ORCID}) [copyright holder]

Authors:
\itemize{
  \item Diego Hernangómez \email{diego.hernangomezherrero@gmail.com} (\href{https://orcid.org/0000-0001-8457-4658}{ORCID}) [copyright holder]
}

Other contributors:
\itemize{
  \item Jindra Lacko (\href{https://orcid.org/0000-0002-0375-5156}{ORCID}) [contributor, reviewer]
  \item Alex White [contributor]
  \item OpenStreetMap contributors (For the osm_amenities data) [copyright holder]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geo-amenity.R
\encoding{UTF-8}
\name{geo_amenity}
\alias{geo_amenity}
\title{Look up OpenStreetMap amenities}
\usage{
geo_amenity(
  bbox,
  amenity,
  lat = "lat",
  long = "lon",
  limit = 1,
  full_results = FALSE,
  return_addresses = TRUE,
  verbose = FALSE,
  nominatim_server = "https://nominatim.openstreetmap.org/",
  progressbar = TRUE,
  custom_query = list(),
  strict = FALSE
)
}
\arguments{
\item{bbox}{A numeric vector, an \code{\link[sf:st_sf]{sf}} object or an
\code{\link[sf:st_sfc]{sfc}} object specifying a bounding box (viewbox) used to
limit the search. Numeric vectors must contain \strong{longitude} (\code{x}) and
\strong{latitude} (\code{y}) in the form \verb{(xmin, ymin, xmax, ymax)}. See \strong{Details}.}

\item{amenity}{A character vector of amenities to look up,
for example \code{c("pub", "restaurant")}. See \link{osm_amenities}.}

\item{lat}{A character string specifying the name of the latitude column in
the output. Defaults to \code{"lat"}.}

\item{long}{A character string specifying the name of the longitude column
in the output. Defaults to \code{"lon"}.}

\item{limit}{A positive integer specifying the maximum number of results to
return per query. Nominatim returns at most 50 results per query.}

\item{full_results}{A logical value indicating whether to return all
available fields from the Nominatim API. If \code{FALSE}, only query metadata,
location data and requested address columns are returned.}

\item{return_addresses}{A logical value indicating whether to include
single-line addresses in the results.}

\item{verbose}{A logical value indicating whether to display detailed
messages in the console.}

\item{nominatim_server}{A character string specifying the base URL of the
Nominatim server. Defaults to
\code{"https://nominatim.openstreetmap.org/"}.}

\item{progressbar}{A logical value indicating whether to display a progress
bar when processing multiple queries.}

\item{custom_query}{A named list of additional API parameters, for example
\code{list(countrycodes = "US")}. See \strong{Details}.}

\item{strict}{A logical value indicating whether to keep only results inside
\code{bbox}. If \code{FALSE} (the default), Nominatim may return results outside the
bounding box.}
}
\value{
A \code{\link[dplyr:tibble]{tibble}} with the results that match the query.
}
\description{
Looks up OpenStreetMap \link[=osm_amenities]{amenities} within a bounding box of
the form \verb{(xmin, ymin, xmax, ymax)}. Results are returned as a
\link[dplyr:tibble]{tibble}. Use \code{\link[=geo_amenity_sf]{geo_amenity_sf()}} to return an
\code{\link[sf:st_sf]{sf}} object instead.
}
\details{
Bounding boxes can be located using online tools such as
\url{https://boundingbox.klokantech.com/}.

For a full list of valid amenities, see
\url{https://wiki.openstreetmap.org/wiki/Key:amenity} and \link{osm_amenities}.

See \url{https://nominatim.org/release-docs/latest/api/Search/} for additional
parameters to pass to \code{custom_query}.
}
\examples{
\dontshow{if (nominatim_check_access()) withAutoprint(\{ # examplesIf}
\donttest{
# Define a bounding box around Times Square, New York.
bbox <- c(
  -73.9894467311, 40.75573629,
  -73.9830630737, 40.75789245
)

geo_amenity(
  bbox = bbox,
  amenity = "restaurant"
)

# Search for multiple amenities.
geo_amenity(
  bbox = bbox,
  amenity = c("restaurant", "pub")
)

# Increase `limit` and use strict filtering.
geo_amenity(
  bbox = bbox,
  amenity = c("restaurant", "pub"),
  limit = 10,
  strict = TRUE
)
}
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=bbox_to_poly]{bbox_to_poly()}} for converting bounding box coordinates to a
\CRANpkg{sf} polygon.

Amenity lookup functions and data:
\code{\link[=geo_amenity_sf]{geo_amenity_sf()}},
\code{\link{osm_amenities}}
}
\concept{amenity}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bbox-to-poly.R
\encoding{UTF-8}
\name{bbox_to_poly}
\alias{bbox_to_poly}
\title{Convert a bounding box to an \code{\link[sf:st_sfc]{sfc}} \code{POLYGON} object}
\usage{
bbox_to_poly(bbox = NA, xmin = NA, ymin = NA, xmax = NA, ymax = NA, crs = 4326)
}
\arguments{
\item{bbox}{A numeric vector of four bounding box coordinates in the form
\code{c(xmin, ymin, xmax, ymax)}.}

\item{xmin, ymin, xmax, ymax}{A numeric value specifying an individual bounding
box coordinate. Use these arguments as an alternative to \code{bbox}.}

\item{crs}{coordinate reference system, something suitable as input to \link[sf]{st_crs}}
}
\value{
An \code{\link[sf:st_sfc]{sfc}} object with \code{POLYGON} geometry and the coordinate
reference system specified by \code{crs}.
}
\description{
Converts bounding box coordinates to an \code{\link[sf:st_sfc]{sfc}} object with
\code{POLYGON} geometry.
}
\details{
Bounding boxes can be located using online tools such as
\url{https://boundingbox.klokantech.com/}.
}
\examples{
\dontshow{if (nominatim_check_access() && requireNamespace("ggplot2")) withAutoprint(\{ # examplesIf}
# Convert the bounding box for Germany.
bbox_GER <- c(5.86631529, 47.27011137, 15.04193189, 55.09916098)

bbox_GER_sf <- bbox_to_poly(bbox_GER)

library(ggplot2)

ggplot(bbox_GER_sf) +
  geom_sf()
\donttest{
# Extract the bounding box of an `sf` object.
sfobj <- geo_lite_sf("seychelles", points_only = FALSE)

sfobj

# Require at least one non-empty object.
if (!all(sf::st_is_empty(sfobj))) {
  bbox <- sf::st_bbox(sfobj)

  bbox

  bbox_sfobj <- bbox_to_poly(bbox)

  ggplot(bbox_sfobj) +
    geom_sf(fill = "lightblue", alpha = 0.5) +
    geom_sf(data = sfobj, fill = "wheat")
}
}
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[sf:st_as_sfc]{sf::st_as_sfc()}} and \code{\link[sf:st_sfc]{sf::st_sfc()}}.

Spatial helper and output functions:
\code{\link[=geo_address_lookup_sf]{geo_address_lookup_sf()}},
\code{\link[=geo_amenity_sf]{geo_amenity_sf()}},
\code{\link[=geo_lite_sf]{geo_lite_sf()}},
\code{\link[=geo_lite_struct_sf]{geo_lite_struct_sf()}},
\code{\link[=reverse_geo_lite_sf]{reverse_geo_lite_sf()}}
}
\concept{spatial}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reverse-geo-lite-sf.R
\encoding{UTF-8}
\name{reverse_geo_lite_sf}
\alias{reverse_geo_lite_sf}
\title{Reverse geocode coordinates and return \CRANpkg{sf} objects}
\usage{
reverse_geo_lite_sf(
  lat,
  long,
  address = "address",
  full_results = FALSE,
  return_coords = TRUE,
  verbose = FALSE,
  nominatim_server = "https://nominatim.openstreetmap.org/",
  progressbar = TRUE,
  custom_query = list(),
  points_only = TRUE
)
}
\arguments{
\item{lat}{A numeric vector of latitude values in the range
\eqn{\left[-90, 90 \right]}.}

\item{long}{A numeric vector of longitude values in the range
\eqn{\left[-180, 180 \right]}.}

\item{address}{A character string specifying the name of the address column
in the output. Defaults to \code{"address"}.}

\item{full_results}{A logical value indicating whether to return all
available fields from the Nominatim API. If \code{FALSE}, only query metadata,
geometry and requested address columns are returned.}

\item{return_coords}{A logical value indicating whether to return the input
coordinates with the results.}

\item{verbose}{A logical value indicating whether to display detailed
messages in the console.}

\item{nominatim_server}{A character string specifying the base URL of the
Nominatim server. Defaults to
\code{"https://nominatim.openstreetmap.org/"}.}

\item{progressbar}{A logical value indicating whether to display a progress
bar when processing multiple queries.}

\item{custom_query}{A named list of API-specific parameters, for example
\code{list(zoom = 3)}. See \strong{Details}.}

\item{points_only}{A logical value indicating whether to return only point
geometries. If \code{FALSE}, the API may return other geometry types. See
\strong{About geometry types}.}
}
\value{
An \code{\link[sf:st_sf]{sf}} object with the results that match the query.
}
\description{
Reverse geocodes latitude and longitude coordinates and returns matching
results as an \code{\link[sf:st_sf]{sf}} object. Latitude values must be in
\eqn{\left[-90, 90 \right]} and longitude values in
\eqn{\left[-180, 180 \right]}. Use \code{\link[=reverse_geo_lite]{reverse_geo_lite()}} to return a
\link[dplyr:tibble]{tibble} instead.
}
\details{
See \url{https://nominatim.org/release-docs/latest/api/Reverse/} for additional
parameters to pass to \code{custom_query}.
}
\section{About zooming}{


Set \code{custom_query = list(zoom = 3)} to adjust the output. Selected zoom
levels correspond to these address details:\tabular{ll}{
   \strong{zoom} \tab \strong{address_detail} \cr
   \code{3} \tab country \cr
   \code{5} \tab state \cr
   \code{8} \tab county \cr
   \code{10} \tab city \cr
   \code{14} \tab suburb \cr
   \code{16} \tab major streets \cr
   \code{17} \tab major and minor streets \cr
   \code{18} \tab building \cr
}
}

\section{About geometry types}{


The \code{points_only} argument controls whether the results contain only points.
All Nominatim results have at least a point geometry.

When \code{points_only = FALSE}, the geometry type depends on the matching
feature. Administrative areas and major buildings are returned as polygons,
rivers and roads are returned as lines and amenities may still be returned
as points.

This function is vectorized, allowing multiple addresses to be searched.
With \code{points_only = FALSE}, multiple geometry types may be returned.
}

\examples{
\dontshow{if (nominatim_check_access()) withAutoprint(\{ # examplesIf}
\donttest{
library(ggplot2)

# Define the Colosseum coordinates.
col_lon <- 12.49309
col_lat <- 41.89026

# Return the Colosseum as a polygon.
col_sf <- reverse_geo_lite_sf(
  lat = col_lat,
  long = col_lon,
  points_only = FALSE
)

dplyr::glimpse(col_sf)

if (!all(sf::st_is_empty(col_sf))) {
  ggplot(col_sf) +
    geom_sf()
}

# Return the city of Rome by using the same coordinates with zoom 10.

rome_sf <- reverse_geo_lite_sf(
  lat = col_lat,
  long = col_lon,
  custom_query = list(zoom = 10),
  points_only = FALSE
)

dplyr::glimpse(rome_sf)

if (!all(sf::st_is_empty(rome_sf))) {
  ggplot(rome_sf) +
    geom_sf()
}
}
\dontshow{\}) # examplesIf}
}
\seealso{
Reverse geocoding functions:
\code{\link[=reverse_geo_lite]{reverse_geo_lite()}}

Spatial helper and output functions:
\code{\link[=bbox_to_poly]{bbox_to_poly()}},
\code{\link[=geo_address_lookup_sf]{geo_address_lookup_sf()}},
\code{\link[=geo_amenity_sf]{geo_amenity_sf()}},
\code{\link[=geo_lite_sf]{geo_lite_sf()}},
\code{\link[=geo_lite_struct_sf]{geo_lite_struct_sf()}}
}
\concept{reverse}
\concept{spatial}
